
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLibrary.h"
#include "dbPCellHeader.h"
#include "edtInstPropertiesPage.h"
#include "edtPCellParametersPage.h"
#include "edtDialogs.h"
#include "layDialogs.h"
#include "layObjectInstPath.h"
#include "layLayoutView.h"
#include "layCellSelectionForm.h"
#include "layExceptions.h"
#include "tlString.h"

#include <string>

namespace edt
{


// -------------------------------------------------------------------------
//  helper function to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

static db::Coord
coord_from_dcoord (double d, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DCoord dc = t.ctrans (d / dbu);
  db::Coord c = db::Coord (floor (dc + 0.5));
#if ! defined(HAVE_64BIT_COORD)   // for 64bit there is not accuracy reserve
  if (fabs (dc - c) > 1e-6) { 
    if (! du) {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit (%g)").c_str (), d, dbu);
    } else {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit").c_str (), d);
    }
  }
#endif

  return c;
}

static db::Point
coord_from_dpoint (const db::DPoint &dp, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DPoint dpt = t * (dp * (1.0 / dbu));

  db::Coord cx = coord_from_dcoord (dpt.x (), 1.0, false, db::CplxTrans ());
  db::Coord cy = coord_from_dcoord (dpt.y (), 1.0, false, db::CplxTrans ());

  return db::Point (cx, cy);
}

// -------------------------------------------------------------------------
//  InstPropertiesPage implementation

InstPropertiesPage::InstPropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), mp_service (service), m_enable_cb_callback (true), mp_pcell_parameters (0)
{
  m_selection_ptrs.reserve (service->selection ().size ());
  for (edt::Service::obj_iterator s = service->selection ().begin (); s != service->selection ().end (); ++s) {
    m_selection_ptrs.push_back (s);
  }
  m_index = 0;
  m_prop_id = 0;
  mp_service->clear_highlights ();

  setupUi (this);
  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (sel_pb, SIGNAL (clicked ()), this, SLOT (show_cell ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (browse_pb, SIGNAL (clicked ()), this, SLOT (browse_cell ()));
  connect (lib_cbx, SIGNAL (currentIndexChanged (int)), this, SLOT (library_changed (int)));
  connect (cell_name_le, SIGNAL (textChanged (const QString &)), this, SLOT (cell_name_changed (const QString &)));

  QHBoxLayout *layout = new QHBoxLayout (pcell_tab);
  layout->setMargin (0);
  pcell_tab->setLayout (layout);
}

InstPropertiesPage::~InstPropertiesPage ()
{
  mp_service->restore_highlights ();
}

void
InstPropertiesPage::library_changed (int)
{
BEGIN_PROTECTED
  update_pcell_parameters ();
END_PROTECTED
}

void
InstPropertiesPage::cell_name_changed (const QString &)
{
BEGIN_PROTECTED
  update_pcell_parameters ();
END_PROTECTED
}

void
InstPropertiesPage::browse_cell ()
{
BEGIN_PROTECTED

  //  find the layout the cell has to be looked up: that is either the layout of the current instance or 
  //  the library selected
  db::Layout *layout = 0;
  db::Library *lib = 0;
  if (lib_cbx->current_library ()) {
    lib = lib_cbx->current_library ();
    layout = &lib->layout ();
  } else {
    edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
    const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
    layout = &cv->layout ();
  }

  lay::LibraryCellSelectionForm form (this, layout, "browse_lib_cell");
  if (lib) {
    form.setWindowTitle (tl::to_qstring (tl::translate ("Select Cell - Library: ") + lib->get_description ()));
  }

  std::pair<bool, db::pcell_id_type> pc = layout->pcell_by_name (tl::to_string (cell_name_le->text ()).c_str ());
  if (pc.first) {
    form.set_selected_pcell_id (pc.second);
  } else {
    std::pair<bool, db::cell_index_type> c = layout->cell_by_name (tl::to_string (cell_name_le->text ()).c_str ());
    if (c.first) {
      form.set_selected_cell_index (c.second);
    }
  }

  if (form.exec ()) {
    if (form.selected_cell_is_pcell ()) {
      cell_name_le->setText (tl::to_qstring (layout->pcell_header (form.selected_pcell_id ())->get_name ()));
    } else if (layout->is_valid_cell_index (form.selected_cell_index ())) {
      cell_name_le->setText (tl::to_qstring (layout->cell_name (form.selected_cell_index ())));
    }
    update_pcell_parameters ();
  }

END_PROTECTED
}

void
InstPropertiesPage::show_props ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  lay::UserPropertiesForm props_form (this);
  if (props_form.show (mp_service->view (), pos->cv_index (), m_prop_id)) {
    if (! readonly () && m_prop_id != pos->back ().inst_ptr.prop_id ()) {

      const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
      db::Cell &cell = cv->layout ().cell (pos->cell_index ());

      db::Transaction transaction (cv->layout ().manager (), tl::translate ("Change instance properties"));

      db::Instance new_inst = cell.replace_prop_id (pos->back ().inst_ptr, m_prop_id);

      if (new_inst != pos->back ().inst_ptr) {

        //  change selection to new shape
        lay::ObjectInstPath new_sel (*pos);
        new_sel.back ().inst_ptr = new_inst;

        mp_service->select (*pos, lay::Editable::Reset);
        mp_service->select (new_sel, lay::Editable::Add);

        // locate the new selection
        for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
          ;
        tl_assert (pos != mp_service->selection ().end ());
        m_selection_ptrs [m_index] = pos;

        mp_service->selection_to_view ();

      }

      update ();

    }
  }
}

void
InstPropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  mp_service->view ()->dbu_coordinates (dbu_cb->isChecked ());
  mp_service->view ()->absolute_coordinates (abs_cb->isChecked ());

  update ();
}

void 
InstPropertiesPage::back ()
{
  m_index = m_selection_ptrs.size ();
}

void 
InstPropertiesPage::front ()
{
  m_index = 0;
}

bool 
InstPropertiesPage::at_begin () const
{
  return (m_index == 0);
}

bool 
InstPropertiesPage::at_end () const
{
  return (m_index == m_selection_ptrs.size ());
}

void 
InstPropertiesPage::operator-- ()
{
  --m_index;
}

void 
InstPropertiesPage::operator++ ()
{
  ++m_index;
}

void 
InstPropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

void 
InstPropertiesPage::update ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
  tl_assert (pos->is_cell_inst ());

  mp_service->highlight (m_index);

  m_enable_cb_callback = false;
  dbu_cb->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_cb->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  double dbu = cv->layout ().dbu ();

  std::string info (tl::translate ("Cell "));
  info += cv->layout ().cell_name (pos->cell_index ());
  info_lbl->setText (tl::to_qstring (info));

  db::Layout *def_layout = &cv->layout ();
  db::cell_index_type def_cell_index = pos->back ().inst_ptr.cell_index ();

  std::pair<db::Library *, db::cell_index_type> dl = def_layout->defining_library (def_cell_index);
  lib_cbx->set_current_library (dl.first);
  if (dl.first) {
    def_layout = &dl.first->layout ();
    def_cell_index = dl.second;
  }

  std::pair<bool, db::pcell_id_type> pci = def_layout->is_pcell_instance (def_cell_index);
  if (pci.first && def_layout->pcell_declaration (pci.second)) {
    cell_name_le->setText (tl::to_qstring (def_layout->pcell_header (pci.second)->get_name ()));
  } else {
    cell_name_le->setText (tl::to_qstring (def_layout->cell_name (def_cell_index)));
  }

  db::Point rowv, columnv;
  unsigned long rows, columns;

  db::CplxTrans gt;
  if (abs_cb->isChecked ()) {
    gt = pos->trans ();
  }

  bool du = dbu_cb->isChecked ();

  db::Trans t (pos->back ().inst_ptr.front ());

  if (pos->back ().inst_ptr.is_regular_array (rowv, columnv, rows, columns)) {

    array_grp->setChecked (true);
    rows_le->setText (tl::to_qstring (tl::to_string (rows)));
    columns_le->setText (tl::to_qstring (tl::to_string (columns)));
    row_x_le->setText (tl::to_qstring (coord_to_string ((gt * rowv).x (), dbu, du)));
    row_y_le->setText (tl::to_qstring (coord_to_string ((gt * rowv).y (), dbu, du)));
    column_x_le->setText (tl::to_qstring (coord_to_string ((gt * columnv).x (), dbu, du)));
    column_y_le->setText (tl::to_qstring (coord_to_string ((gt * columnv).y (), dbu, du)));

    if (! pos->back ().array_inst.at_end ()) {

      //  show the array indices
      long row = pos->back ().array_inst.index_a ();
      long column = pos->back ().array_inst.index_b ();
      inst_lbl->setText (tl::to_qstring (tl::sprintf (tl::translate ("This is instance [%ld,%ld] of array with"), row, column)));

    }

  } else {

    array_grp->setChecked (false);
    rows_le->setText (QString ());
    columns_le->setText (QString ());
    row_x_le->setText (QString ());
    row_y_le->setText (QString ());
    column_x_le->setText (QString ());
    column_y_le->setText (QString ());
    inst_lbl->setText (QString ());

  }

  pos_x_le->setText (tl::to_qstring (coord_to_string ((gt * t.disp ()).x (), dbu, du)));
  pos_y_le->setText (tl::to_qstring (coord_to_string ((gt * t.disp ()).y (), dbu, du)));
  angle_le->setText (tl::to_qstring (tl::to_string (pos->back ().inst_ptr.complex_trans ().angle ())));
  mirror_cbx->setChecked (t.is_mirror ());
  mag_le->setText (tl::to_qstring (tl::to_string (pos->back ().inst_ptr.complex_trans ().mag ())));

  m_prop_id = pos->back ().inst_ptr.prop_id ();

  update_pcell_parameters ();
}

void 
InstPropertiesPage::show_cell ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  lay::CellView::unspecific_cell_path_type path (mp_service->view ()->cellview (pos->cv_index ()).combined_unspecific_path ());
  for (lay::ObjectInstPath::iterator p = pos->begin (); p != pos->end (); ++p) {
    path.push_back (p->inst_ptr.cell_index ());
  }

  mp_service->view ()->set_current_cell_path (pos->cv_index (), path);
}

void
InstPropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), *m_selection_ptrs [m_index]);
}

bool 
InstPropertiesPage::readonly ()
{
  return ! mp_service->view ()->is_editable ();
}

void 
InstPropertiesPage::apply ()
{
  lay::LayerState layer_state = mp_service->view ()->layer_snapshot ();

  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  bool du = dbu_cb->isChecked ();
  double dbu = cv->layout ().dbu ();

  db::Layout *layout;
  db::Library *lib = lib_cbx->current_library ();

  //  find the layout the cell has to be looked up: that is either the layout of the current instance or 
  //  the library selected
  if (lib) {

    layout = &lib->layout ();

  } else {

    edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
    const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
    layout = &cv->layout ();

  }

  std::pair<bool, db::cell_index_type> ci = layout->cell_by_name (tl::to_string (cell_name_le->text ()).c_str ());
  std::pair<bool, db::pcell_id_type> pci = layout->pcell_by_name (tl::to_string (cell_name_le->text ()).c_str ());
  if (! ci.first && ! pci.first) {
    throw tl::Exception (tl::translate ("Not a valid cell name: %s").c_str (), tl::to_string (cell_name_le->text ()).c_str ());
  }

  db::Cell &cell = cv->layout ().cell (pos->cell_index ());

  db::cell_index_type inst_cell_index = ci.second;

  //  instantiate the PCell
  if (pci.first) {
    tl_assert (mp_pcell_parameters != 0);
    tl_assert (layout->pcell_declaration (pci.second) == mp_pcell_parameters->pcell_decl ());
    inst_cell_index = layout->get_pcell_variant (pci.second, mp_pcell_parameters->get_parameters ());
  }

  //  reference the library
  if (lib) {
    layout = & cv->layout ();
    inst_cell_index = layout->get_lib_proxy (lib, inst_cell_index);
  }

  //  detect recursions in the hierarchy
  std::set<db::cell_index_type> called;
  cv->layout ().cell (inst_cell_index).collect_called_cells (called);
  if (inst_cell_index == cell.cell_index () || called.find (cell.cell_index ()) != called.end ()) {
    throw tl::Exception (tl::translate ("Trying to build a recursive hierarchy").c_str ());
  }

  double x = 0.0, y = 0.0;
  tl::from_string (tl::to_string (pos_x_le->text ()), x);
  tl::from_string (tl::to_string (pos_y_le->text ()), y);

  db::CplxTrans t;
  if (abs_cb->isChecked ()) {
    t = pos->trans ().inverted ();
  }

  db::Point disp (coord_from_dpoint (db::DPoint (x, y), dbu, du, t));

  bool mirror = mirror_cbx->isChecked ();
  double angle_deg = 0.0;
  tl::from_string (tl::to_string (angle_le->text ()), angle_deg);

  double mag = 0.0;
  tl::from_string (tl::to_string (mag_le->text ()), mag);

  angle_deg -= (floor (angle_deg / 360.0) + 1.0) * 360.0;
  while (angle_deg < -1e-6) {
    angle_deg += 360.0;
  }

  unsigned int angle = (unsigned int) floor (angle_deg / 90.0 + 1e-6);
  bool is_complex = ((fabs (mag - 1.0) > 1e-6 || fabs (angle_deg - angle * 90.0) > 1e-6));

  db::CellInstArray inst;

  if (array_grp->isChecked ()) {

    double cx = 0.0, cy = 0.0;
    double rx = 0.0, ry = 0.0;
    unsigned long rows = 0, cols = 0;

    tl::from_string (tl::to_string (column_x_le->text ()), cx);
    tl::from_string (tl::to_string (column_y_le->text ()), cy);
    tl::from_string (tl::to_string (row_x_le->text ()), rx);
    tl::from_string (tl::to_string (row_y_le->text ()), ry);
    tl::from_string (tl::to_string (rows_le->text ()), rows);
    tl::from_string (tl::to_string (columns_le->text ()), cols);

    db::Point rv (coord_from_dpoint (db::DPoint (rx, ry), dbu, du, t));
    db::Point cv (coord_from_dpoint (db::DPoint (cx, cy), dbu, du, t));

    if (is_complex) {
      inst = db::CellInstArray (db::CellInst (inst_cell_index), 
                                db::CplxTrans (mag, angle_deg, mirror, disp), rv, cv, rows, cols);
    } else {
      inst = db::CellInstArray (db::CellInst (inst_cell_index), 
                                db::Trans (angle, mirror, disp), rv, cv, rows, cols);
    }

  } else {
    if (is_complex) {
      inst = db::CellInstArray (db::CellInst (inst_cell_index), db::CplxTrans (mag, angle_deg, mirror, disp));
    } else {
      inst = db::CellInstArray (db::CellInst (inst_cell_index), db::Trans (angle, mirror, disp));
    }
  }

  if (! (pos->back ().inst_ptr.cell_inst () == inst)) {

    db::Instance new_inst = cell.replace (pos->back ().inst_ptr, inst);
    if (new_inst != pos->back ().inst_ptr) {

      //  change selection to new shape
      lay::ObjectInstPath new_sel (*pos);
      new_sel.back ().inst_ptr = new_inst;

      mp_service->select (*pos, lay::Editable::Reset);
      mp_service->select (new_sel, lay::Editable::Add);

      // locate the new selection
      for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
        ;
      tl_assert (pos != mp_service->selection ().end ());
      m_selection_ptrs [m_index] = pos;

      mp_service->selection_to_view ();

    }

    mp_service->view ()->add_new_layers (layer_state);

    //  remove superfluous proxies
    cv->layout ().cleanup ();

    update ();

  }
}

void
InstPropertiesPage::update_pcell_parameters ()
{
  db::Layout *layout;

  //  find the layout the cell has to be looked up: that is either the layout of the current instance or 
  //  the library selected
  if (lib_cbx->current_library ()) {

    layout = &lib_cbx->current_library ()->layout ();

  } else {

    edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
    const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
    layout = &cv->layout ();

  }

  std::pair<bool, db::pcell_id_type> pc = layout->pcell_by_name (tl::to_string (cell_name_le->text ()).c_str ());
  std::pair<bool, db::cell_index_type> cc = layout->cell_by_name (tl::to_string (cell_name_le->text ()).c_str ());

  //  by the way, update the foreground color of the cell edit box as well (red, if not valid)
  QPalette pl = cell_name_le->palette ();
  if (! pc.first && ! cc.first) {
    pl.setColor (QPalette::Text, Qt::red);
    pl.setColor (QPalette::Base, QColor (Qt::red).lighter (180));
  } else {
    pl.setColor (QPalette::Text, palette ().color (QPalette::Text));
    pl.setColor (QPalette::Base, palette ().color (QPalette::Base));
  }
  cell_name_le->setPalette (pl);

  if (pc.first && layout->pcell_declaration (pc.second)) {

    const db::PCellDeclaration *pcell_decl = layout->pcell_declaration (pc.second);

    std::vector<tl::Variant> parameters;

    edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
    const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
    db::Cell &cell = cv->layout ().cell (pos->cell_index ());
    std::pair<bool, db::pcell_id_type> pci = cell.is_pcell_instance (pos->back ().inst_ptr);
    const db::Library *pci_lib = cv->layout ().defining_library (pos->back ().inst_ptr.cell_index ()).first;

    //  fetch the parameters of the current instance if it matches the selected PCell and in that
    //  case use the current parameters
    if (pci.first) {
      if (pci.second == pc.second && pci_lib == lib_cbx->current_library ()) {
        //  exact match: take the parameters
        parameters = cell.get_pcell_parameters (pos->back ().inst_ptr);
      } else {
        //  otherwise: update the parameter whose name matches, use default for others
        parameters = pcell_decl->map_parameters (cell.get_named_pcell_parameters (pos->back ().inst_ptr));
      }
    }

    if (mp_pcell_parameters && mp_pcell_parameters->pcell_decl () == pcell_decl) {

      //  For identical parameter declarations just set the new values
      //  TODO: formally the declaration could change - compare current declarations vs. new ones?
      //  Better: provide a way to install custom parameter pages ...
      mp_pcell_parameters->set_parameters (parameters);

    } else {

      //  Hint: we shall not delete the page immediately. This gives a segmentation fault in some cases.
      if (mp_pcell_parameters) {
        mp_pcell_parameters->hide ();
        mp_pcell_parameters->deleteLater ();
      }

      mp_pcell_parameters = new PCellParametersPage (pcell_tab, &cv->layout (), mp_service->view (), pos->cv_index (), layout->pcell_declaration (pc.second), parameters);
      pcell_tab->layout ()->addWidget (mp_pcell_parameters);

    }

    param_tab_widget->setTabEnabled (1, true);
    param_tab_widget->setCurrentIndex (1);

  } else {

    //  Hint: we shall not delete the page immediately. This gives a segmentation fault in some cases.
    if (mp_pcell_parameters) {
      mp_pcell_parameters->hide ();
      mp_pcell_parameters->deleteLater ();
    }

    mp_pcell_parameters = 0;

    param_tab_widget->setCurrentIndex (0);
    param_tab_widget->setTabEnabled (1, false);
  }

}

}

