
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtPropertiesPages.h"
#include "edtDialogs.h"
#include "layDialogs.h"
#include "layObjectInstPath.h"
#include "layLayoutView.h"
#include "layExceptions.h"
#include "tlString.h"

#include <string>

namespace edt
{

// -------------------------------------------------------------------------
//  ShapePropertiesPage implementation

ShapePropertiesPage::ShapePropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), 
    mp_service (service), m_enable_cb_callback (true)
{
  m_selection_ptrs.reserve (service->selection ().size ());
  for (edt::Service::obj_iterator s = service->selection ().begin (); s != service->selection ().end (); ++s) {
    m_selection_ptrs.push_back (s);
  }
  m_index = 0;
  m_prop_id = 0;
  mp_service->clear_highlights ();
}

ShapePropertiesPage::~ShapePropertiesPage ()
{
  mp_service->restore_highlights ();
}

void 
ShapePropertiesPage::setup ()
{
  connect (dbu_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));

  m_enable_cb_callback = false;
  dbu_checkbox ()->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_checkbox ()->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;
}

void 
ShapePropertiesPage::back ()
{
  m_index = m_selection_ptrs.size ();
}

void 
ShapePropertiesPage::front ()
{
  m_index = 0;
}

bool 
ShapePropertiesPage::at_begin () const
{
  return (m_index == 0);
}

bool 
ShapePropertiesPage::at_end () const
{
  return (m_index == m_selection_ptrs.size ());
}

void 
ShapePropertiesPage::operator-- ()
{
  --m_index;
}

void 
ShapePropertiesPage::operator++ ()
{
  ++m_index;
}

void 
ShapePropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

bool 
ShapePropertiesPage::dbu_units () const
{
  return dbu_checkbox ()->isChecked ();
}

bool 
ShapePropertiesPage::abs_trans () const
{
  return abs_checkbox ()->isChecked ();
}

db::CplxTrans
ShapePropertiesPage::trans () const
{
  if (abs_trans ()) {
    return m_selection_ptrs[m_index]->trans ();
  } else {
    return db::CplxTrans ();
  }
}

void 
ShapePropertiesPage::display_mode_changed (bool)
{
BEGIN_PROTECTED

  if (! m_enable_cb_callback) {
    return;
  }

  update_shape ();

END_PROTECTED
}

void 
ShapePropertiesPage::update ()
{
  mp_service->highlight (m_index);

  update_shape ();
}

void 
ShapePropertiesPage::apply ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());

  tl_assert (! pos->is_cell_inst ());

  db::Shapes &shapes = cv->layout ().cell (pos->cell_index ()).shapes (pos->layer ());
  double dbu = cv->layout ().dbu ();

  if (pos->shape ().is_array_member ()) {
    throw tl::Exception (tl::translate ("Shape array members cannot be changed"));
  }

  db::Shape new_shape = do_apply (shapes, pos->shape (), dbu);

  if (new_shape != pos->shape ()) {

    //  change selection to new shape
    lay::ObjectInstPath new_sel (*pos);
    new_sel.set_shape (new_shape);

    mp_service->select (*pos, lay::Editable::Reset);
    mp_service->select (new_sel, lay::Editable::Add);

    // locate the new selection
    for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
      ;
    tl_assert (pos != mp_service->selection ().end ());
    m_selection_ptrs [m_index] = pos;

    mp_service->selection_to_view ();

  } 

  mp_service->handle_guiding_shape_changes ();

  update ();
}

void 
ShapePropertiesPage::update_shape ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  double dbu = cv->layout ().dbu ();

  tl_assert (! pos->is_cell_inst ());

  std::string layer (tl::translate ("Layer "));

  std::string ln = cv->layout ().get_properties (pos->layer ()).to_string ();
  for (lay::LayerPropertiesConstIterator lp = mp_service->view ()->begin_layers (); ! lp.at_end (); ++lp) {
    if (lp->cellview_index () == int (pos->cv_index ()) && lp->layer_index () == int (pos->layer ())) {
      ln = lp->display_string (mp_service->view (), true, true);
      break;
    }
  }
  layer += ln;

  layer += ", ";
  layer += tl::translate ("Cell ");
  layer += cv->layout ().cell_name (pos->cell_index ());

  mp_service->view ()->set_current_layer (pos->cv_index (), cv->layout ().get_properties (pos->layer ()));

  m_prop_id = pos->shape ().prop_id ();

  do_update (pos->shape (), dbu, layer);
}

void
ShapePropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), *m_selection_ptrs [m_index]);
}

void
ShapePropertiesPage::show_props ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  lay::UserPropertiesForm props_form (this);
  if (props_form.show (mp_service->view (), pos->cv_index (), m_prop_id)) {
    if (! readonly () && m_prop_id != pos->shape ().prop_id ()) {

      const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
      db::Shapes &shapes = cv->layout ().cell (pos->cell_index ()).shapes (pos->layer ());

      db::Transaction transaction (cv->layout ().manager (), tl::translate ("Change shape properties"));

      db::Shape new_shape = shapes.replace_prop_id (pos->shape (), m_prop_id);

      if (new_shape != pos->shape ()) {

        //  change selection to new shape
        lay::ObjectInstPath new_sel (*pos);
        new_sel.set_shape (new_shape);

        mp_service->select (*pos, lay::Editable::Reset);
        mp_service->select (new_sel, lay::Editable::Add);

        // locate the new selection
        for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
          ;
        tl_assert (pos != mp_service->selection ().end ());
        m_selection_ptrs [m_index] = pos;

        mp_service->selection_to_view ();

      } 

      update ();

    }
  }
}

bool 
ShapePropertiesPage::readonly ()
{
  return ! mp_service->view ()->is_editable ();
}

// -------------------------------------------------------------------------
//  helper functions to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

static db::Coord
coord_from_dcoord (double d, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DCoord dc = t.ctrans (d / dbu);
  db::Coord c = db::Coord (floor (dc + 0.5));

#if ! defined(HAVE_64BIT_COORD)   // for 64bit there is not accuracy reserve
  if (fabs (dc - c) > 1e-6) { 
    if (! du) {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit (%g)"), d, dbu);
    } else {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit"), d);
    }
  }
#endif

  return c;
}

static db::Point
coord_from_dpoint (const db::DPoint &dp, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DPoint dpt = t * (dp * (1.0 / dbu));

  db::Coord cx = coord_from_dcoord (dpt.x (), 1.0, false, db::CplxTrans ());
  db::Coord cy = coord_from_dcoord (dpt.y (), 1.0, false, db::CplxTrans ());

  return db::Point (cx, cy);
}

static db::Coord
coord_from_string (const char *txt, double dbu, bool du, const db::CplxTrans &t)
{
  double d = 0.0;
  tl::from_string (txt, d);
  return coord_from_dcoord (d, dbu, du, t);
}

static std::string
coords_to_string (const db::DPoint &dp, double dbu, bool du, const char *sep = "\t")
{
  return coord_to_string (dp.x (), dbu, du) + sep + coord_to_string (dp.y (), dbu, du);
}

// -------------------------------------------------------------------------
//  PolygonPropertiesPage implementation

PolygonPropertiesPage::PolygonPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PolygonPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (tl::to_qstring (lname));

  db::Polygon poly;
  shape.polygon (poly);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Polygon::polygon_contour_iterator pt = poly.begin_hull (); pt != poly.end_hull (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }

  for (unsigned int h = 0; h < poly.holes (); ++h) {

    ptlist += "\n/";

    for (db::Polygon::polygon_contour_iterator pt = poly.begin_hole (h); pt != poly.end_hole (h); ++pt) {
      ptlist += "\n";
      ptlist += coords_to_string (t * *pt, dbu, du);
    }

  }

  pointListEdit->setText (tl::to_qstring (ptlist));
  pointCountLabel->setText (tl::to_qstring (tl::sprintf (tl::translate ("(%lu points)"), poly.vertices ())));
}

db::Shape
PolygonPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  std::string text (tl::to_string (pointListEdit->toPlainText ()));
  tl::Extractor ex (text.c_str ());

  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  db::Polygon poly;

  if (*ex.skip () == '(') {

    db::DPolygon dp;
    ex.read (dp);

    poly = db::Polygon::from_double (dp.transformed (db::DCplxTrans (t) * db::DCplxTrans (du ? 1.0 : 1.0 / dbu)));

  } else {

    unsigned int h = 0;
    while (! ex.at_end ()) {

      std::vector <db::Point> points;

      while (! ex.at_end () && ! ex.test ("/")) {

        double dx = 0.0, dy = 0.0;
        ex.read (dx);
        ex.test (",");
        ex.read (dy);
        ex.test (";");

        points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

      }

      if (points.size () < 3) {
        throw tl::Exception (tl::translate ("Polygon must have at least three points"));
      }

      if (h == 0) {
        poly.assign_hull (points.begin (), points.end (), false /*not compressed*/);
      } else {
        poly.insert_hole (points.begin (), points.end (), false /*not compressed*/);
      }

      ++h;

    }

  }

  db::Polygon org_poly;
  shape.polygon (org_poly);
  if (poly != org_poly) {
    //  shape changed - replace the old by the new one
    return shapes.replace (shape, poly);
  } else {
    //  shape did not change
    return shape;
  }
}

// -------------------------------------------------------------------------
//  BoxPropertiesPage implementation

BoxPropertiesPage::BoxPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent),
    m_recursion_sentinel (false), m_tab_index (0), m_dbu (1.0)
{
  setupUi (this);
  setup ();

  connect (mode_tab, SIGNAL (currentChanged (int)), this, SLOT (changed ()));
  connect (x1_le_1, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (y1_le_1, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (x2_le_1, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (y2_le_1, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (w_le_2, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (h_le_2, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (cx_le_2, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (cy_le_2, SIGNAL (editingFinished ()), this, SLOT (changed ()));
  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
BoxPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  m_dbu = dbu;

  layer_lbl->setText (tl::to_qstring (lname));

  db::Box box;
  shape.box (box);
  set_box (box);
}

db::Shape
BoxPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  m_dbu = dbu;

  db::Box box = get_box (mode_tab->currentIndex ());

  db::Box org_box;
  shape.box (org_box);
  if (box != org_box) {
    //  shape changed - replace the old by the new one
    return shapes.replace (shape, box);
  } else {
    //  shape did not change
    return shape;
  }
}

db::Box 
BoxPropertiesPage::get_box (int mode) const
{
  if (mode == 0) {

    double x1 = 0.0, y1 = 0.0, x2 = 0.0, y2 = 0.0;
    tl::from_string (tl::to_string (x1_le_1->text ()), x1);
    tl::from_string (tl::to_string (y1_le_1->text ()), y1);
    tl::from_string (tl::to_string (x2_le_1->text ()), x2);
    tl::from_string (tl::to_string (y2_le_1->text ()), y2);

    db::CplxTrans t = trans ().inverted ();
    bool du = dbu_units ();

    return db::Box (coord_from_dpoint (db::DPoint (x1, y1), m_dbu, du, t), coord_from_dpoint (db::DPoint (x2, y2), m_dbu, du, t));

  } else {

    double cx = 0.0, cy = 0.0, w = 0.0, h = 0.0;
    tl::from_string (tl::to_string (cx_le_2->text ()), cx);
    tl::from_string (tl::to_string (cy_le_2->text ()), cy);
    tl::from_string (tl::to_string (w_le_2->text ()), w);
    tl::from_string (tl::to_string (h_le_2->text ()), h);

    db::CplxTrans t = trans ().inverted ();
    bool du = dbu_units ();

    return db::Box (coord_from_dpoint (db::DPoint (cx - w * 0.5, cy - h * 0.5), m_dbu, du, t), coord_from_dpoint (db::DPoint (cx + w * 0.5, cy + h * 0.5), m_dbu, du, t));

  }
}

void
BoxPropertiesPage::set_box (const db::Box &box)
{
  if (m_recursion_sentinel) {
    return;
  }
  m_recursion_sentinel = true;

  m_tab_index = mode_tab->currentIndex ();

  db::DCplxTrans t = db::DCplxTrans (trans ());
  db::DBox dbox (box);
  bool du = dbu_units ();

  x1_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.lower_left ())).x (), m_dbu, du)));
  y1_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.lower_left ())).y (), m_dbu, du)));
  x2_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.upper_right ())).x (), m_dbu, du)));
  y2_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.upper_right ())).y (), m_dbu, du)));
  cx_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.center ())).x (), m_dbu, du)));
  cy_le_1->setText (tl::to_qstring (coord_to_string ((t (dbox.center ())).y (), m_dbu, du)));
  w_le_1->setText (tl::to_qstring (coord_to_string (t.ctrans (dbox.width ()), m_dbu, du)));
  h_le_1->setText (tl::to_qstring (coord_to_string (t.ctrans (dbox.height ()), m_dbu, du)));

  x1_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.lower_left ())).x (), m_dbu, du)));
  y1_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.lower_left ())).y (), m_dbu, du)));
  x2_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.upper_right ())).x (), m_dbu, du)));
  y2_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.upper_right ())).y (), m_dbu, du)));
  cx_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.center ())).x (), m_dbu, du)));
  cy_le_2->setText (tl::to_qstring (coord_to_string ((t (dbox.center ())).y (), m_dbu, du)));
  w_le_2->setText (tl::to_qstring (coord_to_string (t.ctrans (dbox.width ()), m_dbu, du)));
  h_le_2->setText (tl::to_qstring (coord_to_string (t.ctrans (dbox.height ()), m_dbu, du)));

  m_recursion_sentinel = false;
}

void 
BoxPropertiesPage::changed ()
{
  try {
    set_box (get_box (m_tab_index));
  } catch (...) {
  }
}

// -------------------------------------------------------------------------
//  TextPropertiesPage implementation

TextPropertiesPage::TextPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
TextPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (tl::to_qstring (lname));

  db::Text text;
  shape.text (text);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();
  db::DPoint dp = t * text.trans ().disp ();

  text_le->setText (tl::to_qstring (tl::escape_string (text.string ())));
  x_le->setText (tl::to_qstring (coord_to_string (dp.x (), dbu, du)));
  y_le->setText (tl::to_qstring (coord_to_string (dp.y (), dbu, du)));
  if (text.size () != 0) {
    size_le->setText (tl::to_qstring (coord_to_string (t.ctrans (text.size ()), dbu, du)));
  } else {
    size_le->setText (QString ());
  }
  orient_cbx->setCurrentIndex (text.trans ().rot ());
  halign_cbx->setCurrentIndex (int (text.halign ()) + 1);
  valign_cbx->setCurrentIndex (int (text.valign ()) + 1);
}

db::Shape
TextPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  double x = 0.0, y = 0.0;
  tl::from_string (tl::to_string (x_le->text ()), x);
  tl::from_string (tl::to_string (y_le->text ()), y);

  db::Point tp = coord_from_dpoint (db::DPoint (x, y), dbu, du, t);
  db::Trans tt (orient_cbx->currentIndex (), tp);

  std::string str = tl::unescape_string (tl::to_string (text_le->text ()));

  db::Text text;
  if (! size_le->text ().isEmpty ()) {
    db::Coord size = coord_from_string (tl::to_string (size_le->text ()).c_str (), dbu, du, t);
    text = db::Text (str, tt, size);
  } else {
    text = db::Text (str, tt);
  }

  text.halign (db::HAlign (halign_cbx->currentIndex () - 1));
  text.valign (db::VAlign (valign_cbx->currentIndex () - 1));

  db::Text org_text;
  shape.text (org_text);
  if (text != org_text) {
    //  shape changed - replace the old by the new one
    return shapes.replace (shape, text);
  } else {
    //  shape did not change
    return shape;
  }
}

// -------------------------------------------------------------------------
//  PathPropertiesPage implementation

PathPropertiesPage::PathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PathPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (tl::to_qstring (lname));

  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }
  ptlist_le->setText (tl::to_qstring (ptlist));

  width_le->setText (tl::to_qstring (coord_to_string (t.ctrans (path.width ()), dbu, du)));
  start_ext_le->setText (tl::to_qstring (coord_to_string (t.ctrans (path.extensions ().first), dbu, du)));
  end_ext_le->setText (tl::to_qstring (coord_to_string (t.ctrans (path.extensions ().second), dbu, du)));
  round_cb->setChecked (path.round ());
}

db::Shape
PathPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  std::string text (tl::to_string (ptlist_le->toPlainText ()));
  tl::Extractor ex (text.c_str ());

  std::vector <db::Point> points;

  while (! ex.at_end ()) {

    double dx = 0.0, dy = 0.0;
    ex.read (dx);
    ex.read (dy);

    points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

  }

  if (points.size () < 1) {
    throw tl::Exception (tl::translate ("Path must have at least one point"));
  }

  db::Coord w = coord_from_string (tl::to_string (width_le->text ()).c_str (), dbu, du, t);
  db::Coord se = coord_from_string (tl::to_string (start_ext_le->text ()).c_str (), dbu, du, t);
  db::Coord ee = coord_from_string (tl::to_string (end_ext_le->text ()).c_str (), dbu, du, t);

  db::Path path (points.begin (), points.end (), w, se, ee, round_cb->isChecked ());

  db::Path org_path;
  shape.path (org_path);
  if (path != org_path) {
    //  shape changed - replace the old by the new one
    return shapes.replace (shape, path);
  } else {
    //  shape did not change
    return shape;
  }
}

// -------------------------------------------------------------------------
//  EditablePathPropertiesPage implementation

EditablePathPropertiesPage::EditablePathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
  connect (type_cb, SIGNAL (currentIndexChanged (int)), this, SLOT (type_selected (int)));
}

void 
EditablePathPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (tl::to_qstring (lname));

  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }
  ptlist_le->setText (tl::to_qstring (ptlist));

  db::Coord w = path.width ();
  db::Coord se = path.extensions ().first;
  db::Coord ee = path.extensions ().second;

  width_le->setText (tl::to_qstring (coord_to_string (t.ctrans (w), dbu, du)));

  start_ext_le->setText (tl::to_qstring (coord_to_string (t.ctrans (se), dbu, du)));
  end_ext_le->setText (tl::to_qstring (coord_to_string (t.ctrans (ee), dbu, du)));

  start_ext_le->setEnabled (false);
  end_ext_le->setEnabled (false);

  if (se == 0 && ee == 0) {
    type_cb->setCurrentIndex (0); // flush
  } else if (se == w / 2 && ee == w / 2 && path.round ()) {
    type_cb->setCurrentIndex (3); // round
  } else if (se == w / 2 && ee == w / 2) {
    type_cb->setCurrentIndex (1); // square
  } else {
    type_cb->setCurrentIndex (2); // variable
    start_ext_le->setEnabled (true);
    end_ext_le->setEnabled (true);
  }
}

db::Shape
EditablePathPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  std::string text (tl::to_string (ptlist_le->toPlainText ()));
  tl::Extractor ex (text.c_str ());

  std::vector <db::Point> points;

  while (! ex.at_end ()) {

    double dx = 0.0, dy = 0.0;
    ex.read (dx);
    ex.read (dy);

    points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

  }

  if (points.size () < 1) {
    throw tl::Exception (tl::translate ("Path must have at least one point"));
  }

  db::Coord w = coord_from_string (tl::to_string (width_le->text ()).c_str (), dbu, du, t);

  db::Coord se = 0, ee = 0;
  bool round = false;

  switch (type_cb->currentIndex ()) {
  case 0: // flush
    break;
  case 1: // square
    se = ee = w / 2;
    break;
  case 2: // variable
    se = coord_from_string (tl::to_string (start_ext_le->text ()).c_str (), dbu, du, t);
    ee = coord_from_string (tl::to_string (end_ext_le->text ()).c_str (), dbu, du, t);
    break;
  case 3: // round
    se = ee = w / 2;
    round = true;
    break;
  } 

  db::Path path (points.begin (), points.end (), w, se, ee, round);

  db::Path org_path;
  shape.path (org_path);
  if (path != org_path) {
    //  shape changed - replace the old by the new one
    return shapes.replace (shape, path);
  } else {
    //  shape did not change
    return shape;
  }
}

void
EditablePathPropertiesPage::type_selected (int t)
{
  start_ext_le->setEnabled (t == 2);
  end_ext_le->setEnabled (t == 2);
}

}

