
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiClass
#define _HDR_gsiClass

#include "tlTypeTraits.h"

//  For a comprehensive documentation see gsi.h

namespace gsi
{

// -----------------------------------------------------------------------------
//  Variant binding

/**
 *  @brief A helper function to implement equal as efficiently as possible
 */
template<class T, class I>
bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, I);

template<class T>
bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::true_tag)
{
  return *a == *b;
}

template<class T>
bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::false_tag)
{
  return delegate->equal_impl ((void *) a, (void *) b);
}

/**
 *  @brief A helper function to implement less as efficiently as possible
 */
template<class T, class I>
bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, I);

template<class T>
bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::true_tag)
{
  return *a < *b;
}

template<class T>
bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::false_tag)
{
  return delegate->less_impl ((void *) a, (void *) b);
}

/**
 *  @brief A helper function to implement to_string as efficiently as possible
 */
template<class T, class I>
std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl *delegate, I);

template<class T>
std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl *delegate, tl::true_tag)
{
  return a->to_string ();
}

template<class T>
std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl *delegate, tl::false_tag)
{
  return delegate->to_string_impl ((void *) a);
}

/**
 *  @brief A helper function to implement read as efficiently as possible
 */
template<class T, class I>
void _var_user_read_impl (T *a, tl::Extractor &ex, I);

template<class T>
void _var_user_read_impl (T *a, tl::Extractor &ex, tl::true_tag)
{
  ex.read (*a);
}

template<class T>
void _var_user_read_impl (T *a, tl::Extractor &ex, tl::false_tag)
{
  tl_assert (false);
}

/**
 *  @brief A VariantUserClassBase specialization that links GSI classes and Variant classes
 */
template <class T>
class VariantUserClass
  : public tl::VariantUserClass<T>, private VariantUserClassImpl
{
public:
  VariantUserClass ()
    : VariantUserClassImpl (),
      mp_cls (0), mp_object_cls (0), m_reference (false), m_is_const (false)
  {
    //  .. nothing yet ..
  }

  ~VariantUserClass ()
  {
    mp_cls = 0;
    tl::VariantUserClass<T>::unregister_instance (this, m_reference, m_is_const);
  }

  void initialize (const gsi::ClassBase *cls, const tl::VariantUserClassBase *object_cls, bool reference, bool is_const)
  {
    mp_cls = cls; 
    mp_object_cls = object_cls;
    m_reference = reference;
    m_is_const = is_const;

    VariantUserClassImpl::initialize (cls, this, object_cls, reference, is_const);

    tl::VariantUserClass<T>::register_instance (this, m_reference, m_is_const);
  }

  const tl::EvalClass *eval_cls () const 
  {
    return this;
  }

  virtual bool equal (void *a, void *b) const
  {
    typename tl::type_traits<T>::has_equal_operator f;
    return gsi::_var_user_equal_impl ((const T *) a, (const T *) b, this, f);
  }

  virtual bool less (void *a, void *b) const
  {
    typename tl::type_traits<T>::has_less_operator f;
    return gsi::_var_user_less_impl ((const T *) a, (const T *) b, this, f);
  }

  virtual std::string to_string (void *a) const
  {
    typename tl::type_traits<T>::supports_to_string f;
    return gsi::_var_user_to_string_impl ((const T *) a, this, f);
  }

  void *clone (void *obj) const 
  {
    if (m_reference) {
      return obj;
    } else {
      void *new_obj = mp_cls->create ();
      mp_cls->assign (new_obj, obj);
      return new_obj;
    }
  }

  void *create () const 
  {
    tl_assert (!m_reference);
    return mp_cls->create ();
  }

  void destroy (void *obj) const 
  {
    if (obj && !m_reference) {
      mp_cls->destroy (obj);
    }
  }

  const char *name () const
  {
    return mp_cls ? mp_cls->name ().c_str() : 0;
  }

  void read (void *a, tl::Extractor &ex) const
  {
    typename tl::type_traits<T>::supports_extractor f;
    gsi::_var_user_read_impl ((T *) a, ex, f);
  }

  const gsi::ClassBase *gsi_cls () const
  {
    return mp_cls;
  }

  bool is_class () const 
  { 
    return mp_object_cls != 0; 
  }

  bool is_ref () const 
  { 
    return m_reference; 
  }

  bool is_const () const 
  { 
    return m_is_const; 
  }

private:
  const gsi::ClassBase *mp_cls;
  const tl::VariantUserClassBase *mp_object_cls;
  bool m_reference, m_is_const;

  //  no copying currently
  VariantUserClass &operator= (const VariantUserClass &);
  VariantUserClass (const VariantUserClass &);
};

// -----------------------------------------------------------------------------
//  GSI implementation

template <class X> 
void *_take_and_create (SerialArgs &args, tl::false_tag /*has_copy_ctor*/)
{
  tl_assert (false); // cannot copy object of this type
  return 0;
}

template <class X> 
void *_take_and_create (SerialArgs &from, tl::true_tag /*has_copy_ctor*/)
{
  return from.template take_object<X> ();
}

template <class X> 
void *_get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj), tl::false_tag /*has_copy_ctor*/) 
{
  tl_assert (false); // cannot copy object of this type
  return 0;
}

template <class X> 
void _get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj), tl::true_tag /*has_copy_ctor*/) 
{
  std::vector<X> vv;
  const std::vector<X> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<X> &> (typename gsi::type_traits<const std::vector<X> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<X> *> (typename gsi::type_traits<const std::vector<X> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<X> &> (typename gsi::type_traits<std::vector<X> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<X> *> (typename gsi::type_traits<std::vector<X> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<X> > (typename gsi::type_traits< std::vector<X> >::tag ());
  }
  for (typename std::vector<X>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, new X (*o));
  }
}

template <class X> 
void _get_cptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
{
  std::vector<const X *> vv;
  const std::vector<const X *> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<const X *> &> (typename gsi::type_traits<const std::vector<const X *> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<const X *> *> (typename gsi::type_traits<const std::vector<const X *> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<const X *> &> (typename gsi::type_traits<std::vector<const X *> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<const X *> *> (typename gsi::type_traits<std::vector<const X *> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<const X *> > (typename gsi::type_traits< std::vector<const X *> >::tag ());
  }
  for (typename std::vector<const X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, (void *) *o);
  }
}

template <class X> 
void _get_ptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
{
  std::vector<X *> vv;
  const std::vector<X *> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<X *> &> (typename gsi::type_traits<const std::vector<X *> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<X *> *> (typename gsi::type_traits<const std::vector<X *> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<X *> &> (typename gsi::type_traits<std::vector<X *> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<X *> *> (typename gsi::type_traits<std::vector<X *> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<X *> > (typename gsi::type_traits< std::vector<X *> >::tag ());
  }
  for (typename std::vector<X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, *o);
  }
}

template <class X> 
void _destroy (X *x, tl::false_tag /*has public dtor*/)
{
  tl_assert (false); // cannot delete object of this type
}

template <class X> 
void _destroy (X *x, tl::true_tag /*has public dtor*/)
{
  delete x;
}

template <class X> 
void _push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects, tl::false_tag /*has_copy_ctor*/) 
{
  tl_assert (false); // cannot copy object of this type
}

template <class X> 
void _push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects, tl::true_tag /*has_copy_ctor*/) 
{
  tl_assert (a.inner != 0);

  std::vector<X> vv;
  std::vector<X> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<X> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back (*(X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void _push_cptr_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects)
{
  tl_assert (a.inner != 0);

  std::vector<const X *> vv;
  std::vector<const X *> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<const X *> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back ((const X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void _push_ptr_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects)
{
  tl_assert (a.inner != 0);

  std::vector<X *> vv;
  std::vector<X *> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<X *> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back ((X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void *_create (tl::false_tag)
{
  throw tl::Exception (tl::translate ("Object cannot be created here"));
  return 0;
}

template <class X> 
void *_create (tl::true_tag)
{
  return new X ();
}

template <class X> 
void *_clone (tl::false_tag, const void *other)
{
  throw tl::Exception (tl::translate ("Object cannot be copied here"));
  return 0;
}

template <class X> 
void *_clone (tl::true_tag, const void *other)
{
  return new X (*(const X *)other);
}

template <class X> 
void _assign (tl::false_tag /*has_copy_ctor*/, void *, const void *)
{
  throw tl::Exception (tl::translate ("Object cannot be copied here"));
}

template <class X> 
void _assign (tl::true_tag /*has_copy_ctor*/, void *dest, const void *src)
{
  *(X *)dest = *(const X *)src;
}

template <class X> 
void _copy_to (SerialArgs &at, void *from, tl::false_tag /*has_copy_ctor*/) 
{
  throw tl::Exception (tl::translate ("Object cannot be copied here"));
}

template <class X> 
void _copy_to (SerialArgs &at, void *from, tl::true_tag /*has_copy_ctor*/) 
{
  at.set_value (typename type_traits<X>::tag (), *((const X *)from));
}

inline void *get_client_data (void *) 
{
  return 0;
}

inline void *get_client_data (gsi::ObjectBase *x) 
{
  return x->data ();
}

/**
 *  @brief The declaration of a specific class
 *
 *  This class declares all methods that are required to instantiate or copy an object
 *  or to call it's methods in some generic way.
 *  The ClassExt declaration shall be used by extensions. For the basic declaration use
 *  Class which provides the singleton declaration object for that type.
 */
template <class X>
class ClassExt
  : public ClassBase
{
public:
  typedef typename tl::type_traits<X>::has_copy_constructor has_copy_ctor;
  typedef typename tl::type_traits<X>::has_default_constructor has_default_ctor;
  typedef typename tl::type_traits<X>::has_public_destructor has_public_dtor;

  ClassExt (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (0, name, 0, doc, mm)
  {
    // ...
  }

  ClassExt (const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (0, name, &base, doc, mm)
  {
    // ...
  }

  ClassExt (const std::string &name, const std::string &doc = std::string ())
    : ClassBase (0, name, 0, doc, Methods ())
  {
    // ...
  }

  ClassExt (const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (0, name, &base, doc, Methods ())
  {
    // ...
  }

  ClassExt (const ObjectConnector &connector, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (&connector, name, 0, doc, mm)
  {
    // ...
  }

  ClassExt (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (&connector, name, &base, doc, mm)
  {
    // ...
  }

  ClassExt (const ObjectConnector &connector, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (&connector, name, 0, doc, Methods ())
  {
    // ...
  }

  ClassExt (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (&connector, name, &base, doc, Methods ())
  {
    // ...
  }

  void initialize ()
  {
    ClassBase::initialize ();
    m_var_cls.initialize (this, 0, false, false);
    m_var_cls_r.initialize (this, 0, true, false);
    m_var_cls_c.initialize (this, 0, false, true);
    m_var_cls_rc.initialize (this, 0, true, true);
    m_var_cls_cls.initialize (this, &m_var_cls, true, false);
  }

  void *client_data (void *p) const
  {
    void *cd = basic_client_data (p);
    if (cd) {
      return cd;
    } else {
      X *x = (X *)p;
      return get_client_data (x);
    }
  }

  virtual void destroy (void *p) const
  {
    X *x = (X *)p;
    has_public_dtor hpd;
    _destroy (x, hpd);
  }

  virtual void *create () const
  {
    has_default_ctor cst;
    void *r = _create<X> (cst);
    return r;
  }

  virtual void *clone (const void *other) const
  {
    has_copy_ctor cst;
    void *r = _clone<X> (cst, other);
    return r;
  }

  virtual void assign (void *dest, const void *src) const
  {
    has_copy_ctor cst;
    _assign<X> (cst, dest, src);
  }

  virtual void get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj)) const 
  {
    has_copy_ctor cpt;
    tl_assert (a.inner != 0);
    if (a.inner->is_cptr) {
      _get_cptr_vector_of<X> (from, a, data, cb);
    } else if (a.inner->is_ptr) {
      _get_ptr_vector_of<X> (from, a, data, cb);
    } else {
      _get_vector_of<X> (from, a, data, cb, cpt);
    }
  }

  void push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects) const 
  {
    has_copy_ctor cpt;
    tl_assert (a.inner != 0);
    if (a.inner->is_cptr) {
      _push_cptr_vector_of<X> (to, a, temp_objects, objects);
    } else if (a.inner->is_ptr) {
      _push_ptr_vector_of<X> (to, a, temp_objects, objects);
    } else {
      _push_vector_of<X> (to, a, temp_objects, objects, cpt);
    }
  }

  virtual void *take_and_create (SerialArgs &from) const
  {
    has_copy_ctor cpt;
    return _take_and_create<X> (from, cpt);
  }

  virtual void copy_to (SerialArgs &at, void *from) const 
  {
    has_copy_ctor cpt;
    _copy_to<X> (at, from, cpt);
  }

  virtual void *create_obj_iterator (SerialArgs &from) const
  {
    return from.template get_value<gsi::IterAdaptorAbstractBase *> (typename type_traits <gsi::IterAdaptorAbstractBase *>::tag ());
  }

  virtual bool obj_iterator_at_end (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->at_end ();
  }

  virtual void *take_iterator_obj (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->v_get_copy ();
  }

  virtual const void *get_iterator_obj (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->v_get ();
  }

  virtual void inc_obj_iterator (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    i->inc ();
  }

  virtual void delete_obj_iterator (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    delete i;
  }

  virtual ClassBase *declaration () const
  {
    return 0;
  }

  virtual bool can_destroy () const
  {
    has_public_dtor hpd;
    return tl::value_of (hpd);
  }

  virtual bool can_copy () const
  {
    has_copy_ctor cpt;
    return tl::value_of (cpt);
  }

  virtual bool can_default_create () const
  {
    has_default_ctor cpt;
    return tl::value_of (cpt);
  }

  virtual const ClassBase *subclass_decl (const void *p) const 
  {
    if (p == 0) {
      return this;
    }

    try {

      const X *x = (const X *)p;
      const std::type_info &ti_base = typeid (*x);
      const ClassBase *cls = find_subclass_decl (ti_base);
      return cls ? cls : this;

    } catch (std::bad_typeid &) {
      return this;
    }
  }

  virtual const ClassBase *find_subclass_decl (const std::type_info &ti) const 
  {
    if (ti == typeid (X)) {
      return this;
    }

    for (std::vector <const ClassBase *>::const_iterator s = subclasses ().begin (); s != subclasses ().end (); ++s) {
      const ClassBase *cls = (*s)->find_subclass_decl (ti);
      if (cls) {
        return cls;
      }
    }

    return 0;
  }

  virtual bool is_of_type (const std::type_info &ti) const 
  {
    return (ti == typeid (X));
  }

  virtual const tl::VariantUserClassBase *var_cls_cls () const
  {
    return &m_var_cls_cls;
  }

  virtual const tl::VariantUserClassBase *var_cls (bool reference, bool is_const) const
  {
    if (reference && is_const) {
      return &m_var_cls_rc;
    } else if (reference) {
      return &m_var_cls_r;
    } else if (is_const) {
      return &m_var_cls_c;
    } else {
      return &m_var_cls;
    }
  }

private:
  gsi::VariantUserClass<X> m_var_cls;
  gsi::VariantUserClass<X> m_var_cls_r;
  gsi::VariantUserClass<X> m_var_cls_c;
  gsi::VariantUserClass<X> m_var_cls_rc;
  gsi::VariantUserClass<X> m_var_cls_cls;
};

/**
 *  @brief The declaration of a specific class
 *
 *  This is the final version of the declaration. In addition to ClassExt it provides
 *  a singleton declaration object for this type.
 */
template <class X>
class Class
  : public ClassExt<X>
{
public:
  Class (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassExt<X> (name, mm, doc)
  {
    // ...
  }

  Class (const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassExt<X> (base, name, mm, doc)
  {
    // ...
  }

  Class (const std::string &name, const std::string &doc = std::string ())
    : ClassExt<X> (name, doc)
  {
    // ...
  }

  Class (const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassExt<X> (base, name, doc)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassExt<X> (connector, name, mm, doc)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassExt<X> (connector, base, name, mm, doc)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const std::string &name, const std::string &doc = std::string ())
    : ClassExt<X> (connector, name, doc)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassExt<X> (connector, base, name, doc)
  {
    // ...
  }

  virtual ClassBase *declaration () const
  {
    return &decl;
  }

  static Class<X> decl;
};

/**
 *  @brief Obtain the class declaration for a given class
 * 
 *  This method looks up the declaration object for a given type.
 *  It does so dynamically, since declarations may be located in different
 *  libraries. However, for performance reasons, the definitions are cached. 
 */
template <class X>
const ClassBase *cls_decl ()
{
  //  TODO: needs thread safety? It's rather unlikely that two threads enter this
  //  piece of code at the same time and they interfere when storing the results.
  static const ClassBase *cd = 0;
  if (! cd) {
    for (ClassBase::class_iterator c = ClassBase::begin_classes (); c != ClassBase::end_classes () && ! cd; ++c) {
      if (c->is_of_type (typeid (X)) && c->declaration () == &*c) {
        cd = &*c;
      }
    }
    //  assert, if no GSI binding available for that class
    tl_assert (cd != 0); 
  }
  return cd;
}

}

#endif
