
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsiDecl.h"
#include "tlAssert.h"

#include <QtCore/QMetaMethod>
#include <QtCore/QMetaObject>

#include <cstdio>

namespace gsi
{

// -------------------------------------------------------------------------
//  ClassBase implementation

ClassBase::class_collection *ClassBase::mp_class_collection = 0;
unsigned int ClassBase::m_class_count = 0;

ClassBase::ClassBase (const ObjectConnector *connector, const std::string &name, const ClassBase *base, const std::string &doc, const Methods &mm)
  : m_name (name), mp_base (base), m_doc (doc), m_methods (mm), mp_connector (connector)
{ 
  if (! mp_class_collection) {
    mp_class_collection = new class_collection ();
  }
  mp_class_collection->insert (this, false);
  ++m_class_count;
}

ClassBase::~ClassBase ()
{
  tl_assert (m_class_count > 0);
  if (--m_class_count == 0) {
    delete mp_class_collection;
    mp_class_collection = 0;
  }
}

void  
ClassBase::on_attach (void *p) const
{
  if (mp_base) {
    mp_base->on_attach (p);
  }
  if (mp_connector) {
    mp_connector->on_attach (p);
  }
}

void  
ClassBase::on_detach (void *p) const
{
  if (mp_connector) {
    mp_connector->on_detach (p);
  }
  if (mp_base) {
    mp_base->on_detach (p);
  }
}

bool 
ClassBase::is_derived_from (const ClassBase *base) const
{
  if (base == this) {
    return true;
  } else if (mp_base == 0) {
    return false;
  } else {
    return mp_base->is_derived_from (base);
  }
}

const ClassBase::class_collection &
ClassBase::collection () 
{
  if (mp_class_collection) {
    return *mp_class_collection;
  } else {
    static class_collection empty;
    return empty;
  }
}

void
ClassBase::merge_declarations ()
{
  if (! mp_class_collection) {
    return;
  }

  //  merge the extensions to the main declaration
  static bool merged = false;
  if (! merged) {

    //  HINT: this code block must be called exactly ONCE!
    merged = true;

    //  Now we are sure that all Class objects are initialized properly (through static initialization)
    //  and we can collect all methods.
    for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
      //  consider only non-extensions: the first pass merges methods within the same class definition
      if (c->declaration () == &*c) {
        Methods mm;
        mm.swap (c->declaration ()->methods ());
        for (method_iterator m = mm.begin (); m != mm.end (); ++m) {
          c->declaration ()->add_method ((*m)->clone ());
        }
      }
    }

    std::vector <const gsi::ClassBase *> ext_declarations;

    for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
      //  consider only extensions
      if (c->declaration () != &*c) {
        ClassBase *declaration = c->declaration ();
        if (! declaration) {
          //  fallback: extensions may not link to the singleton declaration - look up the class by typeid
          declaration = const_cast<ClassBase *> (class_by_name (c->name ()));
        }
        for (gsi::ClassBase::method_iterator m = c->begin_methods (); m != c->end_methods (); ++m) {
          declaration->add_method ((*m)->clone ());
        }
        ext_declarations.push_back (&*c);
      }
    }

    //  removed auxiliary classes that have been merged with the "actual" one
    for (std::vector <const gsi::ClassBase *>::const_iterator ed = ext_declarations.begin (); ed != ext_declarations.end (); ++ed) {
      //  TODO: ugly const_cast hack
      mp_class_collection->remove (const_cast<gsi::ClassBase *> (*ed));
    }

  }

  //  collect the subclasses of a class
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    if (c->base ()) {
      //  TODO: ugly const_cast hack
      const_cast<gsi::ClassBase *> (c->base ())->m_subclasses.push_back (&*c);
    }
  }

  //  compute the method offsets (which is the number of methods provided by the base classes)
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    unsigned int mo = 0;
    const gsi::ClassBase *cb = c->base ();
    while (cb) {
      mo += (unsigned int) cb->overloaded_methods ();
      cb = cb->base ();
    }
    //  TODO: ugly const_cast hack
    (const_cast<gsi::ClassBase *> (&*c))->set_method_offset (mo);
  }
}

//  extract protected flag from the method name
static const char *
extract_protected (const char *name, bool &prot)
{
  if (*name == '*' && name[1] && name[1] != '=') {
    ++name;
    prot = true;
  } else {
    prot = false;
  }
  return name;
}

void
ClassBase::add_method (MethodBase *method, bool base_class)
{
  //  take apart the name by synonyms
  //  The name may be composed of several synonym names like "<name1>|<name2>|...".
  //  Each synonym name may start with a "#" indicating a deprecated version.
  bool prot = false;
  tl::Extractor ex (extract_protected (method->name ().c_str (), prot));
  while (! ex.at_end ()) {

    ex.test ("#");  //  deprecated flag
    std::string name;
    while (! ex.at_end () && *ex != '|') {
      if (*ex == '\\') {
        ++ex;
      }
      if (! ex.at_end ()) {
        name += *ex;
        ++ex;
      }
    }
    ex.test ("|");
  
    //  insert the method into the overloaded name map
    NameMap::iterator nm = m_name_map.find (std::make_pair (name, method->is_static ()));
    if (nm != m_name_map.end ()) {
      //  The base classes methods are completely overwritten by the derived class - since the derived class
      //  is entered first, we simply do not extend the overloaded base class method list for base class methods.
      if (! base_class) {
        m_overloaded_methods [nm->second].push_back ((unsigned int) m_methods.size ());
      }
    } else {
      m_name_map.insert (std::make_pair (std::make_pair (name, method->is_static ()), (unsigned int) m_overloaded_methods.size ()));
      m_overloaded_names.push_back (name);
      m_overloaded_methods.push_back (std::vector<unsigned int> ());
      m_overloaded_methods.back ().push_back ((unsigned int) m_methods.size ());
    }
    
  }

  m_methods.add_method (method);
}

const ClassBase *class_by_name (const std::string &name)
{
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    if (c->declaration () == &*c && c->name () == name) {
      return &*c;
    }
  }
  tl_assert (false);
}

bool has_class (const std::string &name)
{
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    if (c->declaration () == &*c && c->name () == name) {
      return true;
    }
  }
  return false;
}

}

