
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiClassBase
#define _HDR_gsiClassBase

#include "tlCollection.h"
#include "tlException.h"
#include "tlExpression.h"

#include "gsiExpression.h"

#include <list>
#include <map>
#include <vector>
#include <typeinfo>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

/**
 *  @brief A class that allows to implement specific behaviour when creating or deleting an object
 *
 *  This is delegate that implements the client-data / C++ object interface by delegation 
 *  rather then by inheritance like gsi::ObjectBase does.
 */
class ObjectConnector
{
public:
  /**
   *  @brief Destructor
   */
  virtual ~ObjectConnector () { }

  /**
   *  @brief If a handler is registered, this method is called when a client object is bound to the given C++ object
   *
   *  @param p The C++ object
   */
  virtual void on_attach (void *) const { }

  /**
   *  @brief If a handler is registered, this method is called when a client object is deleted and the given C++ object is released
   *
   *  @param p The C++ object
   */
  virtual void on_detach (void *) const { }

  /**
   *  @brief Return the client data for the object or 0 if there is no client data
   */
  virtual void *client_data (void *p) const { return 0; }
};

/**
 *  @brief The basic object to declare a class
 */
class KLAYOUT_DLL ClassBase
  : public tl::Collectable
{
public:
  typedef tl::Collection<ClassBase> class_collection; 
  typedef class_collection::const_iterator class_iterator;
  typedef Methods::iterator method_iterator;
  typedef std::map<std::pair<std::string, bool>, unsigned int> NameMap;
  typedef NameMap::const_iterator name_map_iterator;
  typedef std::vector< std::vector<unsigned int> > OverloadedMethodsList;
  typedef std::vector<unsigned int>::const_iterator overloaded_methods_iterator;
  
  ClassBase (const ObjectConnector *mp_connector, const std::string &name, const ClassBase *base, const std::string &doc, const Methods &mm);

  virtual ~ClassBase ();

  const ClassBase *base () const
  {
    return mp_base;
  }

  virtual ClassBase *declaration () const = 0;

  Methods &methods () 
  {
    return m_methods;
  }

  void add_method (MethodBase *method, bool base_class = false);

  const std::string &name () const
  {
    return m_name;
  }

  const std::string &doc () const
  {
    return m_doc;
  }

  /**
   *  @brief Merge base and extension declarations
   *
   *  This method must be called at least once!
   */
  static void merge_declarations ();

  static class_iterator begin_classes () 
  {
    return collection ().begin ();
  }
 
  static class_iterator end_classes () 
  {
    return collection ().end ();
  }
 
  method_iterator begin_methods () const
  {
    return m_methods.begin ();
  }

  method_iterator end_methods () const
  {
    return m_methods.end ();
  }

  name_map_iterator begin_name_map () const
  {
    return m_name_map.begin ();
  }

  name_map_iterator end_name_map () const
  {
    return m_name_map.end ();
  }

  unsigned int overloaded_methods () const
  {
    return (unsigned int) m_overloaded_methods.size ();
  }

  unsigned int method_offset () const
  {
    return m_method_offset;
  }

  void set_method_offset (unsigned int mo)
  {
    m_method_offset = mo;
  }

  overloaded_methods_iterator begin_overloaded_methods (unsigned int index) const
  {
    return m_overloaded_methods [index].begin ();
  }

  overloaded_methods_iterator end_overloaded_methods (unsigned int index) const
  {
    return m_overloaded_methods [index].end ();
  }

  const std::string &overloaded_name (unsigned int index) const
  {
    return m_overloaded_names [index];
  }

  bool is_derived_from (const ClassBase *base) const;

  virtual void attach (void *obj) const { on_attach (obj); }
  virtual void detach (void *obj) const { on_detach (obj); }
  virtual void destroy (void *obj) const = 0;
  virtual void *create () const = 0;
  virtual void *clone (const void *) const = 0;
  virtual void assign (void *, const void *) const = 0;
  virtual void *take_and_create (SerialArgs &from) const = 0;
  virtual void get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj)) const = 0;
  virtual void push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects) const = 0;
  virtual void copy_to (SerialArgs &at, void *from) const = 0;
  virtual void *create_obj_iterator (SerialArgs &from) const = 0;
  virtual bool obj_iterator_at_end (void *iter) const = 0;
  virtual void *take_iterator_obj (void *iter) const = 0;
  virtual const void *get_iterator_obj (void *iter) const = 0;
  virtual void inc_obj_iterator (void *iter) const = 0;
  virtual void delete_obj_iterator (void *iter) const = 0;
  virtual bool can_destroy () const = 0;
  virtual bool can_copy () const = 0;
  virtual bool can_default_create () const = 0;
  virtual bool is_of_type (const std::type_info &ti) const = 0;
  virtual const ClassBase *subclass_decl (const void *p) const = 0;
  virtual const ClassBase *find_subclass_decl (const std::type_info &ti) const = 0;
  virtual void *client_data (void *p) const = 0;
  virtual const tl::VariantUserClassBase *var_cls (bool reference, bool is_const) const = 0;
  virtual const tl::VariantUserClassBase *var_cls_cls () const = 0;

  virtual void initialize () 
  {
    m_methods.initialize ();
  }

protected:
  static const class_collection &collection ();

  const std::vector<const ClassBase *> &subclasses () const
  {
    return m_subclasses;
  }

  void on_attach (void *p) const;
  void on_detach (void *p) const;

  void *basic_client_data (void *p) const
  { 
    if (mp_connector) {
      return mp_connector->client_data (p);
    } else {
      return 0;
    }
  }

private:
  std::string m_name;
  const ClassBase *mp_base;
  std::string m_doc;
  Methods m_methods;
  NameMap m_name_map;
  OverloadedMethodsList m_overloaded_methods;
  unsigned int m_method_offset;
  std::vector<std::string> m_overloaded_names;
  std::vector<const ClassBase *> m_subclasses;
  const ObjectConnector *mp_connector;

  static class_collection *mp_class_collection;
  static unsigned int m_class_count;
};

/**
 *  @brief Accessor to a declaration through name
 * 
 *  The main intention of this method is to provide access to the 
 *  decl objects for external components. First, using this method
 *  is the only way to export the decl symbols to other DLL's (we 
 *  cannot make the static decl member both dllexport and dllimport)
 *  and second this method ensures that the member has been initialized
 *  already.
 *  This method will assert if no class with the given name was
 *  found.
 */
KLAYOUT_DLL const ClassBase *class_by_name (const std::string &name);

/**
 *  @brief Test whether there is a class with the given name
 */
KLAYOUT_DLL bool has_class (const std::string &name);

}

#endif


