
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "tlExpression.h"
#include "gsiExpression.h"
#include "gsiDecl.h"

#include <set>
#include <map>
#include <list>
#include <cstdio>

namespace gsi
{

// ---------------------------------------------------------------------

static std::string 
method_name_for_expr (const std::string &n)
{
  if (n.size () > 0 && n [n.size () - 1] == '?') {
    return std::string (n, 0, n.size () - 1);
  } else {
    return n;
  }
}

// -------------------------------------------------------------------
//  Test if an argument can be converted to the given type

struct TestArgFunc
{
  TestArgFunc (const tl::Variant &arg, const gsi::ArgType &atype)
    : m_arg (arg), m_atype (atype), m_ret (false)
  { }

  bool value () const { return m_ret; }

  template <class R>
  void op ()
  {
    m_ret = m_arg.can_convert_to<R> ();
  }

private:
  const tl::Variant &m_arg;
  const gsi::ArgType &m_atype;
  bool m_ret;
};

bool
test_arg (const gsi::ArgType &atype, const tl::Variant &arg)
{
  //  for const X * or X *, nil is an allowed value
  if ((atype.is_cptr || atype.is_ptr) && arg.is_nil ()) {
    return true;
  }

  if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const ArgType &ainner = *atype.inner;

    if (! arg.is_list ()) {
      return false;
    }

    for (tl::Variant::const_iterator v = arg.begin (); v != arg.end (); ++v) { 

      TestArgFunc f (*v, ainner);

      if (switch_type (ainner.type, f)) {
        return f.value ();
      } else {

        switch (ainner.type) {
        case gsi::T_void:
          tl_assert (false);
          break;
        case gsi::T_object:
          {
            if (! v->is_user ()) {
              return false;
            }
            const gsi::ClassBase *gsi_cls = v->gsi_cls ();
            if (!gsi_cls || !gsi_cls->is_derived_from (ainner.cls)) {
              return false;
            }
          }
          break;
        case gsi::T_object_new:
          tl_assert (false);  // TODO: not implemented yet.
        default:
          break;
        }

      }

    }

    return true;

  } else {

    TestArgFunc f (arg, atype);

    if (switch_type (atype.type, f)) {
      return f.value ();
    } else {

      switch (atype.type) {
      case gsi::T_void:
        return true;
      case gsi::T_object:
        {
          if ((atype.is_ptr || atype.is_cptr) && arg.is_nil ()) {
            return true;
          }
          if (! arg.is_user ()) {
            return false;
          }
          const tl::VariantUserClassBase *cls = arg.user_cls ();
          if (atype.is_cptr || atype.is_cref) {
            return cls && cls->gsi_cls ()->is_derived_from (atype.cls);
          } else {
            return cls && !cls->is_const () && cls->gsi_cls ()->is_derived_from (atype.cls);
          }
        }
      case gsi::T_object_new:
        tl_assert (false);  // An argument cannot be "new"
        return false;
      default:
        return true;
      }

    }

  }
}
      
// -------------------------------------------------------------------
//  Variant to C conversion

template <class R>
struct var2c
{
  static R get (const tl::Variant &rval)
  {
    return rval.to<R> ();
  }
};

template <>
struct var2c<tl::Variant>
{
  static const tl::Variant &get (const tl::Variant &rval)
  {
    return rval;
  }
};

// ---------------------------------------------------------------------
//  

template <class R>
struct set_value
{
  static void set (gsi::SerialArgs &aa, const tl::Variant &arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg.is_nil () && atype.type != gsi::T_var) {
      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments of reference or direct type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (typename gsi::type_traits<R *>::tag (), (R *)0);
      } else {
        aa.set_value (typename gsi::type_traits<const R *>::tag (), (const R *)0);
      }
    } else {

      if (atype.is_ref || atype.is_ptr) {

        // TODO: morph the variant to the requested type and pass it's pointer (requires a non-const reference for arg)
        // -> we would have a reference that can modify the argument (out parameter).
        tmp_stack.push_back (gsi::TempObject ());
        R *v = new R (var2c<R>::get (arg));
        tmp_stack.back ().set (v);

        aa.set_value (typename gsi::type_traits<void *>::tag (), v);

      } else if (atype.is_cref) {
        aa.set_value (typename gsi::type_traits<const R &>::tag (), var2c<R>::get (arg));
      } else if (atype.is_cptr) {
        R r = var2c<R>::get (arg);
        aa.set_value (typename gsi::type_traits<const R *>::tag (), &r);
      } else {
        aa.set_value (typename gsi::type_traits<R>::tag (), var2c<R>::get (arg));
      }

    }
  }
};

template <>
struct set_value<std::string>
{
  static void set (gsi::SerialArgs &aa, const tl::Variant &arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg.is_nil ()) {
      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments of reference or direct type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (gsi::type_traits<std::string *>::tag (), (std::string *)0);
      } else {
        aa.set_value (gsi::type_traits<const std::string *>::tag (), (const std::string *)0);
      }
    } else {

      if (atype.is_ref || atype.is_ptr || atype.is_cref || atype.is_cptr) {

        // TODO: morph the variant to the requested type and pass it's pointer (requires a non-const reference for arg)
        // -> we would have a reference that can modify the argument (out parameter).
        tmp_stack.push_back (gsi::TempObject ());
        std::string *v = new std::string (var2c<std::string>::get (arg));
        tmp_stack.back ().set (v);

        aa.set_value (gsi::type_traits<void *>::tag (), v);

      } else {
        aa.set_value (gsi::type_traits<std::string>::tag (), var2c<std::string>::get (arg));
      }

    }
  }
};

template <>
struct set_value<QString>
{
  static void set (gsi::SerialArgs &aa, const tl::Variant &arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg.is_nil ()) {
      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments of reference or direct type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (gsi::type_traits<QString *>::tag (), (QString *)0);
      } else {
        aa.set_value (gsi::type_traits<const QString *>::tag (), (const QString *)0);
      }
    } else {

      if (atype.is_ref || atype.is_ptr || atype.is_cref || atype.is_cptr) {

        // TODO: morph the variant to the requested type and pass it's pointer (requires a non-const reference for arg)
        // -> we would have a reference that can modify the argument (out parameter).
        tmp_stack.push_back (gsi::TempObject ());
        QString *v = new QString (var2c<QString>::get (arg));
        tmp_stack.back ().set (v);

        aa.set_value (gsi::type_traits<void *>::tag (), v);

      } else {
        aa.set_value (gsi::type_traits<QString>::tag (), var2c<QString>::get (arg));
      }

    }
  }
};

//  TODO: simplify all NPOD types ...
template <>
struct set_value<tl::Variant>
{
  static void set (gsi::SerialArgs &aa, const tl::Variant &arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    //  TODO: clarify: is nil a zero-pointer to a variant or a pointer to a "nil" variant
    if (atype.is_ref || atype.is_ptr || atype.is_cref || atype.is_cptr) {

      // TODO: better to simply pass the tl::Variant's pointer? Would require a non-const Variant ref for example.
      tmp_stack.push_back (gsi::TempObject ());
      tl::Variant *v = new tl::Variant (arg);
      tmp_stack.back ().set (v);

      aa.set_value (gsi::type_traits<void *>::tag (), v);

    } else {
      aa.set_value (gsi::type_traits<tl::Variant>::tag (), var2c<tl::Variant>::get (arg));
    }

  }
};

template <class R>
struct set_vector
{
  static void set (gsi::SerialArgs &aa, const tl::Variant &arr, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (! arr.is_list ()) {
      throw tl::Exception (tl::translate ("Array expected instead of") + " '" + arr.to_string () + "'");
    }

    if (atype.is_ref || atype.is_cref || atype.is_ptr || atype.is_cptr) {

      tmp_stack.push_back (gsi::TempObject ());

      std::vector<R> *v = new std::vector<R> ();
      for (tl::Variant::const_iterator a = arr.begin (); a != arr.end (); ++a) { 
        v->push_back (var2c<R>::get (*a));
      }
      tmp_stack.back ().set (v);

      if (atype.is_cref) {
        aa.set_value (typename gsi::type_traits<const std::vector<R> &>::tag (), *v);
      } else if (atype.is_ref) {
        aa.set_value (typename gsi::type_traits<std::vector<R> &>::tag (), *v);
      } else if (atype.is_cptr) {
        aa.set_value (typename gsi::type_traits<const std::vector<R> *>::tag (), v);
      } else if (atype.is_ptr) {
        aa.set_value (typename gsi::type_traits<std::vector<R> *>::tag (), v);
      }

    } else {
      std::vector<R> *v = aa.set_value (typename gsi::type_traits< std::vector<R> >::tag (), std::vector<R> ());
      for (tl::Variant::const_iterator a = arr.begin (); a != arr.end (); ++a) { 
        v->push_back (var2c<R>::get (*a));
      }
    }
  }
};

// ---------------------------------------------------------------------
//  

struct PushArgFunc
{
  PushArgFunc (const gsi::ArgType &atype, gsi::SerialArgs &aserial, const tl::Variant &arg, std::list<gsi::TempObject> &tmp_stack)
    : m_atype (atype), m_aserial (aserial), m_arg (arg), m_tmp_stack (tmp_stack)
  { }

  template <class R>
  void op () 
  {
    set_value<R>::set (m_aserial, m_arg, m_atype, m_tmp_stack);
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  const tl::Variant &m_arg;
  std::list<gsi::TempObject> &m_tmp_stack;
};

struct PushArgFuncVector
{
  PushArgFuncVector (const gsi::ArgType &atype, gsi::SerialArgs &aserial, const tl::Variant &arg, std::list<gsi::TempObject> &tmp_stack)
    : m_atype (atype), m_aserial (aserial), m_arg (arg), m_tmp_stack (tmp_stack)
  { }

  template <class R>
  void op () 
  {
    set_vector<R>::set (m_aserial, m_arg, m_atype, m_tmp_stack); 
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  const tl::Variant &m_arg;
  std::list<gsi::TempObject> &m_tmp_stack;
};

static void
push_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, const tl::Variant &arg, std::list<gsi::TempObject> &tmp_stack)
{
  PushArgFunc f (atype, aserial, arg, tmp_stack);

  if (switch_type (atype.type, f)) {

    //  handled in switch

  } else if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const ArgType &ainner = *atype.inner;

    PushArgFuncVector f (atype, aserial, arg, tmp_stack);

    if (switch_type (ainner.type, f)) {

      //  handled in switch
    
    } else if (ainner.type == T_object) {

      if (! arg.is_list ()) {
        throw tl::Exception (tl::translate ("Array expected instead of") + " '" + arg.to_string () + "'");
      }

      std::vector<void *> vobj;
      for (tl::Variant::const_iterator a = arg.begin (); a != arg.end (); ++a) { 
        if (! a->is_user ()) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), ainner.cls->name ()));
        }
        const tl::VariantUserClassBase *cls = a->user_cls ();
        if (!cls || !cls->gsi_cls ()->is_derived_from (ainner.cls)) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), ainner.cls->name ()));
        }
        //  TODO: get rid of that const_cast
        vobj.push_back (const_cast<void *> (a->to_user_object ()));
      }

      ainner.cls->push_vector_of (aserial, atype, tmp_stack, vobj);

    } else if (ainner.type == T_object_new) {
      tl_assert (false);  // TODO: not implemented yet.
    } else if (ainner.type == T_void) {
      tl_assert (false);  // Not supported
    }

  } else if (atype.type == gsi::T_object) {

    if (atype.is_ref || atype.is_cref || atype.is_ptr || atype.is_cptr) {

      if (arg.is_nil ()) {
        if (atype.is_ref || atype.is_cref) {
          throw tl::Exception (tl::translate ("Cannot pass nil to reference parameters"));
        }
        aserial.set_value (gsi::type_traits<void *>::tag (), (void *) 0);
      } else {
        if (! arg.is_user ()) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), atype.cls->name ()));
        }
        const tl::VariantUserClassBase *cls = arg.user_cls ();
        if (!cls || !cls->gsi_cls ()->is_derived_from (atype.cls)) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), atype.cls->name ()));
        }
        if (cls->is_const () && (atype.is_ref || atype.is_ptr)) {
          throw tl::Exception (tl::sprintf (tl::translate ("Cannot pass a const reference of class %s to a non-const reference or pointer parameter"), atype.cls->name ()));
        }
        aserial.set_value (gsi::type_traits<void *>::tag (), arg.to_user_object ());
      }

    } else {

      if (! arg.is_user ()) {
        throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), atype.cls->name ()));
      }
      const tl::VariantUserClassBase *cls = arg.user_cls ();
      if (!cls || !cls->gsi_cls ()->is_derived_from (atype.cls)) {
        throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)"), atype.cls->name ()));
      }
      //  TODO: get rid of that const_cast
      atype.cls->copy_to (aserial, const_cast<void *> (arg.to_user_object ()));

    }

  } else if (atype.type == gsi::T_object_new) {
    tl_assert (false);  // An argument cannot be "new"
  }
}

// -------------------------------------------------------------------
//  C to tl::Variant conversion

template <class R> 
struct retargs
{
  static void
  get_value (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    if (atype.is_ref) {
      out = rr.template get_value<R &> (typename gsi::type_traits<R &>::tag ());
    } else if (atype.is_cref) {
      out = rr.template get_value<const R &> (typename gsi::type_traits<const R &>::tag ());
    } else if (atype.is_ptr) {
      R *p = rr.template get_value<R *> (typename gsi::type_traits<R *>::tag ());
      if (p == 0) {
        out = tl::Variant ();
      } else {
        out = *p;
      }
    } else if (atype.is_cptr) {
      const R *p = rr.template get_value<const R *> (typename gsi::type_traits<const R *>::tag ());
      if (p == 0) {
        out = tl::Variant ();
      } else {
        out = *p;
      }
    } else {
      out = rr.template get_value<R> (typename gsi::type_traits<R>::tag ());
    }
  }

  static void
  get_vector (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    if (atype.is_ref) {

      std::vector<R> &v = rr.template get_value<std::vector<R> &> (typename gsi::type_traits<std::vector<R> &>::tag ());

      out = tl::Variant::empty_list ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        out.push (tl::Variant ());
        out.back () = (const R &) *i;
      }

    } else if (atype.is_cref) {

      const std::vector<R> &v = rr.template get_value<const std::vector<R> &> (typename gsi::type_traits<const std::vector<R> &>::tag ());

      out = tl::Variant::empty_list ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        out.push (tl::Variant ());
        out.back () = (const R &) *i;
      }

    } else if (atype.is_ref) {

      std::vector<R> *v = rr.template get_value<std::vector<R> *> (typename gsi::type_traits<std::vector<R> *>::tag ());

      if (v) {
        out = tl::Variant::empty_list ();
        for (typename std::vector<R>::const_iterator i = v->begin (); i != v->end (); ++i) {
          out.push (tl::Variant ());
          out.back () = (const R &) *i;
        }
      } else {
        out = tl::Variant ();
      }

    } else if (atype.is_cref) {

      const std::vector<R> *v = rr.template get_value<const std::vector<R> *> (typename gsi::type_traits<const std::vector<R> *>::tag ());

      if (v) {
        out = tl::Variant::empty_list ();
        for (typename std::vector<R>::const_iterator i = v->begin (); i != v->end (); ++i) {
          out.push (tl::Variant ());
          out.back () = (const R &) *i;
        }
      } else {
        out = tl::Variant ();
      }

    } else {

      std::vector<R> v = rr.template get_value< std::vector<R> > (typename gsi::type_traits< std::vector<R> >::tag ());

      out = tl::Variant::empty_list ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        out.push (tl::Variant ());
        out.back () = (const R &) *i;
      }

    }
  }
};

template <> 
struct retargs<void *>
{
  static void
  get_value (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    tl_assert (!atype.is_ref);
    tl_assert (!atype.is_cref);
    tl_assert (!atype.is_ptr);
    tl_assert (!atype.is_cptr);
    out = size_t (rr.get_value<void *> (gsi::type_traits<void *>::tag ()));
  }

  static void
  get_vector (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    tl_assert (false); //  not implemented
  }
};

template <> 
struct retargs<const char *>
{
  static void
  get_value (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    tl_assert (!atype.is_ref);
    tl_assert (!atype.is_cref);
    tl_assert (!atype.is_ptr);
    tl_assert (!atype.is_cptr);
    out = rr.get_value<const char *> (gsi::type_traits<const char *>::tag ());
  }

  static void
  get_vector (tl::Variant &out, gsi::SerialArgs &rr, const gsi::ArgType &atype)
  {
    tl_assert (false); //  not implemented
  }
};

// -------------------------------------------------------------------
//  Some helper stuff for extracting arrays

struct MakeArrayData
{
  MakeArrayData (tl::Variant *_ret, const tl::VariantUserClassBase *_cls)
    : ret (_ret), cls (_cls)
  { }

  tl::Variant *ret;
  const tl::VariantUserClassBase *cls;
};

static void
make_array_f (void *data, void *cobj)
{
  MakeArrayData *ad = (MakeArrayData *)data;
  ad->ret->push (tl::Variant ());
  ad->ret->back ().set_user (0, cobj, ad->cls);
}

// -------------------------------------------------------------------
//  Pop an argument from the call or return stack

struct PopArgFunc
{
  PopArgFunc (tl::Variant &out, const gsi::ArgType &atype, gsi::SerialArgs &aserial)
    : m_out (out), m_atype (atype), m_aserial (aserial)
  { }

  template <class R>
  void op () 
  {
    retargs<R>::get_value (m_out, m_aserial, m_atype);
  }

private:
  tl::Variant &m_out;
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
};

struct PopArgFuncVector
{
  PopArgFuncVector (tl::Variant &out, const gsi::ArgType &atype, gsi::SerialArgs &aserial)
    : m_out (out), m_atype (atype), m_aserial (aserial)
  { }

  template <class R>
  void op () 
  {
    retargs<R>::get_vector (m_out, m_aserial, m_atype);
  }

private:
  tl::Variant &m_out;
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
};

void
pop_arg (tl::Variant &out, const gsi::ArgType &atype, gsi::SerialArgs &aserial)
{
  out = tl::Variant ();

  PopArgFunc f (out, atype, aserial);

  if (atype.is_iter) {

    tl_assert (false); // not implemented here 

  } else if (switch_type (atype.type, f)) {

    //  type handled in switch

  } else if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const ArgType &ainner = *atype.inner;

    PopArgFuncVector f (out, atype, aserial);

    if (switch_type (ainner.type, f)) {

      //  type handled in switch

    } else if (ainner.type == gsi::T_object) {

      out = tl::Variant::empty_list ();
      const tl::VariantUserClassBase *cls = ainner.cls->var_cls (false, false);
      tl_assert (cls != 0);
      MakeArrayData ad (&out, cls);
      ainner.cls->get_vector_of (aserial, atype, (void *) &ad, &make_array_f);

    } else if (ainner.type == gsi::T_void) {
      tl_assert (false); // not supported
    } else if (ainner.type == gsi::T_object_new) {
      tl_assert (false); // TODO: not implemented
    }

  } else if (atype.type == gsi::T_object) {

    if (atype.is_ptr || atype.is_cptr || atype.is_ref || atype.is_cref) {
      void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
      if (obj == 0) {
        out = tl::Variant ();
      } else {
        const tl::VariantUserClassBase *cls = atype.cls->var_cls (true, atype.is_cref || atype.is_cptr);
        tl_assert (cls != 0);
        out = tl::Variant ();
        out.set_user (0, obj, cls);
      }
    } else {
      const tl::VariantUserClassBase *cls = atype.cls->var_cls (false, false);
      tl_assert (cls != 0);
      out = tl::Variant ();
      out.set_user (0, atype.cls->take_and_create (aserial), cls);
    }

  } else if (atype.type == gsi::T_object_new) {

    void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
    if (obj == 0) {
      out = tl::Variant ();
    } else {
      const tl::VariantUserClassBase *cls = atype.cls->var_cls (false, false);
      tl_assert (cls != 0);
      out = tl::Variant ();
      out.set_user (0, obj, cls);
    }

  }
}

// ---------------------------------------------------------------------
//  Implementation of initialize_expressions

class EvalClassFunction
  : public tl::EvalFunction
{
public:
  EvalClassFunction (const tl::VariantUserClassBase *var_cls)
    : mp_var_cls (var_cls)
  {
    //  .. nothing yet ..
  }

  void execute (const tl::ExpressionParserContext &context, tl::Variant &out, const std::vector<tl::Variant> &args) const 
  {
    if (! args.empty ()) {
      throw tl::Exception (tl::translate ("Class is not a function - use 'new' to create a new object"));
    }
    out = tl::Variant (0, 0, mp_var_cls);
  }

private:
  const tl::VariantUserClassBase *mp_var_cls;
};

//  TODO: this method does a lot more than just initializing "expressions"
//  Basically this method initializes the complete GSI type system.

void KLAYOUT_DLL 
initialize_expressions ()
{
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    //  TODO: get rid of that const cast
    (const_cast<gsi::ClassBase *> (&*c))->initialize ();
  }

  //  merge the extensions to the main declaration
  gsi::ClassBase::merge_declarations ();

  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    //  Initialize the method table once again after we have merged the declarations
    //  TODO: get rid of that const cast
    (const_cast<gsi::ClassBase *> (&*c))->initialize ();

    //  there should be only main declarations since we merged
    tl_assert (c->declaration () == &*c);

    const tl::VariantUserClassBase *cc = c->var_cls_cls ();

    //  register a function that creates a class object (use a function to avoid issues with
    //  late destruction of global variables which the class object is already gone)
    if (cc) {
      tl::Eval::define_global_function (c->name (), new EvalClassFunction (cc));
    }
  }

  tl::initialize_variant_class_table ();
}

// -------------------------------------------------------------------------
//  VariantUserClassImpl implementation

VariantUserClassImpl::VariantUserClassImpl () 
  : mp_cls (0), mp_self (0), mp_object_cls (0), m_reference (false), m_is_const (false)
{ 
  //  .. nothing yet ..
}

void 
VariantUserClassImpl::initialize (const gsi::ClassBase *cls, const tl::VariantUserClassBase *self, const tl::VariantUserClassBase *object_cls, bool reference, bool is_const)
{
  mp_cls = cls;
  mp_self = self;
  mp_object_cls = object_cls;
  m_reference = reference; 
  m_is_const = is_const;

  if (mp_cls) {

    for (gsi::ClassBase::name_map_iterator nm = cls->begin_name_map (); nm != cls->end_name_map (); ++nm) {

      unsigned int mid = nm->second;

      //  the first overloaded method determines the type of method - this implies that all
      //  overloaded methods must be static or none.
      const gsi::MethodBase *m_first = cls->begin_methods () [*cls->begin_overloaded_methods (mid)];

      if (! m_first->is_callback () && (m_first->is_static () == (object_cls != 0))) {
        m_method_ids.insert (std::make_pair (method_name_for_expr (nm->first.first), mid));
      }

    }

  }
}

VariantUserClassImpl::~VariantUserClassImpl () 
{ 
  mp_cls = 0;
}

bool 
VariantUserClassImpl::equal_impl (void *obj, void *other) const 
{
  if (obj) {

    std::map <std::string, int>::const_iterator mm = m_method_ids.find ("==");
    if (mm == m_method_ids.end ()) {

      //  No == method - use object identity
      return (void *) this == other;

    } else {

      tl::ExpressionParserContext context;

      tl::Variant out;

      VariantUserClass<void> ref_cls;
      ref_cls.initialize (mp_cls, mp_object_cls, true, true);
      tl::Variant object (0, obj, &ref_cls);
      std::vector<tl::Variant> vv;
      vv.resize (1, tl::Variant ());
      vv[0].set_user (0, other, &ref_cls);

      execute_gsi (context, out, object, "==", vv);

      return out.to_bool ();

    }

  } else {
    return false; 
  }
}

bool 
VariantUserClassImpl::less_impl (void *obj, void *other) const 
{
  if (obj) {

    std::map <std::string, int>::const_iterator mm = m_method_ids.find ("<");
    if (mm == m_method_ids.end ()) {

      //  No < method - use object pointers
      return (void *) this < other;

    } else {

      tl::ExpressionParserContext context;

      tl::Variant out;

      VariantUserClass<void> ref_cls;
      ref_cls.initialize (mp_cls, mp_object_cls, true, true);
      tl::Variant object (0, obj, &ref_cls);
      std::vector<tl::Variant> vv;
      vv.resize (1, tl::Variant ());
      vv[0].set_user (0, other, &ref_cls);

      execute_gsi (context, out, object, "<", vv);

      return out.to_bool ();

    }

  } else {
    return false; 
  }
}

std::string 
VariantUserClassImpl::to_string_impl (void *obj) const 
{
  if (obj) {

    std::map <std::string, int>::const_iterator mm = m_method_ids.find ("to_s");
    if (mm == m_method_ids.end ()) {

      //  no method to convert the object to a string
      return std::string ();

    } else {

      tl::ExpressionParserContext context;

      tl::Variant out;

      VariantUserClass<void> ref_cls;
      ref_cls.initialize (mp_cls, mp_object_cls, true, true);
      tl::Variant object (0, obj, &ref_cls);
      std::vector<tl::Variant> vv;

      execute_gsi (context, out, object, "to_s", vv);

      return out.to_string ();

    }

  } else {
    return std::string (); 
  }
}

void
VariantUserClassImpl::execute (const tl::ExpressionParserContext &context, tl::Variant &out, tl::Variant &object, const std::string &method, const std::vector<tl::Variant> &args) const
{
  if (mp_object_cls == 0 && method == "is_a") {

    if (args.size () != 1) {
      throw tl::EvalError (tl::translate ("'is_a' method requires exactly one argument"), context);
    }

    bool ret = false;
    if (args [0].is_user ()) {
      const tl::VariantUserClassBase *ub = args [0].user_cls ();
      if (ub && ub->gsi_cls () == mp_cls) {
        ret = true;
      }
    }

    out = ret;

  } else if (mp_object_cls != 0 && method == "new" && args.size () == 0) {

    void *obj = mp_cls->create ();
    if (obj) {
      out.set_user (0, obj, mp_object_cls);
    } else {
      out.reset ();
    }

  } else if (mp_object_cls == 0 && method == "dup") {

    if (args.size () != 0) {
      throw tl::EvalError (tl::translate ("'dup' method does not allow arguments"), context);
    }

    void *obj = mp_cls->create ();
    if (obj) {
      mp_cls->assign (obj, object.to_user_object ());
      out.set_user (0, obj, mp_self);
    } else {
      out.reset ();
    }

  } else {
    try {
      execute_gsi (context, out, object, method, args);
    } catch (tl::EvalError &) {
      throw;
    } catch (tl::Exception &ex) {
      throw tl::EvalError (ex.msg (), context);
    }
  }
}

void
VariantUserClassImpl::execute_gsi (const tl::ExpressionParserContext &context, tl::Variant &out, tl::Variant &object, const std::string &method, const std::vector<tl::Variant> &args) const
{
  std::map <std::string, int>::const_iterator mm = m_method_ids.find (method);
  if (mm == m_method_ids.end ()) {
    throw tl::Exception (tl::translate ("Unknown method") + " '" + method + "'");
  }

  int mid = mm->second;

  //  get number of candidates by argument count
  const gsi::MethodBase *meth = 0;
  unsigned int candidates = 0;
  for (gsi::ClassBase::overloaded_methods_iterator om = mp_cls->begin_overloaded_methods (mid); om != mp_cls->end_overloaded_methods (mid); ++om) {
    const gsi::MethodBase *m = mp_cls->begin_methods () [*om];
    if (m->is_event()) {
      throw tl::Exception (tl::sprintf (tl::translate ("Events are not supported inside expressions (event %s)"), mp_cls->overloaded_name (mid).c_str ()));
    } else if (m->is_callback()) {
      //  ignore callbacks
    } else if (std::distance (m->begin_arguments (), m->end_arguments ()) == (int) args.size ()) {
      ++candidates;
      meth = m;
    }
  }

  //  no candidate -> error
  if (! meth) {
    std::set<unsigned int> nargs;
    for (gsi::ClassBase::overloaded_methods_iterator om = mp_cls->begin_overloaded_methods (mid); om != mp_cls->end_overloaded_methods (mid); ++om) {
      const gsi::MethodBase *m = mp_cls->begin_methods () [*om];
      nargs.insert (std::distance (m->begin_arguments (), m->end_arguments ()));
    }
    std::string nargs_s;
    for (std::set<unsigned int>::const_iterator na = nargs.begin (); na != nargs.end (); ++na) {
      if (na != nargs.begin ()) {
        nargs_s += "/";
      }
      nargs_s += tl::to_string (*na);
    }

    throw tl::Exception (tl::sprintf (tl::translate ("Invalid number of arguments for method %s, class %s (got %d, expected %s)"), mp_cls->overloaded_name (mid), mp_cls->name (), int (args.size ()), nargs_s));
  }

  //  more than one candidate -> refine by checking the arguments
  if (candidates > 1) {
    meth = 0;
    candidates = 0;
    for (gsi::ClassBase::overloaded_methods_iterator om = mp_cls->begin_overloaded_methods (mid); om != mp_cls->end_overloaded_methods (mid); ++om) {
      const gsi::MethodBase *m = mp_cls->begin_methods () [*om];
      bool is_valid = (std::distance (m->begin_arguments (), m->end_arguments ()) == int (args.size ()));
      size_t narg = 0;
      for (gsi::MethodBase::argument_iterator a = m->begin_arguments (); is_valid && a != m->end_arguments (); ++a, ++narg) {
        if (! test_arg (*a, args [narg])) {
          is_valid = false;
        }
      }
      if (is_valid) {
        ++candidates;
        meth = m;
      }
    }
  }

  if (! meth) {
    throw tl::Exception (tl::sprintf (tl::translate ("No method with matching arguments for method %s, class %s"), mp_cls->overloaded_name (mid), mp_cls->name ()));
  }

  if (candidates > 1) {
    throw tl::Exception (tl::sprintf (tl::translate ("Ambiguous overload variants for method %s, class %s - multiple method declarations match arguments"), mp_cls->overloaded_name (mid), mp_cls->name ()));
  }

  if (m_is_const && ! meth->is_const ()) {
    throw tl::Exception (tl::sprintf (tl::translate ("Cannot call non-const method %s, class %s on a const reference"), mp_cls->overloaded_name (mid), mp_cls->name ()));
  }

  if (meth->is_event ()) {

    //  .. not supported yet ..

    out = tl::Variant ();

  } else {

    std::list<gsi::TempObject> tmp_stack;

    gsi::SerialArgs arglist (meth->argsize ());
    gsi::SerialArgs retlist (meth->retsize ());
    size_t narg = 0;
    for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a, ++narg) {
      push_arg (*a, arglist, args [narg], tmp_stack);
    }

    meth->call (object.to_user_object (), arglist, retlist);

    if (meth->ret_type ().is_iter) {

#if 0
      //  .. iterators not supported yet ..

      switch (meth->ret_type ().type) {
      case gsi::T_void:
        break;
      case gsi::T_bool:
        yield_iter<bool>::yield (retlist);
        break;
      case gsi::T_int:
        yield_iter<int>::yield (retlist);
        break;
      case gsi::T_uint:
        yield_iter<unsigned int>::yield (retlist);
        break;
      case gsi::T_long:
        yield_iter<long>::yield (retlist);
        break;
      case gsi::T_ulong:
        yield_iter<unsigned long>::yield (retlist);
        break;
      case gsi::T_longlong:
        yield_iter<long long>::yield (retlist);
        break;
      case gsi::T_ulonglong:
        yield_iter<unsigned long long>::yield (retlist);
        break;
      case gsi::T_double:
        yield_iter<double>::yield (retlist);
        break;
      case gsi::T_string_ccptr:
        yield_iter<const char *>::yield (retlist);
        break;
      case gsi::T_string:
        yield_iter<std::string>::yield (retlist);
        break;
      case gsi::T_var:
        yield_iter<tl::Variant>::yield (retlist);
        break;
      case gsi::T_object:
        {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (meth->ret_type ().cls);
          tl_assert (c != rev_cls_map.end ());
          void *iter = c->first->create_obj_iterator (retlist);
          try {
            while (! c->first->obj_iterator_at_end (iter)) {
              VALUE obj = rb_class_new_instance (0, 0, c->second);
              Proxy *p;
              Data_Get_Struct (obj, Proxy, p);
              p->set (c->first->take_iterator_obj (iter), true, false);
              rb_yield (obj);
              c->first->inc_obj_iterator (iter);
            }
            c->first->delete_obj_iterator (iter);
          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }
        }
        break;
      case gsi::T_object_ref:
      case gsi::T_object_cref:
        {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (meth->ret_type ().cls);
          tl_assert (c != rev_cls_map.end ());
          void *iter = c->first->create_obj_iterator (retlist);
          try {
            while (! c->first->obj_iterator_at_end (iter)) {
              void *cobj = const_cast<void *> (c->first->get_iterator_obj (iter));
              VALUE obj;
              if (cobj == 0) {
                obj = Qnil;
              } else if (meth->ret_type ().cls->client_data (cobj) != 0 && 
                         ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->get_ref () != Qnil) {
                //  we have a that is located in C++ space but is supposed to get attached
                //  a Ruby object. If it already has, we simply return a reference to this
                obj = ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->get_ref ();
              } else {
                obj = rb_class_new_instance (0, 0, c->second);
                Proxy *p;
                Data_Get_Struct (obj, Proxy, p);
                p->set (cobj, false, meth->ret_type ().type == gsi::T_object_cref);
                //  attach the ruby object to the C++ object if possible
                if (meth->ret_type ().cls->client_data (cobj) != 0) {
                  ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->set_ref (obj, true);
                }
              }
              rb_yield (obj);
              c->first->inc_obj_iterator (iter);
            }
            c->first->delete_obj_iterator (iter);
          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }
        }
        break;
      case gsi::T_object_new:
        tl_assert (false);
        break;
      }
#endif

      out = tl::Variant ();

    } else {
      pop_arg (out, meth->ret_type (), retlist);
    }

  }
}

}

