
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiObject
#define _HDR_gsiObject

#include <memory>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

template <class X> class Class;
class ClassBase;

/**
 *  @brief The handler for the client-side data of Object<X>
 *
 *  If the scripting client desires to attach data to the Object<X> objects, it
 *  must reimplement this handler class and register a handler for each ClassBase
 *  object that is registered.
 */
class ClientHandler
{
public:
  ClientHandler () { }
  virtual ~ClientHandler () { }
  virtual void attach (void *&data) = 0;
  virtual void detach (void *&data) = 0;
  //  indicates that the C++ side wants to keep the object and transfer ownership from the script client to a C++ object
  virtual void keep (void *&data) = 0;
  //  indicates that the C++ side no longer wants to keep the object and transfer ownership from C++ back to the script client
  virtual void release (void *&data) = 0;
};

//  TODO: currently the client data is not copied. As a consequence, copying the C++
//  objects does not copy the Ruby part of the object but rather leaves it unitialized.
//  Since however, the client-side data feature is supposed to be used to non-copying 
//  objects mainly, this problem is not considered serious for now.

/**
 *  @brief A base class for objects that are connected to their Ruby counterpart
 */
class KLAYOUT_DLL ObjectBase
{
public:
  ObjectBase () 
    : mp_data (0)
  {
    attach (this);
  }

  ObjectBase (const ObjectBase &) 
    : mp_data (0)
  {
    attach (this);
  }

  ObjectBase &operator= (const ObjectBase &)
  {
    return *this;
  }

  virtual ~ObjectBase ()
  {
    detach (this);
    mp_data = 0;
  }

  void *data () const
  {
    return mp_data;
  }

  ClientHandler *client_handler () const
  {
    return ms_client_handler.get();
  }

  static void register_client_handler (ClientHandler *handler) 
  {
    ms_client_handler.reset (handler);
  }

  static void attach (ObjectBase *obj) 
  {
    if (ms_client_handler.get()) {
      ms_client_handler->attach (obj->mp_data);
    }
  }

  static void detach (ObjectBase *obj) 
  {
    if (ms_client_handler.get() && obj->mp_data) {
      ms_client_handler->detach (obj->mp_data);
    }
  }

  void keep () 
  {
    if (ms_client_handler.get()) {
      ms_client_handler->keep (mp_data);
    }
  }

  void release () 
  {
    if (ms_client_handler.get()) {
      ms_client_handler->release (mp_data);
    }
  }

private:
  static std::auto_ptr<ClientHandler> ms_client_handler;
  void *mp_data;
};

}

#endif

