
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiSerialisation
#define _HDR_gsiSerialisation

#include <list>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

struct ArglistUnderflowException
  : public tl::Exception
{
  ArglistUnderflowException ()
    : tl::Exception (tl::translate ("Too few arguments or no return value supplied"))
  { }
};

struct NilPointerToReference
  : public tl::Exception
{
  NilPointerToReference ()
    : tl::Exception (tl::translate ("nil object passed to a reference"))
  { }
};

/**
 *  @brief This class provides the basic argument serialization mechanism for the C++/scripting interface
 */
class SerialArgs 
{
public:
  SerialArgs ()
    : mp_buffer (0)
  {
    mp_write = mp_read = mp_buffer;
  }

  SerialArgs (unsigned int len)
    : mp_buffer (len > 0 ? new char [len] : 0)
  {
    mp_write = mp_read = mp_buffer;
  }

  ~SerialArgs ()
  {
    if (mp_buffer) {
      delete [] mp_buffer;
    }
    mp_buffer = 0;
  }

  char *cptr ()
  {
    return mp_buffer;
  }

  const char *cptr () const
  {
    return mp_buffer;
  }

  template <class X> 
  inline X *set_value (const direct_tag &, const X &x)
  {
    X *r = (X *)mp_write;
    new (r) X(x);
    mp_write += item_size<X> ();
    return r;
  }

  template <class X> 
  inline void set_value (const ref_tag &, X &x) 
  {
    *((X **)mp_write) = &x;
    mp_write += item_size<X *> ();
  }

  template <class X> 
  inline void set_value (const ptr_tag &, X *x)
  {
    *((X **)mp_write) = x;
    mp_write += item_size<X *> ();
  }

  template <class X> 
  inline void set_value (const pod_cref_tag &, const X &x)
  {
    X *r = (X *)mp_write;
    new (r) X(x);
    mp_write += item_size<X> ();
  }

  template <class X> 
  inline void set_value (const npod_cref_tag &, const X &x)
  {
    *((X const **)mp_write) = &x;
    mp_write += item_size<const X *> ();
  }

  template <class X> 
  inline void set_value (const x_cref_tag &, const X &x)
  {
    *((X const **)mp_write) = &x;
    mp_write += item_size<const X *> ();
  }

  template <class X> 
  inline void set_value (const pod_cptr_tag &, const X *x)
  {
    *(bool *)mp_write = (x != 0);
    mp_write += item_size<bool> ();
    if (x) {
      *(X *)mp_write = *x;
    }
    mp_write += item_size<X> ();
  }

  //  see notes on the serialization for this type:
  template <class X> 
  inline void set_value (const npod_cptr_tag &, const X *x)
  {
    *((X const **)mp_write) = x;
    mp_write += item_size<const X *> ();
  }

  template <class X> 
  inline void set_value (const x_cptr_tag &, const X *x)
  {
    *((X const **)mp_write) = x;
    mp_write += item_size<const X *> ();
  }

  inline void set_value (const vptr_tag &, void *x)
  {
    *((void **)mp_write) = x;
    mp_write += item_size<void *> ();
  }

  inline void set_value (const vptr_tag &, const void *x)
  {
    *((const void **)mp_write) = x;
    mp_write += item_size<const void *> ();
  }

  inline void set_value (const ccptr_tag &, const char *x)
  {
    *((const char **)mp_write) = x;
    mp_write += item_size<const char *> ();
  }

  inline void set_value (const ccptr_tag &, const signed char *x)
  {
    *((const signed char **)mp_write) = x;
    mp_write += item_size<const signed char *> ();
  }

  inline void set_value (const ccptr_tag &, const unsigned char *x)
  {
    *((const unsigned char **)mp_write) = x;
    mp_write += item_size<const unsigned char *> ();
  }

  inline const char *get_value (const ccptr_tag &)
  {
    check_data ();
    const char *r = *((const char **)mp_read);
    mp_read += item_size<const char *> ();
    return r;
  }

  inline void *get_value (const vptr_tag &)
  {
    check_data ();
    void *r = *((void **)mp_read);
    mp_read += item_size<void *> ();
    return r;
  }

  template <class X>
  inline X get_value (const pod_direct_tag &)
  {
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  inline X get_value (const npod_direct_tag &)
  {
    check_data ();
    X x = *(X *)mp_read;
    ((X *)mp_read)->~X ();
    mp_read += item_size<X> ();
    return x;
  }

  template <class X> 
  inline X get_value (const x_tag &)
  {
    check_data ();
    X x = *(X *)mp_read;
    ((X *)mp_read)->~X ();
    mp_read += item_size<X> ();
    return x;
  }

  template <class X> 
  inline X get_value (const vptr_tag &)
  {
    void *r = *((void **)mp_read);
    mp_read += item_size<void *> ();
    // Hint: X can be const unsigned char *, const signed char * as well.
    return (X)r;
  }

  template <class X> 
  inline X get_value (const ccptr_tag &)
  {
    const char *r = *((const char **)mp_read);
    mp_read += item_size<const char *> ();
    // Hint: X can be const unsigned char *, const signed char * as well.
    return (X)r;
  }

  template <class X> 
  inline X get_value (const ref_tag &)
  {
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    value_type *r = *((value_type **)mp_read);
    mp_read += item_size<value_type *> ();
    if (! r) {
      throw NilPointerToReference ();
    }
    return *r;
  }

  template <class X>
  inline X get_value (const pod_cref_tag &)
  {
    //  X is actually an (const X &)
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    const value_type *r = ((const value_type *)mp_read);
    mp_read += item_size<value_type> ();
    return *r;
  }

  template <class X> 
  inline X get_value (const npod_cref_tag &)
  {
    //  X is actually an (const X &)
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    const value_type *r = *((const value_type **)mp_read);
    mp_read += item_size<const value_type *> ();
    if (! r) {
      throw NilPointerToReference ();
    }
    return *r;
  }

  template <class X> 
  inline X get_value (const x_cref_tag &)
  {
    //  X is actually an (const X &)
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    const value_type *r = *((const value_type **)mp_read);
    mp_read += item_size<const value_type *> ();
    if (! r) {
      throw NilPointerToReference ();
    }
    return *r;
  }

  template <class X> 
  inline X get_value (const ptr_tag &)
  {
    //  X is actually an (X *)
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X>
  inline X get_value (const pod_cptr_tag &)
  {
    //  X is actually an (const X *)
    check_data ();
    bool h = *(bool *)mp_read;
    mp_read += item_size<bool> ();
    X r = h ? (X)mp_read : (X)0;
    mp_read += item_size<X> ();
    return r;
  }

  //  see notes on the serialization for this type:
  template <class X> 
  inline X get_value (const npod_cptr_tag &)
  {
    //  X is actually an (const X *)
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  inline X get_value (const x_cptr_tag &)
  {
    //  X is actually an (const X *)
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  inline X *take_object ()
  {
    check_data ();
    X *x = new X (*((X *)mp_read));
    ((X *)mp_read)->~X ();
    mp_read += item_size<X> ();
    return x;
  }

  void check_data () const
  {
    if (! (mp_read && mp_read < mp_write)) {
      throw ArglistUnderflowException ();
    }
  }

private:
  char *mp_buffer;
  char *mp_read, *mp_write;
  std::list<std::string> m_strbufs;
};

}

#endif

