
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiTypes
#define _HDR_gsiTypes

#include <string>
#include <vector>
#include <stdexcept>

#include "tlInternational.h"
#include "tlException.h"

#include <QtCore/QString>

//  For a comprehensive documentation see gsi.h

namespace tl
{
  class Variant;
}

namespace gsi
{

// ---------------------------------------------------------------------------------
//  Type system of GSI

class ClassBase;
template <class X> class Class;
template <class X> struct ClassTag;
template <class I, class R, class V> class IterAdaptor;
template <class R, class V> class IterPtrAdaptor;
template <class R, class V> class ConstIterPtrAdaptor;
template <class I, class R, class V> class FreeIterAdaptor;
template <class X> const ClassBase *cls_decl ();

/**
 *  @brief Determine the size for one item in the serialization buffer
 */
template<class X>
unsigned int item_size () 
{
  return (sizeof (X) + (sizeof (void *) - 1)) & ~(sizeof (void *) - 1); 
}

template<class X>
struct type_tag
{
  // .. intentionally left blank ..
};

/**
 *   @brief Basic GSI type constants
 */
enum BasicType
{
  T_void = 0,
  T_bool = 1,
  T_char = 2,
  T_schar = 3,
  T_uchar = 4,
  T_short = 5,
  T_ushort = 6,
  T_int = 7,
  T_uint = 8,
  T_long = 9,
  T_ulong = 10,
  T_longlong = 11,
  T_ulonglong = 12,
#if defined(HAVE_64BIT_COORD)
  T_int128 = 100,
#endif
  T_double = 13,
  T_float = 14,
  T_var = 15,
  T_string = 16,
  T_qstring = 17,
  T_string_ccptr = 18,
  T_void_ptr = 19,
  T_object = 20,
  T_object_new = 21,
  T_vector = 22
};

/**
 *  @brief Type tags
 *
 *  "pod" are the POD types (bool, char, short, int, long, long long, double, float, unsigned variants)
 *  "npod" types are strings, vectors and variants
 *  "obj" types are objects
 *
 *  The basic tags are:
 *    - pod_direct_tag:     (POD)           serialized as (POD)
 *    - pod_cref_tag:       (const POD &)   serialized as (POD)
 *    - pod_ref_tag:        (POD &)         serialized as (POD *) with pointer != 0
 *    - pod_cptr_tag:       (const POD *)   serialized as (true, POD) or (false)
 *    - pod_ptr_tag:        (POD *)         serialized as (POD *)
 *    - npod_direct_tag:    (NPOD)          serialized as (NPOD)
 *    - npod_cref_tag:      (const NPOD &)  serialized as (NPOD *)
 *                                          (Reason for the latter decision: deserializing is that 
 *                                          way possible without having to call the destructor after
 *                                          the value has been used)
 *    - npod_ref_tag:       (NPOD &)        serialized as (NPOD *) with pointer != 0
 *    - npod_cptr_tag:      (const NPOD *)  serialized as (NPOD *)
 *                                          (Reason for the latter decision: deserializing is that 
 *                                          way possible without having to call the destructor after
 *                                          the value has been used)
 *    - npod_ptr_tag:       (NPOD *)        serialized as (NPOD *)
 *    - x_tag:              (OBJ)           serialized as (OBJ)
 *    - x_cref_tag:         (const OBJ &)   serialized as (OBJ *) with pointer != 0
 *    - x_ref_tag:          (OBJ &)         serialized as (OBJ *) with pointer != 0, object must not be const
 *    - x_cptr_tag:         (const OBJ *)   serialized as (OBJ *)
 *    - x_ptr_tag:          (OBJ *)         serialized as (OBJ *), object must not be const
 */
struct basic_type_tag { };
struct void_tag { };
struct ccptr_tag { };
struct vptr_tag { };

struct direct_tag           : public basic_type_tag { };
struct npod_direct_tag      : public direct_tag { };
struct pod_direct_tag       : public direct_tag { };
struct cref_tag             : public basic_type_tag { };
struct pod_cref_tag         : public cref_tag { };
struct npod_cref_tag        : public cref_tag { };
struct ref_tag              : public basic_type_tag { };
struct pod_ref_tag          : public ref_tag { };
struct npod_ref_tag         : public ref_tag { };
struct cptr_tag             : public basic_type_tag { };
struct pod_cptr_tag         : public cptr_tag { };
struct npod_cptr_tag        : public cptr_tag { };
struct ptr_tag              : public basic_type_tag { };
struct pod_ptr_tag          : public ptr_tag { };
struct npod_ptr_tag         : public ptr_tag { };

struct bool_tag             : public pod_direct_tag { };
struct char_tag             : public pod_direct_tag { };
struct uchar_tag            : public pod_direct_tag { };
struct schar_tag            : public pod_direct_tag { };
struct short_tag            : public pod_direct_tag { };
struct ushort_tag           : public pod_direct_tag { };
struct int_tag              : public pod_direct_tag { };
struct uint_tag             : public pod_direct_tag { };
struct long_tag             : public pod_direct_tag { };
struct ulong_tag            : public pod_direct_tag { };
struct longlong_tag         : public pod_direct_tag { };
struct ulonglong_tag        : public pod_direct_tag { };
#if defined(HAVE_64BIT_COORD)
struct int128_tag           : public pod_direct_tag { };
#endif
struct double_tag           : public pod_direct_tag { };
struct float_tag            : public pod_direct_tag { };

struct vector_tag           : public npod_direct_tag { }; 
struct string_tag           : public npod_direct_tag { };
struct qstring_tag          : public npod_direct_tag { };
struct var_tag              : public npod_direct_tag { };

struct bool_cref_tag        : public pod_cref_tag { };
struct char_cref_tag        : public pod_cref_tag { };
struct uchar_cref_tag       : public pod_cref_tag { };
struct schar_cref_tag       : public pod_cref_tag { };
struct short_cref_tag       : public pod_cref_tag { };
struct ushort_cref_tag      : public pod_cref_tag { };
struct int_cref_tag         : public pod_cref_tag { };
struct uint_cref_tag        : public pod_cref_tag { };
struct long_cref_tag        : public pod_cref_tag { };
struct ulong_cref_tag       : public pod_cref_tag { };
struct longlong_cref_tag    : public pod_cref_tag { };
struct ulonglong_cref_tag   : public pod_cref_tag { };
#if defined(HAVE_64BIT_COORD)
struct int128_cref_tag      : public pod_cref_tag { };
#endif
struct double_cref_tag      : public pod_cref_tag { };
struct float_cref_tag       : public pod_cref_tag { };

struct vector_cref_tag      : public npod_cref_tag { };
struct string_cref_tag      : public npod_cref_tag { };
struct qstring_cref_tag     : public npod_cref_tag { };
struct var_cref_tag         : public npod_cref_tag { };

struct bool_ref_tag         : public pod_ref_tag { };
struct char_ref_tag         : public pod_ref_tag { };
struct uchar_ref_tag        : public pod_ref_tag { };
struct schar_ref_tag        : public pod_ref_tag { };
struct short_ref_tag        : public pod_ref_tag { };
struct ushort_ref_tag       : public pod_ref_tag { };
struct int_ref_tag          : public pod_ref_tag { };
struct uint_ref_tag         : public pod_ref_tag { };
struct long_ref_tag         : public pod_ref_tag { };
struct ulong_ref_tag        : public pod_ref_tag { };
struct longlong_ref_tag     : public pod_ref_tag { };
struct ulonglong_ref_tag    : public pod_ref_tag { };
#if defined(HAVE_64BIT_COORD)
struct int128_ref_tag       : public pod_ref_tag { };
#endif
struct double_ref_tag       : public pod_ref_tag { };
struct float_ref_tag        : public pod_ref_tag { };

struct vector_ref_tag       : public npod_ref_tag { };
struct string_ref_tag       : public npod_ref_tag { };
struct qstring_ref_tag      : public npod_ref_tag { };
struct var_ref_tag          : public npod_ref_tag { };

struct bool_cptr_tag        : public pod_cptr_tag { };
struct char_cptr_tag        : public pod_cptr_tag { };
struct uchar_cptr_tag       : public pod_cptr_tag { };
struct schar_cptr_tag       : public pod_cptr_tag { };
struct short_cptr_tag       : public pod_cptr_tag { };
struct ushort_cptr_tag      : public pod_cptr_tag { };
struct int_cptr_tag         : public pod_cptr_tag { };
struct uint_cptr_tag        : public pod_cptr_tag { };
struct long_cptr_tag        : public pod_cptr_tag { };
struct ulong_cptr_tag       : public pod_cptr_tag { };
struct longlong_cptr_tag    : public pod_cptr_tag { };
struct ulonglong_cptr_tag   : public pod_cptr_tag { };
#if defined(HAVE_64BIT_COORD)
struct int128_cptr_tag      : public pod_cptr_tag { };
#endif
struct double_cptr_tag      : public pod_cptr_tag { };
struct float_cptr_tag       : public pod_cptr_tag { };

struct vector_cptr_tag      : public npod_cptr_tag { };
struct string_cptr_tag      : public npod_cptr_tag { };
struct qstring_cptr_tag     : public npod_cptr_tag { };
struct var_cptr_tag         : public npod_cptr_tag { };

struct bool_ptr_tag         : public pod_ptr_tag { };
struct char_ptr_tag         : public pod_ptr_tag { };
struct uchar_ptr_tag        : public pod_ptr_tag { };
struct schar_ptr_tag        : public pod_ptr_tag { };
struct short_ptr_tag        : public pod_ptr_tag { };
struct ushort_ptr_tag       : public pod_ptr_tag { };
struct int_ptr_tag          : public pod_ptr_tag { };
struct uint_ptr_tag         : public pod_ptr_tag { };
struct long_ptr_tag         : public pod_ptr_tag { };
struct ulong_ptr_tag        : public pod_ptr_tag { };
struct longlong_ptr_tag     : public pod_ptr_tag { };
struct ulonglong_ptr_tag    : public pod_ptr_tag { };
#if defined(HAVE_64BIT_COORD)
struct int128_ptr_tag       : public pod_ptr_tag { };
#endif
struct double_ptr_tag       : public pod_ptr_tag { };
struct float_ptr_tag        : public pod_ptr_tag { };

struct vector_ptr_tag       : public npod_ptr_tag { };
struct string_ptr_tag       : public npod_ptr_tag { };
struct qstring_ptr_tag      : public npod_ptr_tag { };
struct var_ptr_tag          : public npod_ptr_tag { };

struct x_tag                : public npod_direct_tag { };
struct x_cref_tag           : public npod_cref_tag { };
struct x_cptr_tag           : public npod_cptr_tag { };
struct x_cptr_cref_tag      : public x_cptr_tag { };
struct x_cptr_ref_tag       : public x_cptr_tag { };
struct x_ref_tag            : public npod_ref_tag { };
struct x_ptr_tag            : public npod_ptr_tag { };
struct x_ptr_cref_tag       : public x_ptr_tag { };
struct x_ptr_ref_tag        : public x_ptr_tag { };


/**
 *  @brief Type traits that map a C++ type to a certain tag
 */
template <class X> struct type_traits;

template <>
struct type_traits<void>
{
  typedef void_tag tag;
  typedef void value_type;
  typedef void inner_type;
  static BasicType code () { return T_void; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <class T, class V, enum BasicType TC> 
struct plain_type_traits 
{
  typedef T tag;
  typedef V value_type;
  typedef void inner_type;
  static BasicType code () { return TC; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <> struct type_traits<bool>                        : plain_type_traits<bool_tag, bool, T_bool> { };
template <> struct type_traits<char>                        : plain_type_traits<char_tag, char, T_char> { };
template <> struct type_traits<signed char>                 : plain_type_traits<schar_tag, signed char, T_schar> { };
template <> struct type_traits<unsigned char>               : plain_type_traits<uchar_tag, unsigned char, T_uchar> { };
template <> struct type_traits<short>                       : plain_type_traits<short_tag, short, T_short> { };
template <> struct type_traits<unsigned short>              : plain_type_traits<ushort_tag, unsigned short, T_ushort> { };
template <> struct type_traits<int>                         : plain_type_traits<int_tag, int, T_int> { };
template <> struct type_traits<unsigned int>                : plain_type_traits<uint_tag, unsigned int, T_uint> { };
template <> struct type_traits<long>                        : plain_type_traits<long_tag, long, T_long> { };
template <> struct type_traits<unsigned long>               : plain_type_traits<ulong_tag, unsigned long, T_ulong> { };
template <> struct type_traits<long long>                   : plain_type_traits<longlong_tag, long long, T_longlong> { };
template <> struct type_traits<unsigned long long>          : plain_type_traits<ulonglong_tag, unsigned long long, T_ulonglong> { };
#if defined(HAVE_64BIT_COORD)
template <> struct type_traits<__int128>                    : plain_type_traits<int128_tag, __int128, T_int128> { };
#endif
template <> struct type_traits<double>                      : plain_type_traits<double_tag, double, T_double> { };
template <> struct type_traits<float>                       : plain_type_traits<float_tag, float, T_float> { };
template <> struct type_traits<std::string>                 : plain_type_traits<string_tag, std::string, T_string> { };
template <> struct type_traits<QString>                     : plain_type_traits<qstring_tag, QString, T_qstring> { };
template <> struct type_traits<tl::Variant>                 : plain_type_traits<var_tag, tl::Variant, T_var> { };

template <>
struct type_traits<void *>
{
  typedef vptr_tag tag;
  typedef void *value_type;
  typedef void inner_type;
  static BasicType code () { return T_void_ptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const void *>
{
  typedef vptr_tag tag;
  typedef void *value_type;
  typedef void inner_type;
  static BasicType code () { return T_void_ptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const char *>
{
  typedef ccptr_tag tag;
  typedef const char *value_type;
  typedef void inner_type;
  static BasicType code () { return T_string_ccptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const unsigned char *>
{
  typedef ccptr_tag tag;
  typedef const char *value_type;
  typedef void inner_type;
  static BasicType code () { return T_string_ccptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const signed char *>
{
  typedef ccptr_tag tag;
  typedef const char *value_type;
  typedef void inner_type;
  static BasicType code () { return T_string_ccptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <> struct type_traits<const bool &>                : plain_type_traits<bool_cref_tag, bool, T_bool> { };
template <> struct type_traits<const char &>                : plain_type_traits<char_cref_tag, char, T_char> { };
template <> struct type_traits<const signed char &>         : plain_type_traits<schar_cref_tag, signed char, T_schar> { };
template <> struct type_traits<const unsigned char &>       : plain_type_traits<uchar_cref_tag, unsigned char, T_uchar> { };
template <> struct type_traits<const short &>               : plain_type_traits<short_cref_tag, short, T_short> { };
template <> struct type_traits<const unsigned short &>      : plain_type_traits<ushort_cref_tag, unsigned short, T_ushort> { };
template <> struct type_traits<const int &>                 : plain_type_traits<int_cref_tag, int, T_int> { };
template <> struct type_traits<const unsigned int &>        : plain_type_traits<uint_cref_tag, unsigned int, T_uint> { };
template <> struct type_traits<const long &>                : plain_type_traits<long_cref_tag, long, T_long> { };
template <> struct type_traits<const unsigned long &>       : plain_type_traits<ulong_cref_tag, unsigned long, T_ulong> { };
template <> struct type_traits<const long long &>           : plain_type_traits<longlong_cref_tag, long long, T_longlong> { };
template <> struct type_traits<const unsigned long long &>  : plain_type_traits<ulonglong_cref_tag, unsigned long long, T_ulonglong> { };
#if defined(HAVE_64BIT_COORD)
template <> struct type_traits<const __int128 &>            : plain_type_traits<int128_cref_tag, __int128, T_int128> { };
#endif
template <> struct type_traits<const double &>              : plain_type_traits<double_cref_tag, double, T_double> { };
template <> struct type_traits<const float &>               : plain_type_traits<float_cref_tag, float, T_float> { };
template <> struct type_traits<const std::string &>         : plain_type_traits<string_cref_tag, std::string, T_string> { };
template <> struct type_traits<const QString &>             : plain_type_traits<qstring_cref_tag, QString, T_qstring> { };
template <> struct type_traits<const tl::Variant &>         : plain_type_traits<var_cref_tag, tl::Variant, T_var> { };

template <> struct type_traits<bool &>                      : plain_type_traits<bool_ref_tag, bool, T_bool> { };
template <> struct type_traits<char &>                      : plain_type_traits<char_ref_tag, char, T_char> { };
template <> struct type_traits<signed char &>               : plain_type_traits<schar_ref_tag, signed char, T_schar> { };
template <> struct type_traits<unsigned char &>             : plain_type_traits<uchar_ref_tag, unsigned char, T_uchar> { };
template <> struct type_traits<short &>                     : plain_type_traits<short_ref_tag, short, T_short> { };
template <> struct type_traits<unsigned short &>            : plain_type_traits<ushort_ref_tag, unsigned short, T_ushort> { };
template <> struct type_traits<int &>                       : plain_type_traits<int_ref_tag, int, T_int> { };
template <> struct type_traits<unsigned int &>              : plain_type_traits<uint_ref_tag, unsigned int, T_uint> { };
template <> struct type_traits<long &>                      : plain_type_traits<long_ref_tag, long, T_long> { };
template <> struct type_traits<unsigned long &>             : plain_type_traits<ulong_ref_tag, unsigned long, T_ulong> { };
template <> struct type_traits<long long &>                 : plain_type_traits<longlong_ref_tag, long long, T_longlong> { };
template <> struct type_traits<unsigned long long &>        : plain_type_traits<ulonglong_ref_tag, unsigned long long, T_ulonglong> { };
#if defined(HAVE_64BIT_COORD)
template <> struct type_traits<__int128 &>                  : plain_type_traits<int128_ref_tag, __int128, T_int128> { };
#endif
template <> struct type_traits<double &>                    : plain_type_traits<double_ref_tag, double, T_double> { };
template <> struct type_traits<float &>                     : plain_type_traits<float_ref_tag, float, T_float> { };
template <> struct type_traits<std::string &>               : plain_type_traits<string_ref_tag, std::string, T_string> { };
template <> struct type_traits<QString &>                   : plain_type_traits<qstring_ref_tag, QString, T_qstring> { };
template <> struct type_traits<tl::Variant &>               : plain_type_traits<var_ref_tag, tl::Variant, T_var> { };

template <> struct type_traits<const bool *>                : plain_type_traits<bool_cptr_tag, bool, T_bool> { };
// that is bound to a string, not a single char
// template <> struct type_traits<const char *>             : plain_type_traits<char_cptr_tag, char, T_char> { };
// template <> struct type_traits<const unsigned char *>       : plain_type_traits<uchar_cptr_tag, unsigned char, T_uchar> { };
// template <> struct type_traits<const signed char *>         : plain_type_traits<schar_cptr_tag, signed char, T_schar> { };
template <> struct type_traits<const short *>               : plain_type_traits<short_cptr_tag, short, T_short> { };
template <> struct type_traits<const unsigned short *>      : plain_type_traits<ushort_cptr_tag, unsigned short, T_ushort> { };
template <> struct type_traits<const int *>                 : plain_type_traits<int_cptr_tag, int, T_int> { };
template <> struct type_traits<const unsigned int *>        : plain_type_traits<uint_cptr_tag, unsigned int, T_uint> { };
template <> struct type_traits<const long *>                : plain_type_traits<long_cptr_tag, long, T_long> { };
template <> struct type_traits<const unsigned long *>       : plain_type_traits<ulong_cptr_tag, unsigned long, T_ulong> { };
template <> struct type_traits<const long long *>           : plain_type_traits<longlong_cptr_tag, long long, T_longlong> { };
template <> struct type_traits<const unsigned long long *>  : plain_type_traits<ulonglong_cptr_tag, unsigned long long, T_ulonglong> { };
#if defined(HAVE_64BIT_COORD)
template <> struct type_traits<const __int128 *>            : plain_type_traits<int128_cptr_tag, __int128, T_int128> { };
#endif
template <> struct type_traits<const double *>              : plain_type_traits<double_cptr_tag, double, T_double> { };
template <> struct type_traits<const float *>               : plain_type_traits<float_cptr_tag, float, T_float> { };
template <> struct type_traits<const std::string *>         : plain_type_traits<string_cptr_tag, std::string, T_string> { };
template <> struct type_traits<const QString *>             : plain_type_traits<qstring_cptr_tag, QString, T_qstring> { };
template <> struct type_traits<const tl::Variant *>         : plain_type_traits<var_cptr_tag, tl::Variant, T_var> { };

template <> struct type_traits<bool *>                      : plain_type_traits<bool_ptr_tag, bool, T_bool> { };
// that is not bound to avoid interpretation problems between string and char *
// template <> struct type_traits<char *>                   : plain_type_traits<char_ptr_tag, char, T_char> { };
// template <> struct type_traits<signed char *>               : plain_type_traits<schar_ptr_tag, signed char, T_schar> { };
// template <> struct type_traits<unsigned char *>             : plain_type_traits<uchar_ptr_tag, unsigned char, T_uchar> { };
template <> struct type_traits<short *>                     : plain_type_traits<short_ptr_tag, short, T_short> { };
template <> struct type_traits<unsigned short *>            : plain_type_traits<ushort_ptr_tag, unsigned short, T_ushort> { };
template <> struct type_traits<int *>                       : plain_type_traits<int_ptr_tag, int, T_int> { };
template <> struct type_traits<unsigned int *>              : plain_type_traits<uint_ptr_tag, unsigned int, T_uint> { };
template <> struct type_traits<long *>                      : plain_type_traits<long_ptr_tag, long, T_long> { };
template <> struct type_traits<unsigned long *>             : plain_type_traits<ulong_ptr_tag, unsigned long, T_ulong> { };
template <> struct type_traits<long long *>                 : plain_type_traits<longlong_ptr_tag, long long, T_longlong> { };
template <> struct type_traits<unsigned long long *>        : plain_type_traits<ulonglong_ptr_tag, unsigned long long, T_ulonglong> { };
#if defined(HAVE_64BIT_COORD)
template <> struct type_traits<__int128 *>                  : plain_type_traits<int128_ptr_tag, __int128, T_int128> { };
#endif
template <> struct type_traits<double *>                    : plain_type_traits<double_ptr_tag, double, T_double> { };
template <> struct type_traits<float *>                     : plain_type_traits<float_ptr_tag, float, T_float> { };
template <> struct type_traits<std::string *>               : plain_type_traits<string_ptr_tag, std::string, T_string> { };
template <> struct type_traits<QString *>                   : plain_type_traits<qstring_ptr_tag, QString, T_qstring> { };
template <> struct type_traits<tl::Variant *>               : plain_type_traits<var_ptr_tag, tl::Variant, T_var> { };

template <class X>
struct type_traits<const X>
{
  typedef typename type_traits<X>::tag tag;
  typedef typename type_traits<X>::value_type value_type;
  typedef void inner_type;
  static BasicType code () { return type_traits<X>::code (); }
  static const ClassBase *cls_decl () { return type_traits<X>::cls_decl (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits
{
  typedef x_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X &>
{
  typedef x_cref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X *>
{
  typedef x_cptr_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X * const &>
{
  typedef x_cptr_cref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X * &>
{
  typedef x_cptr_ref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X &>
{
  typedef x_ref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X *>
{
  typedef x_ptr_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X * const &>
{
  typedef x_ptr_cref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X * &>
{
  typedef x_ptr_ref_tag tag;
  typedef X value_type;
  typedef void inner_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return gsi::cls_decl<X> (); }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< std::vector<X> >
{
  typedef vector_tag tag;
  typedef std::vector<X> value_type;
  typedef X inner_type;
  static BasicType code () { return T_vector; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< const std::vector<X> & >
{
  typedef vector_cref_tag tag;
  typedef std::vector<X> value_type;
  typedef X inner_type;
  static BasicType code () { return T_vector; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< std::vector<X> & >
{
  typedef vector_ref_tag tag;
  typedef std::vector<X> value_type;
  typedef X inner_type;
  static BasicType code () { return T_vector; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_iter () { return false; }
};

template <class I, class R, class V>
struct type_traits< IterAdaptor<I, R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  typedef void inner_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_iter () { return true; }
};

template <class R, class V>
struct type_traits< IterPtrAdaptor<R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  typedef void inner_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_iter () { return true; }
};

template <class R, class V>
struct type_traits< ConstIterPtrAdaptor<R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  typedef void inner_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_iter () { return true; }
};

template <class I, class R, class V>
struct type_traits< FreeIterAdaptor<I, R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  typedef void inner_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_iter () { return true; }
};

inline bool ref_predicate (const void_tag &)            { return false; }
inline bool ref_predicate (const ccptr_tag &)           { return false; }
inline bool ref_predicate (const vptr_tag &)            { return false; }
inline bool ref_predicate (const basic_type_tag &)      { return false; }
inline bool ref_predicate (const ref_tag &)             { return true; }

inline bool cref_predicate (const void_tag &)           { return false; }
inline bool cref_predicate (const ccptr_tag &)          { return false; }
inline bool cref_predicate (const vptr_tag &)           { return false; }
inline bool cref_predicate (const basic_type_tag &)     { return false; }
inline bool cref_predicate (const cref_tag &)           { return true; }

inline bool ptr_predicate (const void_tag &)            { return false; }
inline bool ptr_predicate (const ccptr_tag &)           { return false; }
inline bool ptr_predicate (const vptr_tag &)            { return false; }
inline bool ptr_predicate (const basic_type_tag &)      { return false; }
inline bool ptr_predicate (const ptr_tag &)             { return true; }

inline bool cptr_predicate (const void_tag &)           { return false; }
inline bool cptr_predicate (const ccptr_tag &)          { return false; }
inline bool cptr_predicate (const vptr_tag &)           { return false; }
inline bool cptr_predicate (const basic_type_tag &)     { return false; }
inline bool cptr_predicate (const cptr_tag &)           { return true; }

//  see serialization notes above
template <class X> inline size_t size_attribute (const void_tag &)         { return 0; }
template <class X> inline size_t size_attribute (const basic_type_tag &)   { return item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const pod_direct_tag &)   { return item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const pod_cref_tag &)     { return item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const pod_ref_tag &)      { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const pod_cptr_tag &)     { return item_size<bool> () + item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const pod_ptr_tag &)      { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const ccptr_tag &)        { return item_size<const char *> (); }
template <class X> inline size_t size_attribute (const vptr_tag &)         { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const npod_direct_tag &)  { return item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const npod_cref_tag &)    { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const npod_ref_tag &)     { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const npod_cptr_tag &)    { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const npod_ptr_tag &)     { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const x_tag &)            { return item_size<typename type_traits<X>::value_type> (); }
template <class X> inline size_t size_attribute (const x_cref_tag &)       { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const x_ref_tag &)        { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const x_cptr_tag &)       { return item_size<void *> (); }
template <class X> inline size_t size_attribute (const x_ptr_tag &)        { return item_size<void *> (); }

/**
 *  @brief Generic argument type declaration
 *
 *  The type declaration carries the basic type code and flags, if the argument 
 *  is an iterator delivering the given types of a vector of the given types and
 *  adds class informations if the type is "T_object", "T_object_ref", "T_object_cref" or "T_object_new".
 */
struct KLAYOUT_DLL ArgType
{
  ArgType ();
  ~ArgType ();
  ArgType (const ArgType &other);

  ArgType &operator= (const ArgType &other);

  template <class X>
  void init () 
  { 
    type        = type_traits<X>::code ();
    is_iter     = type_traits<X>::is_iter ();
    cls         = type_traits<X>::cls_decl ();

    is_ref      = ref_predicate (typename type_traits<X>::tag ());
    is_ptr      = ptr_predicate (typename type_traits<X>::tag ());
    is_cref     = cref_predicate (typename type_traits<X>::tag ());
    is_cptr     = cptr_predicate (typename type_traits<X>::tag ());
    size        = (unsigned int) size_attribute<X> (typename type_traits<X>::tag ());

    if (inner) {
      delete inner;
      inner = 0;
    }

    if (type_traits<typename type_traits<X>::inner_type>::code () != T_void) {
      inner = new ArgType;
      inner->init<typename type_traits<X>::inner_type> ();
    }
  }

  template <class X>
  void init_new_object () 
  { 
    type        = T_object_new;
    is_iter     = type_traits<X>::is_iter ();
    cls         = type_traits<X>::cls_decl ();

    is_ref      = ref_predicate (typename type_traits<X>::tag ());
    is_ptr      = ptr_predicate (typename type_traits<X>::tag ());
    is_cref     = cref_predicate (typename type_traits<X>::tag ());
    is_cptr     = cptr_predicate (typename type_traits<X>::tag ());
    size        = (unsigned int) size_attribute<X> (typename type_traits<X>::tag ());

    if (inner) {
      delete inner;
      inner = 0;
    }

    if (type_traits<typename type_traits<X>::inner_type>::code () != T_void) {
      inner = new ArgType;
      inner->init<typename type_traits<X>::inner_type> ();
    }
  }

  void check_type (const ArgType &a) const;

  bool operator== (const ArgType &b) const
  {
    if ((inner == 0) != (b.inner == 0)) {
      return false;
    }
    if (inner && *inner != *b.inner) {
      return false;
    }
    return type == b.type && is_iter == b.is_iter && 
           is_ref == b.is_ref && is_cref == b.is_cref && is_ptr == b.is_ptr && is_cptr == b.is_cptr && 
           cls == b.cls;
  }

  bool operator!= (const ArgType &b) const
  {
    return !operator== (b);
  }

  std::string to_string () const;

  BasicType type;
  ArgType *inner;
  bool is_ref : 1;
  bool is_ptr : 1;
  bool is_cref : 1;
  bool is_cptr : 1;
  bool is_iter : 1;
  mutable const ClassBase *cls;
  unsigned int size;
};

struct KLAYOUT_DLL IncompatibleTypeException
  : public tl::Exception
{
  IncompatibleTypeException (const ArgType &got, const ArgType &want)
    : tl::Exception (tl::translate ("Incompatible types: got '") + got.to_string () + tl::translate ("', want '") + want.to_string () + "'")
  { }
};

inline void ArgType::check_type (const ArgType &a) const
{
  if (*this != a) {
    throw IncompatibleTypeException (a, *this);
  }
}

/**
 *  @brief A helper class that provides a base class for wrapped pointers for TempObject
 */
class TempObjectBase 
{
public:
  virtual ~TempObjectBase () { }
};

/**
 *  @brief A helper class that provides a wrapper to a custom type for TempObject
 */
template <class X>
class TempObjectCont : public TempObjectBase 
{
public:
  TempObjectCont (X *x) : mp_x (x) { }
  ~TempObjectCont () { delete mp_x; }
private:
  X *mp_x;
};

/**
 *  @brief A simple autopointer class to hold references to temporary objects
 */
class TempObject
{
public:
  TempObject () : mp_b (0) { }

  template <class X>
  TempObject (X *x) : mp_b (0) 
  { 
    set (x);  
  }

  ~TempObject () {
    if (mp_b) {
      delete mp_b;
    }
    mp_b = 0;
  }

  template <class X>
  void set (X *x) { 
    tl_assert (mp_b == 0);
    mp_b = new TempObjectCont<X> (x); 
  }

private:
  TempObjectBase *mp_b;
};

// ---------------------------------------------------------------------------------
//  Utility: determine non-case "base" of X (nc_x(const X) == X)

template <class X>
struct non_const_x {
  typedef X nc_x;
};

template <class X>
struct non_const_x<const X> {
  typedef X nc_x;
};

// ---------------------------------------------------------------------------------
//  Utility: determine, if X is const 

template <class X>
struct is_const_x {
  static bool value () { return false; }
};

template <class X>
struct is_const_x<const X> {
  static bool value () { return true; }
};

// -------------------------------------------------------------------
//  A utility to map a type enum to a type
//
//  Is must be used this way:
//
//  struct func {
//    template <class X> void op () { do_it_with_X; }
//  };
//
//  if (! switch_type(type, func())) {
//    do_it_with_other_types(type);
//  }

template <class F>
bool switch_type (gsi::BasicType type, F &f)
{
  switch (type) {
  default:
    return false;
  case gsi::T_bool:     
    f.template op<bool> ();
    return true;
  case gsi::T_char:      
    f.template op<char> ();
    return true;
  case gsi::T_schar:     
    f.template op<signed char> ();
    return true;
  case gsi::T_uchar:     
    f.template op<unsigned char> ();
    return true;
  case gsi::T_short:      
    f.template op<short> ();
    return true;
  case gsi::T_ushort:     
    f.template op<unsigned short> ();
    return true;
  case gsi::T_int:      
    f.template op<int> ();
    return true;
  case gsi::T_uint:     
    f.template op<unsigned int> ();
    return true;
  case gsi::T_long:     
    f.template op<long> ();
    return true;
  case gsi::T_ulong:    
    f.template op<unsigned long> ();
    return true;
  case gsi::T_longlong:     
    f.template op<long long> ();
    return true;
  case gsi::T_ulonglong:     
    f.template op<unsigned long long> ();
    return true;
#if defined(HAVE_64BIT_COORD)
  case gsi::T_int128:     
    f.template op<__int128> ();
    return true;
#endif
  case gsi::T_double:   
    f.template op<double> ();
    return true;
  case gsi::T_float:   
    f.template op<float> ();
    return true;
  case gsi::T_void_ptr:   
    f.template op<void *> ();
    return true;
  case gsi::T_string_ccptr:   
    f.template op<const char *> ();
    return true;
  case gsi::T_string:   
    f.template op<std::string> ();
    return true;
  case gsi::T_qstring:   
    f.template op<QString> ();
    return true;
  case gsi::T_var:   
    f.template op<tl::Variant> ();
    return true;
  }
}

}

#endif

