
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layGSIHelpProvider.h"
#include "layHelpSource.h"
#include "tlClassRegistry.h"

#include "gsiDecl.h"
#include "gsiDeclBasic.h"

#include <QtCore/QUrl> 
#include <QtGui/QColor>
#include <QtGui/QPalette>
#include <QtGui/QApplication>

#include <cstdio>
#include <cctype>
#include <algorithm>

namespace lay
{

// --------------------------------------------------------------------------------------
//  Some utilities

static std::string class_doc_path (const std::string &c)
{
  std::string cn;
  for (const char *cp = c.c_str (); *cp; ++cp) {
    if (*cp == ':' && cp[1] == ':') {
      cn += "_";
      ++cp;
    } else {
      cn += *cp;
    }
  }
  return "/code/class_" + cn + ".xml";
}

static std::string class_doc_path (const std::string &c, const std::string &m)
{
  if (c.empty ()) {
    return "#m_" + m;
  } else {
    std::string cn;
    for (const char *cp = c.c_str (); *cp; ++cp) {
      if (*cp == ':' && cp[1] == ':') {
        cn += "_";
        ++cp;
      } else {
        cn += *cp;
      }
    }
    return "/code/class_" + cn + ".xml#m_" + m;
  }
}

std::string escape_xml_with_formatting (const std::string &s, bool &in_code)
{
  std::string r;
  r.reserve (s.size ());
  for (tl::Extractor sc (s.c_str ()); *sc; ) {
    if (*sc == '<') {
      r += "&lt;";
      ++sc;
    } else if (*sc == '>') {
      r += "&gt;";
      ++sc;
    } else if (*sc == '&') {
      r += "&amp;";
      ++sc;
    } else if (*sc == '@') {
      if (sc.test ("@<")) {

        //  HTML tag
        r += "<";
        while (*sc && *sc != '>') {
          r += *sc;
          ++sc;
        }
        if (*sc == '>') {
          r += ">";
          ++sc;
        }

      } else if (sc.test ("@li")) {
        r += "<li>";
      } else if (sc.test ("@/li")) {
        r += "</li>";
      } else if (sc.test ("@ul")) {
        r += "<ul>";
      } else if (sc.test ("@/ul")) {
        r += "</ul>";
      } else if (sc.test ("@b")) {
        r += "<b>";
      } else if (sc.test ("@/b")) {
        r += "</b>";
      } else if (sc.test ("@u")) {
        r += "<u>";
      } else if (sc.test ("@/u")) {
        r += "</u>";
      } else if (sc.test ("@i")) {
        r += "<i>";
      } else if (sc.test ("@/i")) {
        r += "</i>";
      } else if (sc.test ("@pre") || sc.test ("@code")) {
        in_code = true;
        r += "<pre>";
      } else if (sc.test ("@/pre") || sc.test ("@/code")) {
        in_code = false;
        r += "</pre>";
      } else if (sc.test ("@@")) {
        r += "@";
      } else {
        r += "@";
        ++sc;
      }
    } else {
      r += *sc;
      ++sc;
    }
  }
  return r;
}

struct DocumentationParser 
{
  DocumentationParser (const std::string &formatted_doc)
    : hidden (false), qt_class (false)
  {
    tl::Extractor ex (formatted_doc.c_str ());
    while (*ex) {

      if (*ex == '@') {

        if (ex.test ("@hide")) {
          hidden = true;
        } else if (ex.test ("@qt")) {
          qt_class = true;
        } else if (ex.test ("@brief")) {

          ex.read (brief_doc, "\n");

        } else if (ex.test ("@alias")) {

          ex.read (alias, "\n");

        } else if (ex.test ("@return") || ex.test ("@returns")) {

          ex.read  (ret_val, "\n");

        } else if (ex.test ("@args")) {

          std::string a;
          ex.try_read (a, "\n");
          a = tl::trim (a);
          if (! a.empty ()) {
            args = tl::split (a, ",");
            for (std::vector<std::string>::iterator i = args.begin (); i != args.end (); ++i) {
              *i = tl::trim (*i);
            }
          }

        } else if (ex.test ("@param")) {

          std::string n;
          ex.try_read (n);
          params.push_back (std::make_pair (n, std::string ()));
          ex.read (params.back ().second, "\n");

        } else {
          doc += "@";
          ++ex;
        }

      } else {
        doc += *ex;
        ++ex;
      }

    }
  }

  std::string doc_html ()
  {
    std::string r;
    r = "<p>";

    size_t p = 0;
    size_t pe = std::string::npos;
    bool in_code = false;
    while ((pe = doc.find ("\n\n", p)) != std::string::npos) {
      r += escape_xml_with_formatting (std::string (doc, p, pe - p), in_code);
      p = pe;
      if (in_code) {
        while (p < doc.size () && doc [p] == '\n') {
          r += doc [p];
          ++p;
        }
      } else {
        while (p < doc.size () && doc [p] == '\n') {
          ++p;
        }
        if (p < doc.size ()) {
          r += "</p><p>";
        }
      }
    }

    if (p < doc.size ()) {
      r += escape_xml_with_formatting (std::string (doc, p), in_code);
    }

    r += "</p>";
    return r;
  }

  bool hidden, qt_class;
  std::string doc;
  std::string brief_doc;
  std::string alias;
  std::vector<std::string> args;
  std::string ret_val;
  std::vector<std::pair <std::string, std::string> > params;
};

//  A cache for the parsed class documentation

static std::map <std::string, DocumentationParser> s_cls_doc;

static DocumentationParser &cls_documentation (const gsi::ClassBase *cls)
{
  std::map <std::string, DocumentationParser>::iterator cd = s_cls_doc.find (cls->name ());
  if (cd != s_cls_doc.end ()) {
    return cd->second;
  } else {
    return s_cls_doc.insert (std::make_pair (cls->name (), DocumentationParser (cls->doc ()))).first->second;
  }
}

static const std::string &aliased_name (const gsi::ClassBase *cls)
{
  const std::string &alias = cls_documentation (cls).alias;
  if (alias.empty ()) {
    return cls->name ();
  } else {
    return alias;
  }
}

static std::string
replace_references (const std::string &t, const gsi::ClassBase *cls_base)
{
  std::string r;

  size_t q = 0;
  size_t p;
  while ((p = t.find ("\\", q)) != std::string::npos) {

    r += std::string (t, q, p - q);

    size_t pp = ++p;
    while (p < t.size () && (t[p] == '_' || isalnum (t [p]))) {
      ++p;
    }
    if (p < t.size () && (t[p] == '?' || t [p] == '=')) {
      ++p;
    }

    std::string id (t, pp, p - pp);

    std::string mid;

    if (p < t.size () && t[p] == '#') {
      pp = ++p;
      while (p < t.size () && (t[p] == '_' || isalnum (t [p]))) {
        ++p;
      }
      if (p < t.size () && (t[p] == '?' || t [p] == '=')) {
        ++p;
      }
      mid = std::string (t, pp, p - pp);
    }

    bool found = false;

    for (gsi::ClassBase::name_map_iterator nm = cls_base->begin_name_map (); nm != cls_base->end_name_map () && !found; ++nm) {
      if (nm->first.first == id) {
        r += "<a href=\"" + escape_xml (class_doc_path (std::string (), id)) + "\">" + escape_xml (id) + "</a>";
        found = true;
      }
    }

    for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes () && !found; ++c) {
      if (c->name () == id) {
        r += "<a href=\"";
        if (mid.empty ()) {
          r += escape_xml (class_doc_path (id));
        } else {
          r += escape_xml (class_doc_path (id, mid));
        }
        r += "\">";
        if (mid.empty ()) {
          r += escape_xml (id);
        } else if (id.empty ()) {
          r += escape_xml (mid);
        } else {
          r += escape_xml (id) + "#" + escape_xml (mid);
        }
        r += "</a>";
        found = true;
      }
    }

    if (! found) {
      r += escape_xml (id);
      if (! mid.empty ()) {
        r += "#";
        r += mid;
      }
    }

    q = p;

  }

  r += std::string (t, q);

  return r;
}

// --------------------------------------------------------------------------------------
//  Implementation

std::string 
GSIHelpProvider::folder () const
{
  return "code";
}

std::string 
GSIHelpProvider::title () const
{
  return tl::translate ("API Reference");
}

void 
GSIHelpProvider::toc (std::vector<std::string> &t)
{
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {

    DocumentationParser &doc = cls_documentation (&*c);

    // skip hidden classes
    if (doc.hidden) {
      continue;
    }

    t.push_back (class_doc_path (c->name ()));

  }
}

QDomDocument
GSIHelpProvider::get (const std::string &path) const
{
  QUrl url = QUrl::fromEncoded (path.c_str ());
  QString url_path = url.path ();
  QRegExp class_doc_url (QString::fromAscii ("^/code/class_(.*)\\.xml$"));

  std::string text;
  if (url_path == QString::fromAscii ("/code/index.xml")) {
    text = produce_class_index ();
  } else if (class_doc_url.indexIn (url_path) == 0) {
    text = produce_class_doc (tl::to_string (class_doc_url.cap (1)));
  } else {
    throw tl::Exception (tl::translate ("Page not found: ") + path);
  }

  QDomDocument doc;
  QString errorMsg;
  int errorLine = 0 ;
  if (! doc.setContent (QByteArray (text.c_str (), text.size ()), true, &errorMsg, &errorLine)) {
    throw tl::Exception (tl::to_string (errorMsg) + ", in line " + tl::to_string (errorLine) + " of " + path);
  }
  
  return doc;
}

std::string  
GSIHelpProvider::produce_class_index () const
{
  std::ostringstream os;

  os << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl
     << "<!DOCTYPE language SYSTEM \"klayout_doc.dtd\">" << std::endl
     << std::endl;

  os << "<doc><title>" << tl::translate ("Class Index") << "</title>" << std::endl;

  std::vector <std::pair <std::string, std::string> > class_names;
  std::vector <std::pair <std::string, std::string> > qt_class_names;
  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    DocumentationParser &doc = cls_documentation (&*c);
    if (! doc.hidden) {
      if (! doc.qt_class) {
        class_names.push_back (std::make_pair (c->name (), doc.brief_doc));
      } else {
        qt_class_names.push_back (std::make_pair (c->name (), doc.brief_doc));
      }
      os << "<topic-ref href=\"" << escape_xml (class_doc_path (c->name ())) << "\"/>" << std::endl;
    }
  }

  if (! qt_class_names.empty ()) {
    os << tl::translate ("Find Qt class documentation") << " <a href=\"#qtclasses\">" << tl::translate ("here") << "</a>" << std::endl;
  }

  if (! class_names.empty ()) {
    
    std::sort (class_names.begin (), class_names.end ());

    os << "<h2>KLayout classes</h2>" << std::endl;
    os << "<table>" << std::endl;
    int n = 0;
    for (std::vector <std::pair <std::string, std::string> >::const_iterator cc = class_names.begin (); cc != class_names.end (); ++cc, ++n) {
      os << "<tr class=\"row" << (n % 2)  << "\">" << std::endl;
      os << "<td><a href=\"" << escape_xml (class_doc_path (cc->first)) << "\">" << escape_xml (cc->first) << "</a></td><td>&nbsp;&nbsp;</td><td><nobr>" << escape_xml (cc->second) << "</nobr></td></tr>" << std::endl;
    }
    os << "</table>" << std::endl;

  }

  if (! qt_class_names.empty ()) {
    
    std::sort (qt_class_names.begin (), qt_class_names.end ());

    os << "<a name=\"qtclasses\"/><h2>Qt classes</h2>" << std::endl;
    os << "<table>" << std::endl;
    int n = 0;
    for (std::vector <std::pair <std::string, std::string> >::const_iterator cc = qt_class_names.begin (); cc != qt_class_names.end (); ++cc, ++n) {
      os << "<tr class=\"row" << (n % 2)  << "\">" << std::endl;
      os << "<td><a href=\"" << escape_xml (class_doc_path (cc->first)) << "\">" << escape_xml (cc->first) << "</a></td><td>&nbsp;&nbsp;</td><td><nobr>" << escape_xml (cc->second) << "</nobr></td></tr>" << std::endl;
    }
    os << "</table>" << std::endl;

  }

  os << "</doc>" << std::endl;

  return os.str ();

}

static std::string
type_to_s (const gsi::ArgType &a, bool linked, bool for_return)
{
  std::string s;
  switch (a.type) {
  case gsi::T_void_ptr:
    s += "void *"; break;
  case gsi::T_void:
    s += "void"; break;
  case gsi::T_bool:
    s += "bool"; break;
  case gsi::T_char:
    s += "char"; break;
  case gsi::T_schar:
    s += "signed char"; break;
  case gsi::T_uchar:
    s += "unsigned char"; break;
  case gsi::T_short:
    s += "short"; break;
  case gsi::T_ushort:
    s += "unsigned short"; break;
  case gsi::T_int:
    s += "int"; break;
  case gsi::T_uint:
    s += "unsigned int"; break;
  case gsi::T_long:
    s += "long"; break;
  case gsi::T_ulong:
    s += "unsigned long"; break;
  case gsi::T_longlong:
    s += "long long"; break;
  case gsi::T_ulonglong:
    s += "unsigned long long"; break;
  case gsi::T_double:
    s += "double"; break;
  case gsi::T_float:
    s += "float"; break;
  case gsi::T_string_ccptr:
  case gsi::T_string:
  case gsi::T_qstring:
    s += "string"; break;
  case gsi::T_var:
    s += "variant"; break;
  case gsi::T_object_new:
    if (linked) {
      s += "new <a href=\"" + escape_xml (class_doc_path (aliased_name (a.cls))) + "\">" + escape_xml (aliased_name (a.cls)) + "</a>";
    } else {
      s += "new " + aliased_name (a.cls);
    }
    break;
  case gsi::T_object:
    if (a.is_cptr || (! for_return && a.is_cref)) {
      s = "const ";
    }
    if (linked) {
      s += "<a href=\"" + escape_xml (class_doc_path (aliased_name (a.cls))) + "\">" + escape_xml (aliased_name (a.cls)) + "</a>";
    } else {
      s += aliased_name (a.cls);
    }
    break;
  case gsi::T_vector:
    if (a.inner) {
      s += type_to_s (*a.inner, linked, false);
    } 
    s += "[]";
    break;
  }
  if (a.is_cptr || a.is_ptr) {
    s += " ptr";
  }
  return s;
}

static bool 
is_deprecated (const std::string &n, const gsi::MethodBase *m)
{
  std::string nn = "#" + n;
  size_t p = m->name ().find (nn);
  return (p != std::string::npos && (p + nn.size () >= m->name ().size () || ! (isalnum (m->name ()[p + nn.size ()]) || m->name ()[p + nn.size ()] == '_')));
}

//  extract protected flag from the method name
static const char *
extract_protected (const char *name, bool &prot)
{
  if (*name == '*' && name[1] && name[1] != '=') {
    ++name;
    prot = true;
  } else {
    prot = false;
  }
  return name;
}

static std::string 
deprecated_alias (const std::string &n, const gsi::MethodBase *m)
{
  bool prot = false;
  std::string pname (extract_protected (m->name ().c_str (), prot));
  size_t spos = pname.find ("|");
  if (spos != std::string::npos) {
    return std::string (pname, 0, spos);
  } else {
    return std::string ();
  }
}

static bool 
is_protected (const gsi::MethodBase *m)
{
  bool prot = false;
  extract_protected (m->name ().c_str (), prot);
  return prot;
}

static std::string 
method_attributes (const gsi::MethodBase *method, DocumentationParser &doc, bool without_static = false, bool without_prot = true)
{
  std::string r;
  if (method->is_event ()) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "event";
  }
  if (method->is_callback ()) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "virtual";
  }
  if (! without_static && method->is_static ()) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "static";
  }
  if (method->is_const ()) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "const";
  }
  if (method->ret_type ().is_iter) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "iter";
  }
  if (! without_prot && is_protected (method)) {
    if (! r.empty ()) {
      r += ",";
    }
    r += "protected";
  }
  return r;
}

static std::string 
method_return (const gsi::MethodBase *method, DocumentationParser &doc, bool linked = false)
{
  return type_to_s (method->ret_type (), linked, true);
}

static std::string
method_arguments (const gsi::MethodBase *method, DocumentationParser &doc, bool linked = false, const char *sep = "<br/>")
{
  std::string r;
  if (method->begin_arguments () == method->end_arguments ()) {

    if (! doc.args.empty ()) {

      r += "(";

      //  special case of external method - just dump the argument names
      for (size_t i = 0; i < doc.args.size (); ++i) {
        if (i > 0) {
          r += ",";
          r += sep;
          r += "&nbsp;";
        }
        r += escape_xml (doc.args [i]);
      }

      r += ")";

    }

    return r;

  } else {

    r += "(";

    int n = 0;
    for (gsi::MethodBase::argument_iterator a = method->begin_arguments (); a != method->end_arguments (); ++a, ++n) {
      if (n > 0) {
        r += ",";
        r += sep;
        r += "&nbsp;";
      }
      r += type_to_s (*a, linked, false);
      r += " ";
      if (n < int (doc.args.size ())) {
        r += escape_xml (doc.args [n]);
      } else {
        r += "arg" + tl::to_string (n + 1);
      }
    }

    r += ")";

  }

  return r;
}

class DummyMethod
  : public gsi::MethodBase
{
public:
  DummyMethod (const std::string &name, const std::string &doc, bool c, bool s)
    : gsi::MethodBase (name, doc, c, s)
  { }

  virtual void call (void *cls, gsi::SerialArgs &args, gsi::SerialArgs &ret) const { } 

  virtual gsi::MethodBase *clone () const { return 0; }
};

class DummyMethodDefaultCtor
  : public DummyMethod
{
public:
  DummyMethodDefaultCtor (const gsi::ClassBase *cls)
    : DummyMethod ("new", "@brief Creates a new object of this class\n", false, true)
  {
    gsi::ArgType ret;
    ret.is_ref = true;
    ret.type = gsi::T_object;
    ret.cls = cls;
    set_return (ret);
  }
};

class DummyMethodDestroyed
  : public DummyMethod
{
public:
  DummyMethodDestroyed (const gsi::ClassBase *cls)
    : DummyMethod ("destroyed?", "@brief Returns a value indicating whether the object was already destroyed\nThis method returns true, if the object was destroyed, either explicitly or by the C++ side.\n" 
                                  "The latter may happen, if the object is owned by a C++ object which got destroyed itself.", true, false)
  {
    gsi::ArgType ret;
    ret.type = gsi::T_bool;
    set_return (ret);
  }
};

class DummyMethodIsConst
  : public DummyMethod
{
public:
  DummyMethodIsConst (const gsi::ClassBase *cls)
    : DummyMethod ("is_const_object?", "@brief Returns a value indicating whether the reference is a const reference\nThis method returns true, if self is a const reference.\n" 
                                       "In that case, only const methods may be called on self.", true, false)
  {
    gsi::ArgType ret;
    ret.type = gsi::T_bool;
    set_return (ret);
  }
};

class DummyMethodCreate
  : public DummyMethod
{
public:
  DummyMethodCreate (const gsi::ClassBase *cls)
    : DummyMethod ("create", "@brief Ensures the C++ object is created\n"
                             "Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. "
                             "Usually C++ objects are created on demand and not necessarily when the script object is created.", false, false)
  {
  }
};

class DummyMethodDup
  : public DummyMethod
{
public:
  DummyMethodDup (const gsi::ClassBase *cls)
    : DummyMethod ("dup", "@brief Creates a copy of self\n", true, false)
  {
    gsi::ArgType ret;
    ret.is_cref = true;
    ret.type = gsi::T_object;
    ret.cls = cls;
    set_return (ret);
  }
};

class DummyMethodAssign
  : public DummyMethod
{
public:
  DummyMethodAssign (const gsi::ClassBase *cls)
    : DummyMethod ("assign", "@args other\n@brief Assign the contents of another object to self\nThis method assigns the contents of another object to self. "
                             "This is a deep copy that does not only copy the reference but the actual content.\n", true, false)
  {
    gsi::ArgType a;
    a.is_cref = true;
    a.type = gsi::T_object;
    a.cls = cls;
    add_arg (a);
  }
};

std::string
GSIHelpProvider::produce_class_doc (const std::string &cls) const
{
  std::ostringstream os;

  const gsi::ClassBase *cls_obj = 0;

  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
    if (c->name () == cls) {
      cls_obj = &*c;
      break;
    }
  }

  if (! cls_obj) {
    return "Unknown class: " + cls;
  }

  DocumentationParser &class_doc = cls_documentation (cls_obj);

  os << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl
     << "<!DOCTYPE language SYSTEM \"klayout_doc.dtd\">" << std::endl
     << std::endl;

  os << "<doc><title>" << tl::translate ("API reference - Class") << " " << escape_xml (cls) << "</title>" << std::endl;

  os << "<keyword name=\"" << escape_xml (cls) << "\"/>" << std::endl;

  os << "<p>"
     << "<link href=\"/about/rba_notation.xml\"/>"
     << "</p>" << std::endl;

  os << "<p><b>" << tl::translate ("Description") << "</b>: " << escape_xml (class_doc.brief_doc) << "</p>" << std::endl;

  std::vector<const gsi::ClassBase *> classes;
  classes.push_back (cls_obj);

  if (cls_obj->base ()) {

    const gsi::ClassBase *base = cls_obj->base ();
    const gsi::ClassBase *last_cls = cls_obj;
    bool all_collected = false;

    os << "<p><b>" << tl::translate ("Class hierarchy") << "</b>: " << cls_obj->name ();
    while (base) {

      DocumentationParser &bdoc = cls_documentation (base);
      if (! bdoc.alias.empty ()) {
        //  suppress direct base class alias to our class (x_Native for x)
        if (bdoc.alias != last_cls->name ()) {
          os << " &#187; <a href=\"" << escape_xml (class_doc_path (bdoc.alias)) << "\">" << escape_xml (bdoc.alias) << "</a>";
          all_collected = true;
        } else if (! all_collected) {
          classes.push_back (base);
        }
      } else if (! bdoc.hidden) {
        os << " &#187; <a href=\"" << escape_xml (class_doc_path (base->name ())) << "\">" << escape_xml (base->name ()) << "</a>";
        all_collected = true;
      } else if (! all_collected) {
        //  class needs to be mixed into the parent
        classes.push_back (base);
      }

      last_cls = base;
      base = base->base ();

    }

    os << "</p>" << std::endl;

  }

  os << "<h2-index/>" << std::endl;

  os << replace_references (class_doc.doc_html (), cls_obj) << std::endl;

  //  collect the methods of the class and their hidden base classes
  //  (in the reverse order so that derived classes override their super classes methods)

  std::multimap <std::string, const gsi::MethodBase *> mm;

  for (size_t n = classes.size (); n > 0; ) {

    --n;

    //  remove the base classes' definitions if the name match
    for (gsi::ClassBase::name_map_iterator nm = classes [n]->begin_name_map (); nm != classes [n]->end_name_map (); ++nm) {
      std::multimap <std::string, const gsi::MethodBase *>::iterator im;
      while ((im = mm.find (nm->first.first)) != mm.end ()) {
        mm.erase (im);
      }
    }

    //  add the current classes' methods
    for (gsi::ClassBase::name_map_iterator nm = classes [n]->begin_name_map (); nm != classes [n]->end_name_map (); ++nm) {
      unsigned int mid = nm->second;
      for (gsi::ClassBase::overloaded_methods_iterator m = classes [n]->begin_overloaded_methods (mid); m != classes [n]->end_overloaded_methods (mid); ++m) {
        const gsi::MethodBase *mb = classes [n]->begin_methods () [*m];
        DocumentationParser mdoc (mb->doc ());
        if (! mdoc.hidden) {
          mm.insert (std::make_pair (nm->first.first, mb));
        }
      }
    }

  }

  //  Add the default methods not listed in the method table
  
  DummyMethodDefaultCtor def_ctor (cls_obj);
  DummyMethodCreate def_create (cls_obj);
  DummyMethod def_destroy ("destroy", "@brief Explicitly destroy the object\nExplicitly destroy the object on C++ side if it was owned by the Ruby interpreter. Subsequent access to this object will throw an exception.\n" 
                                      "If the object is not owned by Ruby, this method will do nothing.", false, false);
  DummyMethodDestroyed def_destroyed (cls_obj);
  DummyMethodIsConst def_is_const (cls_obj);
  DummyMethodDup def_dup (cls_obj);
  DummyMethodAssign def_assign (cls_obj);

  if (cls_obj->declaration ()) {

    if (cls_obj->can_default_create () && mm.find ("new") == mm.end ()) {
      mm.insert (std::make_pair ("new", &def_ctor));
    }

    mm.insert (std::make_pair ("create", &def_create));

    if (cls_obj->can_destroy ()) {
      mm.insert (std::make_pair ("destroy", &def_destroy));
    }

    mm.insert (std::make_pair ("destroyed?", &def_destroyed));

    mm.insert (std::make_pair ("is_const_object?", &def_is_const));

    if (cls_obj->can_copy ()) {
      mm.insert (std::make_pair ("dup", &def_dup));
      mm.insert (std::make_pair ("assign", &def_assign));
    }

  }

  if (mm.empty ()) {
    return os.str ();
  }

  int n = 0;
  int row = 0;
  bool any = false;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    if (i->second->is_static () && i->first.find ("new") == 0 && ! is_deprecated (i->first, i->second) && ! is_protected (i->second)) {

      if (! any) {
        os << "<h2>" << tl::translate ("Public constructors") << "</h2>" << std::endl;
        os << "<table>" << std::endl;
        any = true;
      }

      DocumentationParser method_doc (i->second->doc ());
      os << "<tr class=\"row" << (row % 2)  << "\">" << std::endl;
      ++row;
      os << "<td>" << method_return (i->second, method_doc) << "</td>";
      os << "<td><b><a href=\"#method" << n << "\">" << escape_xml (i->first) << "</a></b></td>";
      os << "<td>" << method_arguments (i->second, method_doc) << "</td>";
      os << "<td>" << replace_references (escape_xml (method_doc.brief_doc), cls_obj) << "</td>";
      os << "</tr>" << std::endl;

    }

  }

  if (any) {
    os << "</table>" << std::endl;
  }

  n = 0;
  row = 0;
  any = false;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    if (! i->second->is_static () && ! is_deprecated (i->first, i->second) && ! is_protected (i->second)) {

      if (! any) {
        os << "<h2>" << tl::translate ("Public methods") << "</h2>" << std::endl;
        os << "<table>" << std::endl;
        any = true;
      }

      DocumentationParser method_doc (i->second->doc ());
      os << "<tr class=\"row" << (row % 2)  << "\">" << std::endl;
      ++row;
      std::string attr = method_attributes (i->second, method_doc, true /*without static*/);
      if (! attr.empty ()) {
        os << "<td><i>[" << escape_xml (method_attributes (i->second, method_doc)) << "]</i></td>";
      } else {
        os << "<td></td>";
      }
      os << "<td>" << method_return (i->second, method_doc) << "</td>";
      os << "<td><b><a href=\"#method" << n << "\">" << escape_xml (i->first) << "</a></b></td>";
      os << "<td>" << method_arguments (i->second, method_doc) << "</td>";
      os << "<td>" << replace_references (escape_xml (method_doc.brief_doc), cls_obj) << "</td>";
      os << "</tr>" << std::endl;

    }

  }

  if (any) {
    os << "</table>" << std::endl;
  }

  any = false;

  n = 0;
  row = 0;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    if (i->second->is_static () && i->first.find ("new") != 0 && ! is_deprecated (i->first, i->second) && ! is_protected (i->second)) {

      if (! any) {
        any = true;
        os << "<h2>" << tl::translate ("Public static methods and constants") << "</h2>" << std::endl;
        os << "<table>" << std::endl;
      }

      DocumentationParser method_doc (i->second->doc ());
      os << "<tr class=\"row" << (row % 2)  << "\">" << std::endl;
      ++row;
      std::string attr = method_attributes (i->second, method_doc, true /*without static*/);
      if (! attr.empty ()) {
        os << "<td><i>[" << escape_xml (method_attributes (i->second, method_doc)) << "]</i></td>";
      } else {
        os << "<td></td>";
      }
      os << "<td>" << method_return (i->second, method_doc) << "</td>";
      os << "<td><b><a href=\"#method" << n << "\">" << escape_xml (i->first) << "</a></b></td>";
      os << "<td>" << method_arguments (i->second, method_doc) << "</td>";
      os << "<td>" << replace_references (escape_xml (method_doc.brief_doc), cls_obj) << "</td>";
      os << "</tr>" << std::endl;

    }

  }

  if (any) {
    os << "</table>" << std::endl;
  }

  any = false;

  n = 0;
  row = 0;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    if (! is_deprecated (i->first, i->second) && is_protected (i->second)) {

      if (! any) {
        any = true;
        os << "<h2>" << tl::translate ("Protected methods (static, non-static and constructors)") << "</h2>" << std::endl;
        os << "<table>" << std::endl;
      }

      DocumentationParser method_doc (i->second->doc ());
      os << "<tr class=\"row" << (row % 2)  << "\">" << std::endl;
      ++row;
      std::string attr = method_attributes (i->second, method_doc);
      if (! attr.empty ()) {
        os << "<td><i>[" << escape_xml (method_attributes (i->second, method_doc)) << "]</i></td>";
      } else {
        os << "<td></td>";
      }
      os << "<td>" << method_return (i->second, method_doc) << "</td>";
      os << "<td><b><a href=\"#method" << n << "\">" << escape_xml (i->first) << "</a></b></td>";
      os << "<td>" << method_arguments (i->second, method_doc) << "</td>";
      os << "<td>";
      os << "<td>" << replace_references (escape_xml (method_doc.brief_doc), cls_obj) << "</td>";
      os << "</td>";
      os << "</tr>" << std::endl;

    }

  }

  if (any) {
    os << "</table>" << std::endl;
  }

  any = false;

  n = 0;
  row = 0;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    if (is_deprecated (i->first, i->second)) {

      if (! any) {
        any = true;
        os << "<h2>" << tl::translate ("Deprecated methods (protected, public, static, non-static and constructors)") << "</h2>" << std::endl;
        os << "<table>" << std::endl;
      }

      DocumentationParser method_doc (i->second->doc ());
      os << "<tr class=\"row" << (row % 2)  << "\">" << std::endl;
      ++row;
      std::string attr = method_attributes (i->second, method_doc, false, false);
      if (! attr.empty ()) {
        os << "<td><i>[" << escape_xml (method_attributes (i->second, method_doc)) << "]</i></td>";
      } else {
        os << "<td></td>";
      }
      os << "<td>" << method_return (i->second, method_doc) << "</td>";
      os << "<td><b><a href=\"#method" << n << "\">" << escape_xml (i->first) << "</a></b></td>";
      os << "<td>" << method_arguments (i->second, method_doc) << "</td>";
      os << "<td>";
      if (! deprecated_alias (i->first, i->second).empty ()) {
        os << tl::translate ("Use of this method is deprecated. Use") << " " << deprecated_alias (i->first, i->second) << " " << tl::translate ("instead");
      } else {
        os << tl::translate ("Use of this method is deprecated");
      }
      os << "</td>";
      os << "</tr>" << std::endl;

    }

  }

  if (any) {
    os << "</table>" << std::endl;
  }

  n = 0;

  os << "<a name=\"detailed\"/><h2>" << tl::translate ("Detailed description") << "</h2>" << std::endl;

  for (std::multimap <std::string, const gsi::MethodBase *>::const_iterator i = mm.begin (); i != mm.end (); ++i, ++n) {

    DocumentationParser method_doc (i->second->doc ());

    os << "<a name=\"method" << n << "\"/>"
       << "<a name=\"m_" << escape_xml (i->first) << "\"/>"
       << "<keyword title=\"" << tl::translate ("API reference - Class") << " " << escape_xml (cls) << ", " << tl::translate ("Method") << " " << escape_xml (i->first) <<  "\" name=\"" << escape_xml (cls) << "#" << escape_xml (i->first) << "\"/>" << std::endl;
    
    os << "<h3>";
    std::string attr = method_attributes (i->second, method_doc);
    if (! attr.empty ()) {
      os << "<i>[" << escape_xml (attr) << "] </i>";
    }
    os << method_return (i->second, method_doc, true) << " " << escape_xml (i->first) << method_arguments (i->second, method_doc, true, "");
    os << "</h3>" << std::endl;

    os << "<p><b>" << tl::translate ("Description") << "</b>: " << replace_references (escape_xml (method_doc.brief_doc), cls_obj) << "</p>" << std::endl;

    if (! method_doc.params.empty () || ! method_doc.ret_val.empty ()) {
      os << "<table>" << std::endl;
      for (std::vector<std::pair <std::string, std::string> >::const_iterator p = method_doc.params.begin (); p != method_doc.params.end (); ++p) {
        os << "<tr><td><b>" << escape_xml (p->first) << "</b>:</td><td>" << replace_references (escape_xml (p->second), cls_obj) << "</td></tr>" << std::endl;
      }
      if (! method_doc.ret_val.empty ()) {
        os << "<tr><td><b>" << tl::translate ("Returns") << "</b>:</td><td>" << replace_references (escape_xml (method_doc.ret_val), cls_obj) << "</td></tr>" << std::endl;
      }
      os << "</table>" << std::endl;
    }

    if (is_deprecated (i->first, i->second)) {
      if (deprecated_alias (i->first, i->second).empty ()) {
        os << "<p>" << tl::translate ("This method is deprecated.") << "</p>" << std::endl;
      } else {
        os << "<p>" << tl::translate ("This method is deprecated. Use method '") << deprecated_alias (i->first, i->second) << tl::translate ("' instead") << "</p>" << std::endl;
      }
    } else {
      std::string dh = method_doc.doc;
      if (! tl::Extractor (dh.c_str ()).at_end ()) {
        os << "<p>" << replace_references (method_doc.doc_html (), cls_obj) << "</p>" << std::endl;
      }
    }

  }

  os << "</doc>" << std::endl;

  return os.str ();

}

static tl::RegisteredClass<lay::HelpProvider> gsi_help_provider (new GSIHelpProvider (), 1000);

}


