
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_rba
#define _HDR_rba

#include "gsi.h"

namespace rba
{

/**
 *  @brief A piece of backtrace information
 */
struct KLAYOUT_DLL BacktraceElement
{
  /**  
   *  @brief Constructor: create an element with a file and a line information
   */
  BacktraceElement(const std::string &_file, int _line);

  /**  
   *  @brief Constructor: create an element with a file, a line information and more information
   */
  BacktraceElement(const std::string &_file, int _line, const std::string _more_info);

  /**
   *  @brief Default constructor
   */
  BacktraceElement();

  /**
   *  @brief Convert the information to a string
   */
  std::string to_string() const;

  std::string file;
  int line;
  std::string more_info;
};

/**
 *  @brief A basic exception class
 *
 *  TODO: this class should go into gsi namespace
 */
class KLAYOUT_DLL RuntimeError
  : public tl::Exception 
{
public:
  RuntimeError (const char *msg, const char *cls, const std::vector <BacktraceElement> &backtrace)
    : tl::Exception (msg), m_line (-1), m_cls (cls), m_backtrace (backtrace)
  { }

  RuntimeError (const char *msg, const char *sourcefile, int line, const char *cls, const std::vector <BacktraceElement> &backtrace)
    : tl::Exception (msg), m_sourcefile (sourcefile), m_line (line), m_cls (cls), m_backtrace (backtrace)
  { }

  RuntimeError (const RuntimeError &d)
    : tl::Exception (d), m_sourcefile (d.m_sourcefile), m_line (d.m_line), m_cls (d.m_cls), m_context (d.m_context), m_backtrace (d.m_backtrace)
  { }

  const std::string &sourcefile () const 
  { 
    return m_sourcefile; 
  }

  void set_sourcefile (const std::string &sourcefile)
  {
    m_sourcefile = sourcefile;
  }

  int line () const
  { 
    return m_line; 
  }

  void set_line (int line)
  {
    m_line = line;
  }

  const std::string &cls () const
  { 
    return m_cls; 
  }

  void set_cls (const std::string &cls)
  {
    m_cls = cls;
  }

  const std::string &context () const
  { 
    return m_context; 
  }

  void set_context (const std::string &context)
  {
    m_context = context;
  }

  const std::vector<BacktraceElement> &backtrace () const
  { 
    return m_backtrace; 
  }

  virtual std::string msg () const;

  std::string basic_msg () const;

private:
  std::string m_sourcefile;
  int m_line;
  std::string m_cls;
  std::string m_context;
  std::vector<BacktraceElement> m_backtrace;
};

/**
 *  @brief An exception class indicating an exit  
 *
 *  This exception can be thrown by the C++ client code and is translated into the same exception 
 *  on the user side.
 */
class KLAYOUT_DLL ExitException
  : public tl::Exception
{
public:
  ExitException () : tl::Exception ("exit"), m_status (1) { }
  ExitException (int status) : tl::Exception ("exit"), m_status (status) { }

  int status() const { return m_status; }

private:
  int m_status;
};

/**
 *  @brief A class encapsulating a ruby exception
 */
class KLAYOUT_DLL RubyError
  : public RuntimeError
{
public:
  RubyError (const char *msg, const char *cls, const std::vector <BacktraceElement> &backtrace)
    : RuntimeError (msg, cls, backtrace)
  { }

  RubyError (const char *msg, const char *sourcefile, int line, const char *cls, const std::vector <BacktraceElement> &backtrace)
    : RuntimeError (msg, sourcefile, line, cls, backtrace)
  { }

  RubyError (const RubyError &d)
    : RuntimeError (d)
  { }
};

/**
 *  @brief The console interface for output
 *
 *  Using this interface, the interpreter can output text on stdout or stderr streams.
 *  This interface is implemented by ConsoleImpl.
 */
class KLAYOUT_DLL Console
{
public:
  enum output_stream { OS_none, OS_stdout, OS_stderr, OS_echo };

  Console () { }
  virtual ~Console () { }

  virtual void write_str (const char *text, output_stream os) = 0;
};

/**
 *  @brief An interface delivering a stack trace for the ExecutionHandler's trace function
 */
class KLAYOUT_DLL StackTraceProvider
{
public:
  StackTraceProvider () { }
  virtual ~StackTraceProvider () { }
  virtual std::vector<BacktraceElement> stack_trace () const = 0;
  virtual int stack_depth () const = 0;
  virtual size_t scope_index () const = 0;
};

/**
 *  @brief An interface controlling the execution of a script
 *
 *  This interface allows to control the execution of a script. Basically when a script
 *  is executed, the handler allows to intercept the execution and implement breakpoints 
 *  for example
 */
class KLAYOUT_DLL ExecutionHandler
{
public:
  /**
   *  @brief The constructor
   */
  ExecutionHandler () { }

  /**
   *  @brief The destructor 
   */
  virtual ~ExecutionHandler () { }

  /**
   *  @brief Indicates the start of the execution of a block of code
   *
   *  This method is called when the execution begins. It can be used to clear any caches for example.
   */
  virtual void start_exec () { }

  /**
   *  @brief Indicates then end of the execution of a block of code
   *
   *  This method is called after the execution has finished.
   */
  virtual void end_exec () { }

  /**
   *  @brief Indicates that we descend into a call
   */
  virtual void push_call_stack () { }

  /**
   *  @brief Indicates that we ascend into a call
   */
  virtual void pop_call_stack () { }

  /**
   *  @brief Indicates that an exception is thrown
   */
  virtual void exception_thrown (size_t file_id, int line, const std::string &eclass, const std::string &emsg, const StackTraceProvider *stack_trace_provider) { }

  /**
   *  @brief Associate a file path with an ID
   *
   *  This method is supposed to deliver an arbitrary integer ID (which can hold a pointer for
   *  example) for a given file path string. When the trace method is called, this ID is used to
   *  identify the file instead of the heavy file path.
   */
  virtual size_t id_for_path (const std::string &path) { return 0; }

  /**
   *  @brief This method is called during execution
   */
  virtual void trace (size_t file_id, int line, const StackTraceProvider *stack_trace_provider) { }
};

/**
 *  @brief The ruby interpreter wrapper class
 */
class KLAYOUT_DLL RubyInterpreter
{
public:
  RubyInterpreter ();
  ~RubyInterpreter ();

  /**
   *  @brief Add the given path to the search path ($: in ruby)
   */
  void add_path (const std::string &path);

  /**
   *  @brief Requires the given module (ruby "require")
   */
  void require (const std::string &filename);

  /**
   *  @brief Set the given debugger scope
   *
   *  The debugger scope is the element to which the back trace will be reduced.
   *  Specifically this suppresses calls from inner functions called from that file.
   *  This is useful for DSL implementations.
   */
  void set_debugger_scope (const std::string &filename);

  /**
   *  @brief Removes the debugger scope
   */
  void remove_debugger_scope ();

  /**
   *  @brief Ignores the next exception
   *
   *  This is useful for suppressing reraised exceptions in the debugger.
   */
  void ignore_next_exception ();

  /**
   *  @brief Load the given file (ruby "load")
   */
  void load_file (const std::string &filename);

  /**
   *  @brief Evaluates the given string 
   *
   *  The filename and line gives the location at which the evaluation should begin.
   *  This location is indicated in the stack trace and error messages
   */
  void eval_string (const char *string, const char *filename = 0, int line = 1);

  /**
   *  @brief Evaluates the given string and returns the results as string
   *
   *  The filename and line gives the location at which the evaluation should begin.
   *  This location is indicated in the stack trace and error messages
   */
  std::string eval_as_string (const char *string, const char *filename = 0, int line = 1);
 
  /**
   *  @brief Evaluates the given string and prints the result to stdout
   *
   *  The filename and line gives the location at which the evaluation should begin.
   *  This location is indicated in the stack trace and error messages
   */ 
  void eval_string_and_print (const char *string, const char *filename = 0, int line = 1);

  /**
   *  @brief Defines a global variable with the given name and value 
   */
  void define_variable (const std::string &name, const std::string &value);

  /**
   *  @brief Gets a value indicating whether the interpreter is available
   *
   *  TODO: should be generic for other languages.
   */
  bool ruby_available () const;

  /**
   *  @brief Installs the given console for output
   */
  void push_console (Console *console);

  /**
   *  @brief Removes the given console
   */
  void remove_console (Console *console);

  /**
   *  @brief Installs the given execution handler
   *
   *  The execution handler is informed when the interpreter enters code execution (also
   *  from the outside, i.e. a method reimplementing a C++ method or a event handler).
   *  During execution, the handler receives trace events which allow him to intercept
   *  execution.
   */
  void push_exec_handler (ExecutionHandler *exec_handler);

  /**
   *  @brief Removes the given execution handler
   */
  void remove_exec_handler (ExecutionHandler *exec_handler);

  /**
   *  @brief Fetch the version string
   *
   *  Returns an empty string when no Ruby interpreter is installed.
   */
  std::string version () const;

  /**
   *  @brief Provide a first (basic) initialization
   */
  static void initialize ();
};

}

#endif


