
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2015 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "gsiDecl.h"
#include "gsiDeclBasic.h"
#include "gsiExpression.h"
#include "rba.h"
#include "tlString.h"
#include "tlInternational.h"
#include "tlException.h"
#include "tlAssert.h"
#include "tlLog.h"

#include <map>
#include <set>
#include <memory>
#include <iostream>
#include <cctype>

#include <QtCore/QString>

#if !defined(HAVE_RUBY_VERSION_CODE)
#  define HAVE_RUBY_VERSION_CODE 10901
#endif

namespace rba
{

// -------------------------------------------------------------------
//  BacktraceElement implementation

BacktraceElement::BacktraceElement (const std::string &_file, int _line)
  : file (_file), line (_line)
{
  // .. nothing yet ..
}

BacktraceElement::BacktraceElement (const std::string &_file, int _line, const std::string _more_info)
  : file (_file), line (_line), more_info (_more_info)
{
  // .. nothing yet ..
}

BacktraceElement::BacktraceElement ()
  : line (0)
{
  // .. nothing yet ..
}

std::string 
BacktraceElement::to_string() const
{
  if (line > 0) {
    return file + ":" + tl::to_string (line) + ":" + more_info;
  } else {
    return more_info;
  }
}

// -------------------------------------------------------------------
//  RuntimeError implementation

std::string
RuntimeError::basic_msg () const
{
  std::string m = tl::Exception::msg ();
  if (! m_context.empty ()) {
    m += tl::translate (" in ") + m_context;
  }
  return m;
}

std::string 
RuntimeError::msg () const
{
  std::string m = cls ();
  m += ": ";
  m += basic_msg ();

  for (std::vector<BacktraceElement>::const_iterator bt = backtrace ().begin (); bt != backtrace ().end (); ++bt) {
    m += "\n  ";
    m += bt->to_string ();
  }

  return m;
}

}

// -------------------------------------------------------------------
//  This part is available only if Ruby is enabled

#ifdef HAVE_RUBY

#include <ruby.h>
#include <signal.h>

//  define getter for top_self
static VALUE ruby_top_self = Qnil;
static VALUE get_top_self ()
{
  return ruby_top_self;
}

//  define a setter for $0
static void set_progname (VALUE pn)
{
  rb_gv_set ("PROGRAM_NAME", pn);
}

//  Add some compatibility definitions for Ruby 1.8
#if HAVE_RUBY_VERSION_CODE < 10900

#  include "node.h"

//  Compatibility macros for Ruby 1.8
#  if !defined(RSTRING_PTR)
#    define RSTRING_PTR(s) (RSTRING(s)->ptr)
#  endif

#  if !defined(RSTRING_LEN)
#    define RSTRING_LEN(s) (RSTRING(s)->len)
#  endif

#  if !defined(RARRAY_PTR)
#    define RARRAY_PTR(s) (RARRAY(s)->ptr)
#  endif

#  if !defined(RARRAY_LEN)
#    define RARRAY_LEN(s) (RARRAY(s)->len)
#  endif

#  if !defined(RCLASS_SUPER)
#    define RCLASS_SUPER(k) (RCLASS (k)->super)
#  endif

//  Ruby 1.8 does not define NUL2ULL
#  if !defined(NUM2ULL) 
#    define NUM2ULL(x) rb_num2ull((VALUE)x)
#  endif

//  add rb_errinfo and rb_set_errinfo for Ruby 1.8
static VALUE rb_errinfo()
{
  return rb_gv_get ("$!");
}

static void rb_set_errinfo(VALUE v)
{
  rb_gv_set ("$!", v);
}

//  add rb_sourcefile and rb_sourceline for Ruby 1.8
static const char *rb_sourcefile ()
{
  ruby_set_current_source ();
  return ruby_sourcefile;
}

static int rb_sourceline ()
{
  ruby_set_current_source ();
  return ruby_sourceline;
}

//  add an emulation of rb_binding_new
static VALUE rb_binding_new ()
{
  return rb_funcall (get_top_self(), rb_intern ("binding"), 0);
}

#endif

//  Define a method to create a Ruby const char * string from a UTF-8 strings.
//  This method is used for rb_require, rb_load.
static std::string rb_cstring_from_utf8 (const std::string &utf8)
{
	// TODO: implement (how?)
	return utf8;
}

namespace rba
{

static void rba_get_backtrace_from_array (VALUE backtrace, std::vector<BacktraceElement> &bt, unsigned int skip = 0);

// -------------------------------------------------------------------
//  RubyStackTraceProvider definition and implementation

class KLAYOUT_DLL RubyStackTraceProvider
  : public StackTraceProvider
{
public:
  RubyStackTraceProvider (const std::string &scope)
    : m_scope (scope) 
  { }

  virtual std::vector<BacktraceElement> stack_trace () const 
  {
    std::vector<BacktraceElement> bt;
    bt.push_back (BacktraceElement (rb_sourcefile (), rb_sourceline ()));
    rba_get_backtrace_from_array (rb_funcall (rb_mKernel, rb_intern ("caller"), 0), bt, 0);
    return bt;
  }


  virtual size_t scope_index () const
  {
    if (! m_scope.empty ()) {
      std::vector<BacktraceElement> bt = stack_trace ();
      for (size_t i = 0; i < bt.size (); ++i) {
        if (bt[i].file == m_scope) {
          return i;
        }
      }
    }
    return 0;
  }

  virtual int stack_depth () const
  {
    // TODO: there should be a more efficient implementation!!
    std::vector<BacktraceElement> bt;
    bt.push_back (BacktraceElement (rb_sourcefile (), rb_sourceline ()));
    rba_get_backtrace_from_array (rb_funcall (rb_mKernel, rb_intern ("caller"), 0), bt, 0);
    return int (bt.size ());
  }

  //  we could use this for ruby >= 1.9.3
#if 0
  static int
  count_stack_levels(void *arg, VALUE file, int line, VALUE method)
  {
    *(int *)arg += 1;
    return 0;
  }

  extern "C" int rb_backtrace_each (int (*iter)(void *arg, VALUE file, int line, VALUE method), void *arg);

  virutal int stack_depth ()
  {
    int l = 0;
    rb_backtrace_each(count_stack_levels, &l);
    return l;
  }
#endif

private:
  std::string m_scope;
};

// --------------------------------------------------------------------------
//  Global variables

static std::map <VALUE, const gsi::ClassBase *> cls_map;
static std::map <const gsi::ClassBase *, VALUE> rev_cls_map;

static RubyInterpreter *rba_interpreter = 0;
static bool rba_ruby_destroyed = false;

static VALUE saved_stderr = Qnil;
static VALUE saved_stdout = Qnil;
static VALUE stdout_klass = Qnil, stderr_klass = Qnil;

//  TODO: move this into a singleton instance and provide thread-safetiness
static Console *sp_current_console = 0;
static std::vector<Console *> s_consoles;
static ExecutionHandler *sp_current_exec_handler = 0;
static int s_current_exec_level = 0;
static bool s_in_trace = false;
static bool s_exit_on_next = false;
static bool s_block_exceptions = false;
static bool s_ignore_next_exception = false;
static std::string s_debugger_scope;
static VALUE s_current_binding = Qnil;
static std::map<const char *, size_t> s_file_id_map;
static std::vector<ExecutionHandler *> s_exec_handlers;

#define RUBY_BEGIN_EXEC \
  try { \
    s_file_id_map.clear (); \
    s_exit_on_next = false; \
    s_block_exceptions = false; \
    if (s_current_exec_level++ == 0 && sp_current_exec_handler) { \
      sp_current_exec_handler->start_exec (); \
    }

#define RUBY_END_EXEC \
    if (s_current_exec_level > 0 && --s_current_exec_level == 0 && sp_current_exec_handler) { \
      sp_current_exec_handler->end_exec (); \
    } \
  } catch (...) { \
    if (s_current_exec_level > 0 && --s_current_exec_level == 0 && sp_current_exec_handler) { \
      sp_current_exec_handler->end_exec (); \
    } \
    if (! s_exit_on_next) { \
      throw; \
    } \
  } \
  if (s_exit_on_next) { \
    s_exit_on_next = false; \
    throw ExitException (0); \
  }

// -------------------------------------------------------------------
//  Utilities to generate C++ exceptions from Ruby exceptions

static BacktraceElement 
rba_split_bt_information (const char *m, size_t l)
{
  for (size_t i = 0; i + 1 < l; ++i) {

    if (m[i] == ':' && isdigit (m[i + 1])) {

      size_t j = i + 1;
      int line = 0;
      while (j < l && isdigit (m[j])) {
        line = (line * 10) + (int)(m[j] - '0');
        ++j;
      }

      std::string file;
      if (line > 0) {
        file = std::string (m, i);
      }
      if (j == l) {
        return BacktraceElement (file, line);
      } else if (m[j] == ':') {
        return BacktraceElement (file, line, std::string (m, j + 1, l - (j + 1)));
      }

    }

  }

  return BacktraceElement ();
}

static void
rba_get_backtrace_from_array (VALUE backtrace, std::vector<BacktraceElement> &bt, unsigned int skip)
{
  if (TYPE (backtrace) == T_ARRAY) {
    unsigned int len = RARRAY_LEN(backtrace);
    VALUE *el = RARRAY_PTR(backtrace);
    bt.reserve (bt.size () + (len >= skip ? len - skip : 0));
    while (len-- > 0) {
      if (skip > 0) {
        ++el;
        --skip;
      } else {
        VALUE str = StringValue (*el++);
        bt.push_back (rba_split_bt_information (RSTRING_PTR(str), RSTRING_LEN(str)));
      }
    }
  }
}

static void
rba_check_error (unsigned int drop = 0)
{
  VALUE lasterr = rb_errinfo ();
  if (lasterr != Qnil) {

    if (CLASS_OF (lasterr) == rb_eSystemExit) {
      int status = NUM2INT (rb_funcall (lasterr, rb_intern ("status"), 0));
      throw ExitException (status);
    }

    VALUE klass = rb_class_path (CLASS_OF (lasterr));
    std::string eclass = std::string (RSTRING_PTR(klass), RSTRING_LEN(klass));
    VALUE message = rb_obj_as_string(lasterr);
    std::string emsg = std::string (RSTRING_PTR(message), RSTRING_LEN(message));

    std::vector <BacktraceElement> bt;
    rba_get_backtrace_from_array (rb_funcall (lasterr, rb_intern ("backtrace"), 0), bt, 0);
    while (! bt.empty () && drop-- > 0) {
      bt.pop_back ();
    }

    if (! s_debugger_scope.empty ()) {
      for (size_t i = 0; i < bt.size (); ++i) {
        if (bt [i].file == s_debugger_scope) {
          bt.erase (bt.begin (), bt.begin () + i);
          break;
        }
      }
      s_debugger_scope.clear ();
    }

    //  parse the backtrace to get the line number
    BacktraceElement info;
    if (! bt.empty ()) {
      //  use the backtrace
      info = bt.front ();
    } else {
      //  use the error message
      info = rba_split_bt_information (emsg.c_str (), emsg.size ());
    }

    if (info.line > 0) {
      throw RubyError (emsg.c_str (), info.file.c_str (), info.line, eclass.c_str (), bt);
    } else {
      throw RubyError (emsg.c_str (), eclass.c_str (), bt);
    }
  }
}

/**
 *  @brief needed because StringValue is a macro:
 */
static VALUE
rba_string_value_f (VALUE obj)
{
  return StringValue (obj);
}

/**
 *  @brief string value retrieval with check 
 */
static VALUE
rba_safe_string_value (VALUE obj)
{
  VALUE ret;

  RUBY_BEGIN_EXEC 

    rb_set_errinfo (Qnil);
    int error = 0;
    ret = rb_protect (rba_string_value_f, obj, &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC

  return ret;
}

/**
 *  @brief object to string with check 
 */
static VALUE
rba_safe_obj_as_string (VALUE obj)
{
  VALUE ret = Qnil;

  RUBY_BEGIN_EXEC 

    rb_set_errinfo (Qnil);
    int error = 0;
    ret = rb_protect (rb_obj_as_string, obj, &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC

  return ret;
}

/**
 *  @brief extract protected flag from the method name
 */
static const char *
extract_protected (const char *name, bool &prot)
{
  if (*name == '*' && name[1] && name[1] != '=') {
    ++name;
    prot = true;
  } else {
    prot = false;
  }
  return name;
}

/**
 *  @brief Find the class declaration from the Ruby object
 */
const gsi::ClassBase *
find_cclass (VALUE k)
{
  std::map <VALUE, const gsi::ClassBase *>::const_iterator cls;

  //  find the class that is bound to C++ (maybe a super class)
  while (k != rb_cObject) { 
    cls = cls_map.find (k);
    if (cls == cls_map.end ()) {
      //  if not found advance to super class
      k = RCLASS_SUPER (k);
    } else {
      break;
    }
  }

  tl_assert (cls != cls_map.end ());
  return cls->second;
}

// -------------------------------------------------------------------
//  Some utilities for handling Ruby errors

/**
 *  @brief Initialize the interpreter
 *
 *  It was learned (by bitter experience), that in particular rb_protect cannot simply be 
 *  called but requires some safety measures, at least in Ruby 1.8.5 as possibly 1.8.x. 
 *  What happens in rb_protect is that a RNode object is instantiated in some kind
 *  of compile steps that builds a temporary executable statement. Deep inside this
 *  functionality, the rb_sourcefile () variable is copied into the RNode object, as this
 *  variable is usually updated with the current sourcefile name. 
 *  In effect, the string referenced by this variable participates in the GC mark&sweep
 *  steps which leads to unpredicable results, if this variable is not set to a valid 
 *  string (ruby) buffer or 0.
 *
 *  As a consequence, this function must be called before rb_protect and likely other xx_protect
 *  derivatives.
 */
static void rb_protect_init ()
{
#if HAVE_RUBY_VERSION_CODE<10900
  ruby_sourcefile = 0;
  ruby_sourceline = 0;
#endif
}

//  A protect wrapper for rb_funcall2

struct rb_funcall2_params
{
  VALUE obj;
  ID id;
  int argc;
  VALUE *args;
};

static VALUE
rb_funcall2_wrap (VALUE args)
{
  //  HINT: this ugly cast is required since there is only one form of rb_protect
  const rb_funcall2_params *p = (const rb_funcall2_params *) args;
  return rb_funcall2 (p->obj, p->id, p->argc, p->args);
}

static VALUE rb_funcall2_checked (VALUE obj, ID id, int argc, VALUE *args)
{
#if HAVE_RUBY_VERSION_CODE>=10900
  //  Hint: calling of methods on terminated objects cannot really be avoided in all cases - 
  //  for example when the destructor triggers some callback as is the case in Qt events
  //  (i.e. childEvent is triggered when a child is removed and may happen on a parent which
  //  is terminated but not destroyed yet). To avoid problems in that case we simply ignore the 
  //  call.
  if (TYPE (obj) == T_ZOMBIE) {
    return Qnil;
  }
#endif

  VALUE ret = Qnil;

  RUBY_BEGIN_EXEC

    rb_funcall2_params p;
    p.obj  = obj;
    p.id   = id;
    p.argc = argc;
    p.args = args;

    int error = 0;
    //  HINT: the ugly (VALUE) cast is required since there is only one form of rb_protect
    rb_protect_init (); // see above
    ret = rb_protect (&rb_funcall2_wrap, (VALUE) &p, &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC

  return ret;
}

//  A protect wrapper for calling "eval" on an object

struct rb_f_eval_params
{
  int argc;
  VALUE *argv;
  VALUE self;
};

static VALUE
rb_f_eval_wrap (VALUE args)
{
  //  HINT: this ugly cast is required since there is only one form of rb_protect
  const rb_f_eval_params *p = (const rb_f_eval_params *) args;
  return rb_funcall2 (p->self, rb_intern ("eval"), p->argc, p->argv);
}

static VALUE 
rb_f_eval_checked (int argc, VALUE *argv, VALUE self)
{
  VALUE ret = Qnil;

  RUBY_BEGIN_EXEC 

    rb_f_eval_params p;
    p.argc = argc;
    p.argv = argv;
    p.self = self;

    int error = 0;
    //  HINT: the ugly (VALUE) cast is required since there is only one form of rb_protect
    rb_protect_init (); // see above
    ret = rb_protect (&rb_f_eval_wrap, (VALUE) &p, &error);
    if (error) {
      //  drop the eval call inside rb_f_eval_wrap from the stack, hence the "1" parameter
      rba_check_error (1);
    }

  RUBY_END_EXEC

  return ret;
}

static VALUE do_eval_string (const char *expr, const char *file, int line)
{
  rb_set_errinfo (Qnil);

  if (file) {
    set_progname (rb_str_new (file, strlen (file)));
  } else {
    const char *e = "-e";
    set_progname (rb_str_new (e, strlen (e)));
  }

  int argc;
  VALUE args[4];
  args[0] = rb_str_new (expr, strlen (expr));
  //  use the current binding if there is one. This allows to eval in the context of a current trace callback
  //  when eval is called from the trace handler.
  if (! s_in_trace) {
    args[1] = rb_const_get(rb_cObject, rb_intern("TOPLEVEL_BINDING")); // taken from ruby.c
  } else {
    if (s_current_binding == Qnil) {
      s_current_binding = rb_binding_new ();
      rb_gc_register_address (&s_current_binding);
    }
    args[1] = s_current_binding;
  }
  if (file) {
    args[2] = rb_str_new (file, strlen (file));
    args[3] = INT2NUM(line);
    argc = 4;
  } else {
    argc = 2;
  }

  return rb_f_eval_checked (argc, args, get_top_self ());
}

//  A protect wrapper for rb_class_new_instance

struct rb_class_new_instance_param
{
  int argc;
  VALUE *argv;
  VALUE klass;
};

static VALUE
rb_class_new_instance_wrap (VALUE args)
{
  //  HINT: this ugly cast is required since there is only one form of rb_protect
  const rb_class_new_instance_param *p = (const rb_class_new_instance_param *) args;
  return rb_class_new_instance (p->argc, p->argv, p->klass);
}

static VALUE 
rb_class_new_instance_checked (int argc, VALUE *argv, VALUE klass)
{
  VALUE ret = Qnil;

  RUBY_BEGIN_EXEC 

    rb_class_new_instance_param p;
    p.argc = argc;
    p.argv = argv;
    p.klass = klass;

    int error = 0;
    //  HINT: the ugly (VALUE) cast is required since there is only one form of rb_protect
    rb_protect_init (); // see above
    ret = rb_protect (&rb_class_new_instance_wrap, (VALUE) &p, &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC

  return ret;
}

/**
 *  @brief Gets the class as a string
 */
std::string 
rb_class_name (VALUE self)
{
  if (TYPE (self) != T_CLASS) {
    self = rb_class_of (self);
  }

  VALUE str = rb_obj_as_string (self);
  return std::string (RSTRING_PTR (str), RSTRING_LEN (str));
}

// -------------------------------------------------------------------
//  The GSI object proxy stored within the Ruby object

class Proxy 
  : public gsi::Callee
{
public:
  struct CallbackFunction
  {
    CallbackFunction (ID id, VALUE o, const gsi::MethodBase *m)
      : method_id (id), rb_obj (o), method (m)
    {
      //  .. nothing yet ..
    }

    ID method_id;
    VALUE rb_obj;
    const gsi::MethodBase *method;
  };

  Proxy (const gsi::ClassBase *_cls_decl)
    : m_cls_decl (_cls_decl), m_obj (0), m_owned (false), m_const_ref (false), m_destroyed (false), m_can_destroy (false)
  { }

  ~Proxy ();

  void destroy ();
  void set (void *obj, bool owned, bool const_ref, bool can_destroy, VALUE self);
  void detach ();

  const gsi::ClassBase *cls_decl () const 
  {
    return m_cls_decl;
  }

  bool destroyed () const 
  {
    return m_destroyed;
  }

  bool const_ref () const 
  {
    return m_const_ref;
  }

  void set_const_ref (bool c) 
  {
    m_const_ref = c;
  }

  void *obj (VALUE self);

  bool owned () const 
  {
    return m_owned;
  }

  int add_callback (const CallbackFunction &vf)
  {
    m_cbfuncs.push_back (vf);
    return int (m_cbfuncs.size () - 1);
  }

  void clear_callbacks ()
  {
    m_cbfuncs.clear ();
  }

  void register_event_proc (const gsi::MethodBase *meth, VALUE proc);

  virtual void call (int id, gsi::SerialArgs &args, gsi::SerialArgs &ret);

private:
  const gsi::ClassBase *m_cls_decl;
  void *m_obj;
  bool m_owned : 1;
  bool m_const_ref : 1;
  bool m_destroyed : 1;
  bool m_can_destroy : 1;

  std::vector<CallbackFunction> m_cbfuncs; 
  std::map<const gsi::MethodBase *, VALUE> m_proc_table;
};

// --------------------------------------------------------------------------
//  Ruby client data 

/**
 *  @brief The Ruby data structure representing the link from a C++ object to the Ruby object
 */
class RBAClientData
{
public:
  /**
   *  @brief Default constructor
   */
  RBAClientData ()
    : m_self (Qnil), m_locked (0)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Destructor
   */
  ~RBAClientData () 
  {
    //  once the object is destroyed, we may release the Ruby reference
    if (m_locked > 0) {
      rb_gc_unregister_address (&m_self);
    }

    if (m_self != Qnil) {
      //  Hint: we cannot use Data_Get_Struct here since this one checks the type and
      //  in some cases (i.e. Qt's recursive destruction of objects already marked in the GC),
      //  the type is T_ZOMBIE
      Proxy *p = (Proxy *) DATA_PTR (m_self);
      p->detach ();
    }

    m_self = Qnil;
    m_locked = 0;
  }

  /**
   *  @brief Resets the Ruby reference to nil
   */
  void reset_ref ()
  {
    //  release the object reference if required
    if (m_locked > 0) {
      rb_gc_unregister_address (&m_self);
    }

    m_self = Qnil;
    m_locked = 0;
  }

  /**
   *  @brief Sets the Ruby reference
   *
   *  @arg s The Ruby reference
   *  @arg lock true, if the Ruby object should be locked, i.e. a reference should be kept
   *
   *  Using lock = true is equivalent to call keep after set_ref
   */
  void set_ref (VALUE s, bool lock)
  {
    tl_assert (m_self == Qnil);
    m_self = s;

    //  this avoids that our Ruby reference is deleted by the GC
    if (lock) {
      keep ();
    }
  }

  /**
   *  @brief Gets the Ruby reference
   */
  VALUE get_ref () const
  {
    return m_self;
  }

  /**
   *  @brief Release the Ruby reference
   *
   *  After having released the ruby references, the GC will destroy the object.
   *  keep() and release() do reference counting, i.e. the object is released when release() is called
   *  as many times as keep().
   */
  void release () 
  {
    if (m_locked > 0) {
      --m_locked;
      if (m_locked == 0) {
        //  make the GC remove the proxy if no longer used
        rb_gc_unregister_address (&m_self);
      }
    }
  }

  /**
   *  @brief Keep a Ruby reference
   *
   *  After calling keep(), the GC will no longer destroy the object. keep() can be 
   *  cancelled with release().
   */
  void keep () 
  {
    if (m_locked == 0) {
      //  make the GC not remove the proxy if no longer used
      rb_gc_register_address (&m_self);
    }
    ++m_locked;
  }

  /**
   *  @brief Returns true, if the object is locked.
   */
  bool locked () const
  {
    return m_locked > 0;
  }

private:
  VALUE m_self;
  int m_locked;
};

// --------------------------------------------------------------------------
//  Ruby client data handler

class RBAClientHandler
  : public gsi::ClientHandler
{
public:
  virtual void attach (void *&data) 
  {
    data = new RBAClientData ();
  }

  virtual void detach (void *&data) 
  {
    void *o = data;
    data = 0;
    delete (RBAClientData *) o;
  }

  virtual void release (void *&data)
  {
    if (data) {
      ((RBAClientData *) data)->release ();
    }
  }

  virtual void keep (void *&data)
  {
    if (data) {
      ((RBAClientData *) data)->keep ();
    }
  }
};

// -------------------------------------------------------------------
//  Type checks 
//
//  The type checks are somewhat more picky than the ruby2c functions.
//  The type checks are used to resolve overridden methods, so being picky
//  might be more appropriate.

template <class T>
struct test_type;

template <>
struct test_type<bool>
{
  static bool test (VALUE rval, bool loose)
  {
    if (loose) {
      return true;  // everything can be converted to bool
    } else {
      unsigned int t = TYPE (rval);
      return (t == T_FALSE || t == T_TRUE || t == T_NIL);
    }
  }
};

//  used for other integer types as well:
template <>
struct test_type<int>
{
  static bool test (VALUE rval, bool loose)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FIXNUM || t == T_BIGNUM || (loose && t == T_FLOAT));
  }
};

template <>
struct test_type<char>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<signed char>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<unsigned char>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<short>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<unsigned short>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<unsigned int>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<long>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<unsigned long>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<long long>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

template <>
struct test_type<unsigned long long>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};

#if defined(HAVE_64BIT_COORD)
template <>
struct test_type<__int128>
{
  static bool test (VALUE rval, bool loose) { return test_type<int>::test (rval, loose); }
};
#endif

template <>
struct test_type<double>
{
  static bool test (VALUE rval, bool loose)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FLOAT || (loose && (t == T_FIXNUM || t == T_BIGNUM)));
  }
};

template <>
struct test_type<float>
{
  static bool test (VALUE rval, bool loose) { return test_type<double>::test (rval, loose); }
};

template <>
struct test_type<void *>
{
  static bool test (VALUE rval, bool loose)
  {
    return test_type<size_t>::test (rval, loose);
  }
};

//  used for strings in general:
template <>
struct test_type<const char *>
{
  static bool test (VALUE rval, bool loose)
  {
    return TYPE (rval) == T_STRING;
  }
};

template <>
struct test_type<std::string>
{
  static bool test (VALUE rval, bool loose) { return test_type<const char *>::test (rval, loose); }
};

template <>
struct test_type<QString>
{
  static bool test (VALUE rval, bool loose) { return test_type<const char *>::test (rval, loose); }
};

template <>
struct test_type<tl::Variant>
{
  static bool test (VALUE rval, bool loose)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FALSE || t == T_TRUE || t == T_NIL || t == T_FIXNUM || t == T_BIGNUM || t == T_FLOAT || t == T_STRING || t == T_ARRAY /*be optimistic*/);
  }
};

// -------------------------------------------------------------------
//  Ruby to C conversion

template <class T>
struct ruby2c;

template <>
struct ruby2c<bool>
{
  static bool get (VALUE rval)
  {
    return bool (RTEST (rval));
  }
};

template <>
struct ruby2c<char>
{
  static char get (VALUE rval)
  {
    return NUM2INT (rval);
  }
};

template <>
struct ruby2c<signed char>
{
  static signed char get (VALUE rval)
  {
    return NUM2INT (rval);
  }
};

template <>
struct ruby2c<unsigned char>
{
  static unsigned char get (VALUE rval)
  {
    return NUM2UINT (rval);
  }
};

template <>
struct ruby2c<short>
{
  static short get (VALUE rval)
  {
    return NUM2INT (rval);
  }
};

template <>
struct ruby2c<unsigned short>
{
  static unsigned short get (VALUE rval)
  {
    return NUM2UINT (rval);
  }
};

template <>
struct ruby2c<int>
{
  static int get (VALUE rval)
  {
    return NUM2INT (rval);
  }
};

template <>
struct ruby2c<unsigned int>
{
  static unsigned int get (VALUE rval)
  {
    return NUM2UINT (rval);
  }
};

template <>
struct ruby2c<long>
{
  static long get (VALUE rval)
  {
    return NUM2LONG (rval);
  }
};

template <>
struct ruby2c<unsigned long>
{
  static unsigned long get (VALUE rval)
  {
    return NUM2ULONG (rval);
  }
};

template <>
struct ruby2c<long long>
{
  static long long get (VALUE rval)
  {
    return NUM2LL (rval);
  }
};

template <>
struct ruby2c<unsigned long long>
{
  static unsigned long long get (VALUE rval)
  {
    return NUM2ULL (rval);
  }
};

#if defined(HAVE_64BIT_COORD)
template <>
struct ruby2c<__int128>
{
  static __int128 get (VALUE rval)
  {
    // TOOD: this is pretty simplistic
    return NUM2DBL (rval);
  }
};
#endif

template <>
struct ruby2c<double>
{
  static double get (VALUE rval)
  {
    return NUM2DBL (rval);
  }
};

template <>
struct ruby2c<float>
{
  static float get (VALUE rval)
  {
    return float (NUM2DBL (rval));
  }
};

template <>
struct ruby2c<std::string>
{
  static std::string get (VALUE rval)
  {
    VALUE str = rba_safe_string_value (rval);
    return std::string (RSTRING_PTR(str), RSTRING_LEN(str));
  }
};

template <>
struct ruby2c<QString>
{
  static QString get (VALUE rval)
  {
    VALUE str = rba_safe_string_value (rval);
    return tl::to_qstring (std::string (RSTRING_PTR(str), RSTRING_LEN(str)));
  }
};

template <>
struct ruby2c<void *>
{
  static void *get (VALUE rval)
  {
    return (void *) ruby2c<size_t>::get (rval);
  }
};

template <>
struct ruby2c<const char *>
{
  static const char *get (VALUE rval)
  {
    VALUE str = rba_safe_string_value (rval);
    return RSTRING_PTR(str);
  }
};

template <>
struct ruby2c<tl::Variant>
{
  static tl::Variant get (VALUE rval)
  {
    if (FIXNUM_P (rval)) {
      return tl::Variant (ruby2c<long>::get (rval));
    } else if (rval == Qnil) {
      return tl::Variant ();
    } else if (rval == Qfalse) {
      return tl::Variant (false);
    } else if (rval == Qtrue) {
      return tl::Variant (true);
    } else if (TYPE (rval) == T_BIGNUM) {
      return tl::Variant (ruby2c<long long>::get (rval));
    } else if (TYPE (rval) == T_FLOAT) {
      return tl::Variant (ruby2c<double>::get (rval));
    } else if (TYPE (rval) == T_ARRAY) {

      unsigned int len = RARRAY_LEN(rval);
      VALUE *el = RARRAY_PTR(rval);

      static std::vector<tl::Variant> empty;
      tl::Variant r (empty.begin (), empty.end ());
      r.get_list ().reserve (len);
      while (len-- > 0) {
        r.get_list ().push_back (ruby2c<tl::Variant>::get (*el++));
      }
      return r;

    } else if (TYPE (rval) == T_DATA) {

      //  some types are supported through "complex" tl::Variant's
      Proxy *p;
      Data_Get_Struct (rval, Proxy, p);

      //  employ the tl::Variant binding capabilities of the Expression binding to derive the 
      //  variant value. Currently we use deep copy mode always since any other solution would
      //  involve reference management.
      return tl::Variant (0, p->cls_decl ()->clone (p->obj (rval)), p->cls_decl ()->var_cls (false, false));

    } else {
      return tl::Variant (ruby2c<const char *>::get (rval));
    }
  }
};

// -------------------------------------------------------------------
//  Return the boxed value pointer for a given basic type from the reference

struct GetBoxValueFunc
{
  GetBoxValueFunc (VALUE arg, std::list<gsi::TempObject> *tmp_stack)
    : m_arg (arg), mp_tmp_stack (tmp_stack), mp_ret (0)
  { }

  void *value () 
  { 
    return mp_ret; 
  }

  template <class R> void op () 
  { 
    if (TYPE (m_arg) != T_DATA) {

      mp_tmp_stack->push_back (gsi::TempObject ());
      R *v = new R (ruby2c<R>::get (m_arg));
      mp_tmp_stack->back ().set (v);
      mp_ret = v;

    } else {

      const gsi::ClassBase *bt = gsi::cls_decl <gsi::Value> ();

      Proxy *p;
      Data_Get_Struct (m_arg, Proxy, p);
      if (!p->cls_decl ()->is_derived_from (bt)) {
        throw tl::Exception (tl::sprintf (tl::translate ("Passing an object to pointer or reference requires a boxed type (RBA::%s)"), bt->name ()));
      }

      gsi::Value *bo = reinterpret_cast<gsi::Value *> (p->obj (m_arg));
      if (bo) {
        mp_ret = bo->value ().template morph<R> ().native_ptr ();
      }

    }
  }

private:
  VALUE m_arg;
  std::list<gsi::TempObject> *mp_tmp_stack;
  void *mp_ret;
};

void *boxed_value_ptr (gsi::BasicType type, VALUE arg, std::list<gsi::TempObject> &tmp_stack)
{
  GetBoxValueFunc f (arg, &tmp_stack);
  switch_type (type, f);
  return f.value ();
}

// -------------------------------------------------------------------

/**
 *  @brief A serialization wrapper (write mode)
 *
 *  The generic class is for POD objects.
 */
template <class R>
struct set_value
{
  static void set (gsi::SerialArgs &aa, VALUE arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg == Qnil) {

      if (atype.is_ref || atype.is_cref) {
        throw tl::Exception (tl::translate ("Arguments or return values of reference type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (typename gsi::type_traits<R *>::tag (), (R *)0);
      } else if (atype.is_cptr) {
        aa.set_value (typename gsi::type_traits<const R *>::tag (), (const R *)0);
      } else {
        aa.set_value (typename gsi::type_traits<R>::tag (), (R)0);
      }

    } else {

      if (atype.is_ref || atype.is_ptr) {
        // references or pointers require a boxed object. Pointers also allow nil.
        void *vc = boxed_value_ptr (atype.type, arg, tmp_stack);
        if (! vc && atype.is_ref) {
          throw tl::Exception (tl::translate ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object"));
        }
        aa.set_value (typename gsi::type_traits<void *>::tag (), vc);
      } else if (atype.is_cref) {
        aa.set_value (typename gsi::type_traits<const R &>::tag (), ruby2c<R>::get (arg));
      } else if (atype.is_cptr) {
        R r = ruby2c<R>::get (arg);
        aa.set_value (typename gsi::type_traits<const R *>::tag (), &r);
      } else {
        aa.set_value (typename gsi::type_traits<R>::tag (), ruby2c<R>::get (arg));
      }

    }
  }
};

/**
 *  @brief A serialization wrapper (write mode)
 *
 *  Specialisation for strings 
 */
//  TODO: simplify all NPOD types ...
template <>
struct set_value <std::string>
{
  static void set (gsi::SerialArgs &aa, VALUE arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg == Qnil) {

      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments or return values of reference or direct type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (gsi::type_traits<std::string *>::tag (), (std::string *)0);
      } else {
        aa.set_value (gsi::type_traits<const std::string *>::tag (), (const std::string *)0);
      }

    } else {

      if (atype.is_ref || atype.is_ptr) {
        // references or pointers require a boxed object. Pointers also allow nil.
        void *vc = boxed_value_ptr (atype.type, arg, tmp_stack);
        if (! vc && atype.is_ref) {
          throw tl::Exception (tl::translate ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object"));
        }
        aa.set_value (gsi::type_traits<void *>::tag (), vc);

      } else if (atype.is_cref || atype.is_cptr) {

        VALUE str = rba_safe_string_value (arg);
        tmp_stack.push_back (gsi::TempObject ());
        std::string *v = new std::string (RSTRING_PTR(str), RSTRING_LEN(str));
        tmp_stack.back ().set (v);

        if (atype.is_cref) {
          aa.set_value (gsi::type_traits<const std::string &>::tag (), *v);
        } else if (atype.is_cptr) {
          aa.set_value (gsi::type_traits<const std::string *>::tag (), v);
        }

      } else {
        VALUE str = rba_safe_string_value (arg);
        aa.set_value (gsi::type_traits<std::string>::tag (), std::string (RSTRING_PTR(str), RSTRING_LEN(str)));
      }

    }
  }
};

/**
 *  @brief A serialization wrapper (write mode)
 *
 *  Specialisation for QString's 
 */
template <>
struct set_value <QString>
{
  static void set (gsi::SerialArgs &aa, VALUE arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg == Qnil) {

      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments or return values of reference or direct type cannot be passed nil"));
      } else if (atype.is_ptr) {
        aa.set_value (gsi::type_traits<std::string *>::tag (), (std::string *)0);
      } else {
        aa.set_value (gsi::type_traits<const std::string *>::tag (), (const std::string *)0);
      }

    } else {

      if (atype.is_ref || atype.is_ptr) {
        // references or pointers require a boxed object. Pointers also allow nil.
        void *vc = boxed_value_ptr (atype.type, arg, tmp_stack);
        if (! vc && atype.is_ref) {
          throw tl::Exception (tl::translate ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object"));
        }
        aa.set_value (gsi::type_traits<void *>::tag (), vc);

      } else if (atype.is_cref || atype.is_cptr) {

        VALUE str = rba_safe_string_value (arg);
        tmp_stack.push_back (gsi::TempObject ());
        QString *v = new QString (tl::to_qstring (std::string (RSTRING_PTR(str), RSTRING_LEN(str))));
        tmp_stack.back ().set (v);

        if (atype.is_cref) {
          aa.set_value (gsi::type_traits<const QString &>::tag (), *v);
        } else if (atype.is_cptr) {
          aa.set_value (gsi::type_traits<const QString *>::tag (), v);
        }

      } else {
        VALUE str = rba_safe_string_value (arg);
        aa.set_value (gsi::type_traits<QString>::tag (), QString (tl::to_qstring (std::string (RSTRING_PTR(str), RSTRING_LEN(str)))));
      }

    }
  }
};

/**
 *  @brief A serialization wrapper (write mode)
 *
 *  Specialisation for variants 
 */
template <>
struct set_value <tl::Variant>
{
  static void set (gsi::SerialArgs &aa, VALUE arg, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (arg == Qnil && (atype.is_ptr || atype.is_cptr)) {

      if (atype.is_ptr) {
        aa.set_value (gsi::type_traits<tl::Variant *>::tag (), (tl::Variant *)0);
      } else {
        aa.set_value (gsi::type_traits<const tl::Variant *>::tag (), (const tl::Variant *)0);
      }

    } else {

      if (atype.is_cref || atype.is_ref || atype.is_cptr || atype.is_ptr) {

        tmp_stack.push_back (gsi::TempObject ());
        tl::Variant *v = new tl::Variant (ruby2c<tl::Variant>::get (arg));
        tmp_stack.back ().set (v);

        if (atype.is_ref) {
          aa.set_value (gsi::type_traits<tl::Variant &>::tag (), *v);
        } else if (atype.is_cref) {
          aa.set_value (gsi::type_traits<const tl::Variant &>::tag (), *v);
        } else if (atype.is_ptr) {
          aa.set_value (gsi::type_traits<tl::Variant *>::tag (), v);
        } else if (atype.is_cptr) {
          aa.set_value (gsi::type_traits<const tl::Variant *>::tag (), v);
        }

      } else {
        aa.set_value (gsi::type_traits<tl::Variant>::tag (), ruby2c<tl::Variant>::get (arg));
      }

    }
  }
};

/**
 *  @brief Test a VALUE for compatibility with a vector of the given type R
 */
template <class R>
struct test_vector
{
  static bool test (VALUE arr, bool loose)
  {
    if (TYPE (arr) != T_ARRAY) {
      return false;
    }

    unsigned int len = RARRAY_LEN(arr);
    VALUE *el = RARRAY_PTR(arr);
    while (len-- > 0) {
      if (! test_type<R>::test (*el++, loose)) {
        return false;
      }
    }

    return true;
  }
};

/**
 *  @brief A serialization wrapper (write mode)
 *
 *  Specialisation for vectors of any kind
 */
template <class R>
struct set_vector
{
  static void set (gsi::SerialArgs &aa, VALUE arr, const gsi::ArgType &atype, std::list<gsi::TempObject> &tmp_stack)
  {
    if (TYPE (arr) != T_ARRAY) {
      throw tl::Exception (tl::sprintf (tl::translate ("Expected array either as argument or return type")));
    }
    unsigned int len = RARRAY_LEN(arr);
    VALUE *el = RARRAY_PTR(arr);

    if (atype.is_cref || atype.is_ref || atype.is_cptr || atype.is_ptr) {

      tmp_stack.push_back (gsi::TempObject ());

      std::vector<R> *v = new std::vector<R> ();
      v->reserve (len);
      while (len-- > 0) {
        v->push_back (ruby2c<R>::get (*el++));
      }
      tmp_stack.back ().set (v);

      if (atype.is_ref) {
        aa.set_value (typename gsi::type_traits<std::vector<R> &>::tag (), *v);
      } else if (atype.is_cref) {
        aa.set_value (typename gsi::type_traits<const std::vector<R> &>::tag (), *v);
      } else if (atype.is_ptr) {
        aa.set_value (typename gsi::type_traits<std::vector<R> *>::tag (), v);
      } else if (atype.is_cptr) {
        aa.set_value (typename gsi::type_traits<const std::vector<R> *>::tag (), v);
      }

    } else {
      std::vector<R> *v = aa.set_value (typename gsi::type_traits< std::vector<R> >::tag (), std::vector<R> ());
      v->reserve (len);
      while (len-- > 0) {
        v->push_back (ruby2c<R>::get (*el++));
      }
    }
  }
};

// -------------------------------------------------------------------
//  C to Ruby conversion

template <class R>
struct c2ruby;

template <>
struct c2ruby<bool>
{
  static VALUE get (bool c)
  {
    return c ? Qtrue : Qfalse;
  }
};

template <>
struct c2ruby<char>
{
  static VALUE get (char c)
  {
    return INT2NUM (c);
  }
};

template <>
struct c2ruby<signed char>
{
  static VALUE get (signed char c)
  {
    return UINT2NUM (c);
  }
};

template <>
struct c2ruby<unsigned char>
{
  static VALUE get (unsigned char c)
  {
    return UINT2NUM (c);
  }
};

template <>
struct c2ruby<short>
{
  static VALUE get (short c)
  {
    return INT2NUM (c);
  }
};

template <>
struct c2ruby<unsigned short>
{
  static VALUE get (unsigned short c)
  {
    return UINT2NUM (c);
  }
};

template <>
struct c2ruby<int>
{
  static VALUE get (int c)
  {
    return INT2NUM (c);
  }
};

template <>
struct c2ruby<unsigned int>
{
  static VALUE get (unsigned int c)
  {
    return UINT2NUM (c);
  }
};

template <>
struct c2ruby<long>
{
  static VALUE get (long c)
  {
    return LONG2NUM (c);
  }
};

template <>
struct c2ruby<unsigned long>
{
  static VALUE get (unsigned long c)
  {
    return ULONG2NUM (c);
  }
};

template <>
struct c2ruby<long long>
{
  static VALUE get (long long c)
  {
    return LL2NUM (c);
  }
};

template <>
struct c2ruby<unsigned long long>
{
  static VALUE get (unsigned long long c)
  {
    return ULL2NUM (c);
  }
};

#if defined(HAVE_64BIT_COORD)
template <>
struct c2ruby<__int128>
{
  static VALUE get (__int128 c)
  {
    // TOOD: this is pretty simplistic
    return rb_float_new (double (c));
  }
};
#endif

template <>
struct c2ruby<double>
{
  static VALUE get (double c)
  {
    return rb_float_new (c);
  }
};

template <>
struct c2ruby<float>
{
  static VALUE get (float c)
  {
    return rb_float_new (c);
  }
};

template <>
struct c2ruby<std::string>
{
  static VALUE get (const std::string &c)
  {
    return rb_str_new (c.c_str (), c.size ());
  }
};

template <>
struct c2ruby<QString>
{
  static VALUE get (const QString &qs)
  {
    if (qs.isNull ()) {
      return Qnil;
    } else {
      std::string c (tl::to_string (qs));
      return rb_str_new (c.c_str (), c.size ());
    }
  }
};

template <>
struct c2ruby<void *>
{
  static VALUE get (void *s)
  {
    return c2ruby<size_t>::get (size_t (s));
  }
};

template <>
struct c2ruby<tl::Variant>
{
  static VALUE get (const tl::Variant &c)
  {
    if (c.is_double ()) {
      return c2ruby<double>::get (c.to_double ());
    } else if (c.is_float ()) {
      return c2ruby<float>::get (c.to_double ());
    } else if (c.is_bool ()) {
      return c2ruby<bool>::get (c.to_bool ());
    } else if (c.is_a_string ()) {
      return c2ruby<std::string>::get (c.to_string ());
    } else if (c.is_long () || c.is_int () || c.is_short () || c.is_char () || c.is_schar ()) {
      return c2ruby<long>::get (c.to_long ());
    } else if (c.is_ulong () || c.is_uint () || c.is_ushort () || c.is_uchar ()) {
      return c2ruby<unsigned long>::get (c.to_ulong ());
    } else if (c.is_longlong ()) {
      return c2ruby<long long>::get (c.to_longlong ());
    } else if (c.is_ulonglong ()) {
      return c2ruby<unsigned long long>::get (c.to_ulonglong ());
    } else if (c.is_list ()) {
      VALUE ret = rb_ary_new ();
      for (std::vector<tl::Variant>::const_iterator i = c.get_list ().begin (); i != c.get_list ().end (); ++i) {
        rb_ary_push (ret, c2ruby<tl::Variant>::get (*i));
      }
      return ret;
    } else if (c.is_user ()) {

      const gsi::ClassBase *cls = c.gsi_cls ();
      if (cls) {

        void *obj = cls->clone (c.to_user_object ());

        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (cls);
        tl_assert (c != rev_cls_map.end ());
        VALUE ret = rb_class_new_instance_checked (0, 0, c->second);
        Proxy *p;
        Data_Get_Struct (ret, Proxy, p);
        p->set (obj, true, false, false, ret);
        return ret;

      } else {
        //  not a known type -> return nil
        return Qnil;
      }

    } else {
      return Qnil;
    }
  }
};

template <>
struct c2ruby<const char *>
{
  static VALUE get (const char *s)
  {
    if (! s) {
      s = "(null)";
    }
    return rb_str_new (s, strlen (s));
  }
};

/**
 *  @brief Deseralisation wrapper 
 *
 *  The default implementation is for POD types, strings and variants
 */
template <class R> 
struct get_value
{
  static VALUE get (gsi::SerialArgs &rr, const gsi::ArgType &arg)
  {
    if (arg.is_ref) {
      return c2ruby<R>::get (rr.template get_value<R &> (typename gsi::type_traits<R &>::tag ()));
    } else if (arg.is_cref) {
      return c2ruby<R>::get (rr.template get_value<const R &> (typename gsi::type_traits<const R &>::tag ()));
    } else if (arg.is_ptr) {
      R *p = rr.template get_value<R *> (typename gsi::type_traits<R *>::tag ());
      if (p) {
        return c2ruby<R>::get (*p);
      } else {
        return Qnil;
      }
    } else if (arg.is_cptr) {
      const R *p = rr.template get_value<const R *> (typename gsi::type_traits<const R *>::tag ());
      if (p) {
        return c2ruby<R>::get (*p);
      } else {
        return Qnil;
      }
    } else {
      return c2ruby<R>::get (rr.template get_value<R> (typename gsi::type_traits<R>::tag ()));
    }
  }
};

/**
 *  @brief Deseralisation wrapper: specialization for const char *
 *
 *  Without that would would have to handle void *&, void * const &, ...
 *  TODO: right now these types are not supported.
 */
template <> 
struct get_value<void *>
{
  static VALUE get (gsi::SerialArgs &rr, const gsi::ArgType &arg)
  {
    tl_assert (! arg.is_cref);
    tl_assert (! arg.is_ref);
    tl_assert (! arg.is_cptr);
    tl_assert (! arg.is_ptr);
    return c2ruby<void *>::get (rr.get_value<void *> (gsi::type_traits<void *>::tag ()));
  }
};

/**
 *  @brief Deseralisation wrapper: specialization for const char *
 *
 *  Without that would would have to handle const char *&, const char * const &, ...
 *  TODO: right now these types are not supported.
 */
template <> 
struct get_value<const char *>
{
  static VALUE get (gsi::SerialArgs &rr, const gsi::ArgType &arg)
  {
    tl_assert (! arg.is_cref);
    tl_assert (! arg.is_ref);
    tl_assert (! arg.is_cptr);
    tl_assert (! arg.is_ptr);
    return c2ruby<const char *>::get (rr.get_value<const char *> (gsi::type_traits<const char *>::tag ()));
  }
};

/**
 *  @brief Deseralisation wrapper 
 *
 *  This implementation is for vectors of the given type
 */
template <class R> 
struct get_vector
{
  static VALUE get (gsi::SerialArgs &rr, const gsi::ArgType &arg)
  {
    if (arg.is_ref) {

      std::vector<R> &v = rr.template get_value<std::vector<R> &> (typename gsi::type_traits<std::vector<R> &>::tag ());

      VALUE ret = rb_ary_new ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        rb_ary_push (ret, c2ruby<R>::get (*i));
      }
      return ret;

    } else if (arg.is_cref) {

      const std::vector<R> &v = rr.template get_value<const std::vector<R> &> (typename gsi::type_traits<const std::vector<R> &>::tag ());

      VALUE ret = rb_ary_new ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        rb_ary_push (ret, c2ruby<R>::get (*i));
      }
      return ret;

    } else if (arg.is_ptr) {

      std::vector<R> *v = rr.template get_value<std::vector<R> *> (typename gsi::type_traits<std::vector<R> *>::tag ());
      if (! v) {
        return Qnil;
      } else {

        VALUE ret = rb_ary_new ();
        for (typename std::vector<R>::const_iterator i = v->begin (); i != v->end (); ++i) {
          rb_ary_push (ret, c2ruby<R>::get (*i));
        }
        return ret;

      }

    } else if (arg.is_cptr) {

      const std::vector<R> *v = rr.template get_value<const std::vector<R> *> (typename gsi::type_traits<const std::vector<R> *>::tag ());
      if (! v) {
        return Qnil;
      } else {

        VALUE ret = rb_ary_new ();
        for (typename std::vector<R>::const_iterator i = v->begin (); i != v->end (); ++i) {
          rb_ary_push (ret, c2ruby<R>::get (*i));
        }
        return ret;

      }

    } else {

      std::vector<R> v = rr.template get_value< std::vector<R> > (typename gsi::type_traits< std::vector<R> >::tag ());

      VALUE ret = rb_ary_new ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        rb_ary_push (ret, c2ruby<R>::get (*i));
      }
      return ret;

    }
  }
};

// -------------------------------------------------------------------
//  Test if an argument can be converted to the given type

//  if atype is a vector:
//      argument must be an array of the given type
//  if atype is a ref:
//      argument must be a boxed type of the required type or an object of the requested class
//  if atype is a ptr:
//      argument must be a boxed type of the required type or an object of the requested class or nil
//  if atype is a cptr:
//      argument must be of requested type or nil
//  otherwise:
//      argument must be of the requested type

struct TestArgFunc
{
  TestArgFunc (VALUE arg, const gsi::ArgType &atype, bool loose)
    : m_arg (arg), m_atype (atype), m_ret (false), m_loose(loose)
  { }

  bool value () const { return m_ret; }

  template <class R>
  void op ()
  {
    if ((m_atype.is_cptr || m_atype.is_ptr) && m_arg == Qnil) {

      //  for const X * or X *, nil is an allowed value
      m_ret = true;

    } else {

      if (m_atype.is_ptr || m_atype.is_ref) {

        //  check if we have a boxed type
        if (TYPE (m_arg) == T_DATA) {
          const gsi::ClassBase *bc = gsi::cls_decl <gsi::Value> ();
          Proxy *p;
          Data_Get_Struct (m_arg, Proxy, p);
          if (p->cls_decl ()->is_derived_from (bc)) {
            m_ret = true;
          }
        }

      }

      if (! m_ret) {
        //  otherwise try a normal match and let the serializer sort out the wrong arguments with
        //  a good error message.
        m_ret = test_type<R>::test (m_arg, m_loose);
      }

    }
  }

private:
  VALUE m_arg;
  const gsi::ArgType &m_atype;
  bool m_ret, m_loose;
};

struct TestArgFuncVector
{
  TestArgFuncVector (VALUE arg, const gsi::ArgType &atype, bool loose)
    : m_arg (arg), m_atype (atype), m_ret (false), m_loose(loose)
  { }

  bool value () const { return m_ret; }

  template <class R>
  void op ()
  {
    m_ret = test_vector<R>::test (m_arg, m_loose);
  }

private:
  VALUE m_arg;
  const gsi::ArgType &m_atype;
  bool m_ret, m_loose;
};

bool
test_arg (const gsi::ArgType &atype, VALUE arg, bool loose)
{
  TestArgFunc f (arg, atype, loose);
  if (switch_type (atype.type, f)) {

    return f.value ();

  } else if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const gsi::ArgType &ainner = *atype.inner;

    TestArgFuncVector f (arg, ainner, loose);
    if (switch_type (ainner.type, f)) {

      return f.value ();

    } else if (ainner.type == gsi::T_object) {

      if (TYPE (arg) != T_ARRAY) {
        return false;
      }
      unsigned int len = RARRAY_LEN(arg);
      VALUE *el = RARRAY_PTR(arg);
      while (len-- > 0) {
        if (TYPE (*el) != T_DATA) {
          return false;
        }
        Proxy *p;
        Data_Get_Struct (*el, Proxy, p);
        if (!p->cls_decl ()->is_derived_from (ainner.cls)) {
          return false;
        }
        ++el;
      }

      return true;

    } else {
      return false;
    }

  } else if (atype.type == gsi::T_object) {

    //  for const X * or X *, nil is an allowed value
    if ((atype.is_cptr || atype.is_ptr) && arg == Qnil) {
      return true;
    }

    //  otherwise, the argument must be of the requested type 
    if (TYPE (arg) != T_DATA) {
      return false;
    }
    Proxy *p;
    Data_Get_Struct (arg, Proxy, p);
    if (!p->cls_decl ()->is_derived_from (atype.cls)) {
      return false;
    }
    if ((atype.is_ref || atype.is_ptr) && p->const_ref ()) {
      return false;
    }

    return true;

  } else {
    return false;
  }
}
      
// -------------------------------------------------------------------
//  Push an argument on the call or return stack

struct PushArgFunc
{
  PushArgFunc (const gsi::ArgType &atype, gsi::SerialArgs &aserial, VALUE arg, std::list<gsi::TempObject> &tmp_stack)
    : m_atype (atype), m_aserial (aserial), m_arg (arg), m_tmp_stack (tmp_stack)
  { }

  template <class R>
  void op () 
  {
    set_value<R>::set (m_aserial, m_arg, m_atype, m_tmp_stack);
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  VALUE m_arg;
  std::list<gsi::TempObject> &m_tmp_stack;
};

struct PushArgFuncVector
{
  PushArgFuncVector (const gsi::ArgType &atype, gsi::SerialArgs &aserial, VALUE arg, std::list<gsi::TempObject> &tmp_stack)
    : m_atype (atype), m_aserial (aserial), m_arg (arg), m_tmp_stack (tmp_stack)
  { }

  template <class R>
  void op () 
  {
    set_vector<R>::set (m_aserial, m_arg, m_atype, m_tmp_stack); 
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  VALUE m_arg;
  std::list<gsi::TempObject> &m_tmp_stack;
};

void
push_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, VALUE arg, std::list<gsi::TempObject> &tmp_stack)
{
  PushArgFunc f (atype, aserial, arg, tmp_stack);
  if (switch_type(atype.type, f)) {

    //  type handled in switch

  } else if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const gsi::ArgType &ainner = *atype.inner;

    //  type handled in switch
    PushArgFuncVector f (atype, aserial, arg, tmp_stack);
    if (switch_type(ainner.type, f)) {

      //  type handled in switch

    } else if (ainner.type == gsi::T_object) {

      if (TYPE (arg) != T_ARRAY) {
        throw tl::Exception (tl::sprintf (tl::translate ("Expected array either as argument or return type")));
      }
      unsigned int len = RARRAY_LEN(arg);
      VALUE *el = RARRAY_PTR(arg);

      std::vector<void *> vobj;
      vobj.reserve (len);
      while (len-- > 0) {
        Proxy *p;
        if (TYPE (*el) != T_DATA) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type inside array (expected argument of class %s)"), ainner.cls->name ()));
        }
        Data_Get_Struct (*el, Proxy, p);
        if (!p->cls_decl ()->is_derived_from (ainner.cls)) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type inside array (expected argument of class %s)"), ainner.cls->name ()));
        }
        vobj.push_back (p->obj (*el));
        ++el;
      }
      ainner.cls->push_vector_of (aserial, atype, tmp_stack, vobj);

      //  TODO: const check?

    } else {
      tl_assert (false); //  unsupported type
    }

  } else if (atype.type == gsi::T_void) {

    //  ignored

  } else if (atype.type == gsi::T_object) {

    if (arg == Qnil) {
      if (! (atype.is_ptr || atype.is_cptr)) {
        throw tl::Exception (tl::translate ("Arguments of reference or direct type cannot be passed nil"));
      } else {
        aserial.set_value (gsi::type_traits<void *>::tag (), (void *) 0);
      }
    } else if (atype.is_ptr || atype.is_cptr || atype.is_ref || atype.is_cref) {
      Proxy *p = 0;
      if (TYPE (arg) == T_DATA) {
        Data_Get_Struct (arg, Proxy, p);
        if (! p->cls_decl ()->is_derived_from (atype.cls)) {
          p = 0;
        }
      }
      if (! p) {
        throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s, got %s)"), atype.cls->name (), rb_class_name (arg).c_str ()));
      }
      if (p->const_ref () && (atype.is_ptr || atype.is_ref)) {
        throw tl::Exception (tl::sprintf (tl::translate ("Cannot pass a const reference of class %s to a non-const reference parameter"), atype.cls->name ()));
      }
      aserial.set_value (gsi::type_traits<void *>::tag (), p->obj (arg));
    } else {
      Proxy *p = 0;
      if (TYPE (arg) == T_DATA) {
        Data_Get_Struct (arg, Proxy, p);
        if (! p->cls_decl ()->is_derived_from (atype.cls)) {
          p = 0;
        }
      }
      if (! p) {
        throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s, got %s)"), atype.cls->name (), rb_class_name (arg).c_str ()));
      }
      atype.cls->copy_to (aserial, p->obj (arg));
    }

  } else {
    tl_assert (false);  // An argument cannot be "new"
  }
}
      
// -------------------------------------------------------------------
//  Correct constness if a reference is const and a non-const reference is required

//  HINT: this is a workaround for the fact that unlike C++, Ruby does not have const or non-const
//  references. Since a reference is identical with the object it points to, there are only const or non-const
//  objects. We deliver const objects first, but if a non-const version is requestes, the
//  object turns into a non-const one. This may be confusing but provides a certain level
//  of "constness", at least until there is another non-const reference for that object.

void correct_constness (VALUE obj, bool const_required)
{
  Proxy *p;
  Data_Get_Struct (obj, Proxy, p);
  if (p && p->const_ref () && ! const_required) {
    //  promote to non-const object
    p->set_const_ref (false);
  }
}

// -------------------------------------------------------------------
//  Some helper stuff for extracting arrays

struct ArrayData
{
  ArrayData (VALUE _ret, const gsi::ArgType &_ainner)
    : ret (_ret), ainner (_ainner)
  { }

  VALUE ret;
  const gsi::ArgType &ainner;
};

static void
make_array_f (void *data, void *cobj)
{
  ArrayData *ad = (ArrayData *)data;
  
  const gsi::ClassBase *clsact = ad->ainner.cls->subclass_decl (cobj);
  std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (clsact);
  tl_assert (c != rev_cls_map.end ());

  VALUE obj = Qnil;

  if (ad->ainner.is_ptr || ad->ainner.is_cptr) {

    if (cobj) {

      if (clsact->client_data (cobj) != 0 && 
          ((RBAClientData *) clsact->client_data (cobj))->get_ref () != Qnil) {

        //  we have a that is located in C++ space but is supposed to get attached
        //  a Ruby object. If it already has, we simply return a reference to this
        obj = ((RBAClientData *) clsact->client_data (cobj))->get_ref ();

        //  correct constness if the object is not supposed to be const
        correct_constness (obj, ad->ainner.is_cptr /*const obj*/);

      } else {

        obj = rb_class_new_instance_checked (0, 0, c->second);
        Proxy *p;
        Data_Get_Struct (obj, Proxy, p);

        p->set (cobj, false, ad->ainner.is_cptr /*const obj*/, ad->ainner.is_ptr /*can destroy*/, obj);

      }

    }

  } else {

    //  create a instance without attaching it
    obj = rb_class_new_instance_checked (0, 0, c->second);
    Proxy *p;
    Data_Get_Struct (obj, Proxy, p);
    p->set (cobj, true, false, false, obj);

  }

  rb_ary_push (ad->ret, obj);
}

// -------------------------------------------------------------------
//  Pop an argument from the call or return stack

struct PopArgFunc
{
  PopArgFunc (const gsi::ArgType &atype, gsi::SerialArgs &aserial)
    : m_atype (atype), m_aserial (aserial), m_ret (Qnil)
  { }

  VALUE value () const { return m_ret; }

  template <class R>
  void op () 
  {
    m_ret = get_value<R>::get (m_aserial, m_atype);
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  VALUE m_ret;
};

struct PopArgFuncVector
{
  PopArgFuncVector (const gsi::ArgType &atype, gsi::SerialArgs &aserial)
    : m_atype (atype), m_aserial (aserial), m_ret (Qnil)
  { }

  VALUE value () const { return m_ret; }

  template <class R>
  void op () 
  {
    m_ret = get_vector<R>::get (m_aserial, m_atype);
  }

private:
  const gsi::ArgType &m_atype;
  gsi::SerialArgs &m_aserial;
  VALUE m_ret;
};

VALUE
pop_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial)
{
  tl_assert (! atype.is_iter); // not implemented here 

  VALUE ret = Qnil;

  PopArgFunc f (atype, aserial);
  if (switch_type (atype.type, f)) {

    ret = f.value ();

  } else if (atype.type == gsi::T_vector) {

    tl_assert (atype.inner != 0);
    const gsi::ArgType &ainner = *atype.inner;

    //  type handled in switch
    PopArgFuncVector f (atype, aserial);
    if (switch_type(ainner.type, f)) {

      ret = f.value ();

    } else if (ainner.type == gsi::T_object) {

      ret = rb_ary_new ();
      ArrayData ad (ret, ainner);
      ainner.cls->get_vector_of (aserial, atype, (void *) &ad, &make_array_f);

    } else if (ainner.type == gsi::T_object_new) {
      tl_assert (false); // not implemented here 
    }

  } else if (atype.type == gsi::T_object) {

    if (atype.is_ptr || atype.is_ref || atype.is_cptr || atype.is_cref) {

      void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
      const gsi::ClassBase *clsact = atype.cls->subclass_decl (obj);

      if (obj == 0) {

        ret = Qnil;

      } else if (atype.is_cref && clsact->can_copy () && clsact->client_data (obj) == 0) {

        //  We copy objects passed by const reference if they are not managed
        //  (derived from gsi::ObjectBase, in that case, client_data is not 0).
        //  Such objects are often exposed internals. First we can't
        //  guarantee the const correctness of references. Second, we
        //  can't guarantee the lifetime of the container will exceed that
        //  of the exposed property. Hence copying is safer.

        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (clsact);
        tl_assert (c != rev_cls_map.end ());

        //  create a instance and copy the value
        ret = rb_class_new_instance_checked (0, 0, c->second);
        Proxy *p;
        Data_Get_Struct (ret, Proxy, p);
        clsact->assign (p->obj (ret), obj);

      } else if (clsact->client_data (obj) != 0 && 
                 ((RBAClientData *) clsact->client_data (obj))->get_ref () != Qnil) {

        //  we have a that is located in C++ space but is supposed to get attached
        //  a Ruby object. If it already has, we simply return a reference to this
        ret = ((RBAClientData *) clsact->client_data (obj))->get_ref ();

        //  correct constness if the object is not supposed to be const
        correct_constness (ret, atype.is_cref || atype.is_cptr);

      } else {

        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (clsact);
        tl_assert (c != rev_cls_map.end ());

        //  create a instance 
        //  TODO: we will create a fresh object here, delete it again and link the 
        //  reference to the exisiting object to the Ruby object. This is not quite
        //  efficient - we should avoid creating and deleting a dummy object first.
        ret = rb_class_new_instance_checked (0, 0, c->second);
        Proxy *p;
        Data_Get_Struct (ret, Proxy, p);
        p->set (obj, false, atype.is_cref || atype.is_cptr, atype.is_ptr, ret);

      }

    } else {

      std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
      tl_assert (c != rev_cls_map.end ());
      //  create a instance without attaching it
      ret = rb_class_new_instance_checked (0, 0, c->second);
      Proxy *p;
      Data_Get_Struct (ret, Proxy, p);
      p->set (atype.cls->take_and_create (aserial), true, false, false, ret);

    }

  } else if (atype.type == gsi::T_object_new) {

    void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
    if (obj == 0) {
      ret = Qnil;
    } else {
      std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
      tl_assert (c != rev_cls_map.end ());
      //  create a instance without attaching it
      ret = rb_class_new_instance_checked (0, 0, c->second);
      Proxy *p;
      Data_Get_Struct (ret, Proxy, p);
      p->set (obj, true, false, false, ret);
    }

  } 

  return ret;
}

// -------------------------------------------------------------------
//  "each" implementation for iterators

template <class R>
struct yield_iter
{
  static void
  yield (gsi::SerialArgs &rr)
  {
    gsi::IterAdaptorAbstractBase *iter = (gsi::IterAdaptorAbstractBase *) rr.template get_value<gsi::IterAdaptorAbstractBase *> (typename gsi::type_traits<gsi::IterAdaptorAbstractBase *>::tag ());
    try {
      while (! iter->at_end ()) {
        rb_yield (c2ruby<R>::get (*((const R *) iter->v_get ())));
        iter->inc ();
      }
      delete iter;
    } catch (...) {
      delete iter;
      throw;
    }
  }
};

// -------------------------------------------------------------------
//  Proxy implementation

Proxy::~Proxy ()
{
  //  unregister the Proc objects
  for (std::map <const gsi::MethodBase *, VALUE>::iterator pt = m_proc_table.begin (); pt != m_proc_table.end (); ++pt) {
    rb_gc_unregister_address (&pt->second);
  }
  m_proc_table.clear ();

  try {
    set (0, false, false, false, Qnil);
  } catch (std::exception &ex) {
    tl::warn << "Caught exception in object destructor: " << ex.what ();
  } catch (tl::Exception &ex) {
    tl::warn << "Caught exception in object destructor: " << ex.msg ();
  } catch (...) {
    tl::warn << "Caught unspecified exception in object destructor";
  }
  m_destroyed = true;
}

void 
Proxy::register_event_proc (const gsi::MethodBase *meth, VALUE proc)
{
  std::map <const gsi::MethodBase *, VALUE>::iterator pt = m_proc_table.find (meth);

  //  unregister an existing Proc object 
  if (pt != m_proc_table.end ()) {
    rb_gc_unregister_address (&pt->second);
    m_proc_table.erase (pt);
  } 

  //  end register the new one
  VALUE *proc_addr = &m_proc_table.insert (std::make_pair (meth, proc)).first->second;
  rb_gc_register_address (proc_addr);
}

void 
Proxy::call (int id, gsi::SerialArgs &args, gsi::SerialArgs &ret) 
{
  const gsi::MethodBase *meth = m_cbfuncs [id].method;

  try {

    VALUE obj;
    ID mid;

    if (meth->is_event ()) {

      //  look up the Proc object in the table and call that ..
      std::map <const gsi::MethodBase *, VALUE>::const_iterator pt = m_proc_table.find (meth);
      if (pt == m_proc_table.end ()) {

        //  event not bound: create a nil return value if required.
        std::list<gsi::TempObject> tmp_stack;
        push_arg (meth->ret_type (), ret, Qnil, tmp_stack);

        //  a Ruby callback must not leave temporary objects
        tl_assert (tmp_stack.empty ());

        return;

      }

      obj = pt->second;
      mid  = rb_intern("call");

    } else {

      tl_assert (id < int (m_cbfuncs.size ()) && id >= 0);
      obj = m_cbfuncs [id].rb_obj;
      mid  = m_cbfuncs [id].method_id;

    }

    std::vector<VALUE> argv;
    argv.reserve (std::distance (meth->begin_arguments (), meth->end_arguments ()));

    for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a) {
      argv.push_back (pop_arg (*a, args));
    }

    VALUE rb_ret = rb_funcall2_checked (obj, mid, argv.size (), argv.empty () ? 0 : &argv [0]);
    
    std::list<gsi::TempObject> tmp_stack;
    push_arg (meth->ret_type (), ret, rb_ret, tmp_stack);

    //  a Ruby callback must not leave temporary objects
    tl_assert (tmp_stack.empty ());

  } catch (rba::RubyError &err) {
    rba::RubyError err_with_context (err);
    err_with_context.set_context (m_cls_decl->name () + "::" + meth->name ());
    throw err_with_context;
  } catch (rba::ExitException &ex) {
    throw;
  } catch (tl::Exception &ex) {
    throw tl::Exception (tl::translate ("Error calling method") + " '" + m_cls_decl->name () + "::" + meth->name () + "': " + ex.msg ());
  } catch (...) {
    throw;
  }
}

void 
Proxy::destroy ()
{
  if (! m_cls_decl) {
    m_obj = 0;
    return;
  }

  if (!(m_owned || m_can_destroy) && m_obj) {
    throw tl::Exception (tl::translate ("Object cannot be destroyed explicitly - it is either a const reference or a direct object"));
  }

  //  first create the object if it was not created yet and check if it has not been 
  //  destroyed already (the former is to ensure that the object is created at least)
  if (! m_obj) {
    if (m_destroyed) {
      throw tl::Exception (tl::translate ("Object has been destroyed already"));
    } else {
      m_obj = m_cls_decl->create ();
      if (m_obj) {
        m_cls_decl->attach (m_obj);
      }
      m_owned = true;
    }
  }

  void *o = 0;
  if (m_owned || m_can_destroy) {
    o = m_obj;
  }
  if (m_obj) {
    m_cls_decl->detach (m_obj);
  }
  detach ();
  if (o) {
    m_cls_decl->destroy (o);
  }
}

void 
Proxy::detach ()
{
  m_obj = 0;
  m_destroyed = true;
  m_const_ref = false;
  m_owned = false;
  m_can_destroy = false;
}

void 
Proxy::set (void *obj, bool owned, bool const_ref, bool can_destroy, VALUE self) 
{
  bool prev_owned = m_owned;

  m_owned = owned;
  m_can_destroy = can_destroy;

  m_const_ref = const_ref;
  if (obj != m_obj) {

    //  cleanup
    {

      const gsi::ClassBase *cls = cls_decl ();
      RBAClientData *cd = 0;
      if (m_obj && cls) {
        cd = (RBAClientData *) cls->client_data (m_obj);
      }

      //  Destroy the object if we are owner. We don't destroy the object if it was locked
      //  (either because we are not owner or from C++ side using keep())
      if (m_obj && cls && prev_owned && (! cd || ! cd->locked ())) {
        void *o = m_obj;
        cls->detach (o);
        m_obj = 0;
        cls->destroy (o);
      }

      if (m_obj) {

        //  Reset the reference
        if (cd) {
          cd->reset_ref ();
        }

        //  we'll get a new object - we have to detach event handlers.
        while (cls) {
          for (gsi::ClassBase::method_iterator m = cls->begin_methods (); m != cls->end_methods (); ++m) {
            if ((*m)->is_event ()) {
              (*m)->set_callback (m_obj, gsi::Callback ());
            }
          }
          //  consider base classes as well.
          cls = cls->base ();
        }

      }

    }

    clear_callbacks ();

    if (m_cls_decl && m_obj) {
      m_cls_decl->detach (m_obj);
    }

    m_obj = obj;

    if (m_cls_decl && m_obj) {
      m_cls_decl->attach (m_obj);
    }

    if (m_obj) {

      const gsi::ClassBase *cls = m_cls_decl;

      //  manage attachement to the C++ object if required
      //  Hold a reference for the GC unless we own the object (read: control the lifetime ourselves)
      if (cls && cls->client_data (m_obj) != 0) {
        ((RBAClientData *) cls->client_data (m_obj))->set_ref (self, !owned);
      }

      // we got a new object - hence we have to attach event handlers.
      // We don't need to install virtual function callbacks because in that case, no overload is possible
      // (the object has been created on C++ side).
      while (cls) {

        for (gsi::ClassBase::method_iterator m = cls->begin_methods (); m != cls->end_methods (); ++m) {

          if ((*m)->is_event ()) {

            //  We can link the event with the callback now.
            int id = add_callback (Proxy::CallbackFunction (0, Qnil, *m));
            (*m)->set_callback (m_obj, gsi::Callback (id, this, (*m)->argsize (), (*m)->retsize ()));

          } else if ((*m)->is_callback ()) {

            //  HINT: all callback may not have aliases nor overloads, but it can have a "*" prefix for protected
            bool prot = false;
            const char *nstr = extract_protected((*m)->name ().c_str (), prot);

            //  There is no place in the ruby API to determine whether a method is defined.
            //  Instead we explicitly call "method_defined?" to check, whether the given method 
            //  is defined.
            VALUE name = rb_str_new (nstr, strlen (nstr));

            for (int prot = 0; prot < 2; ++prot) {

              VALUE rb_ret;
              if (prot) {
                rb_ret = rb_funcall2_checked (rb_class_of (self), rb_intern ("protected_method_defined?"), 1, &name);
              } else {
                rb_ret = rb_funcall2_checked (rb_class_of (self), rb_intern ("method_defined?"), 1, &name);
              }
              if (RTEST (rb_ret)) {

                //  Only if the class defines that method we can link the virtual method call to the 
                //  Ruby method
                ID method_id = rb_intern (nstr);
                int id = add_callback (Proxy::CallbackFunction (method_id, self, *m));
                (*m)->set_callback (m_obj, gsi::Callback (id, this, (*m)->argsize (), (*m)->retsize ()));

                break;

              }

            }

          }

        }

        //  consider base classes as well.
        cls = cls->base ();

      }

    }

  }

  //  now we have a valid object (or nil) - we can reset "destroyed" state. Note: this has to be done 
  //  here because before detach might be called on *this which resets m_destroyed.
  m_destroyed = false;
}

void *
Proxy::obj (VALUE self) 
{
  if (! m_obj) {
    if (m_destroyed) {
      throw tl::Exception (tl::translate ("Object has been destroyed already"));
    } else {
      //  delayed creation of a detached C++ object ..
      set(cls_decl ()->create (), true, m_const_ref, m_can_destroy, self);
    }
  }

  return m_obj;
}

// -------------------------------------------------------------------
//  Ruby API 

#define RBA_TRY \
  VALUE __error_msg = Qnil; \
  int __estatus = 0; \
  VALUE __eclass = Qnil; \
  { \
    try { 

#define RBA_CATCH(where) \
    } catch (std::exception &ex) { \
      __eclass = rb_eRuntimeError; \
      __error_msg = rb_str_new2 ((std::string(ex.what ()) + tl::translate (" in ") + (where)).c_str ()); \
    } catch (rba::ExitException &ex) { \
      __estatus = ex.status (); \
      __eclass = rb_eSystemExit; \
      __error_msg = rb_str_new2 ((ex.msg () + tl::translate (" in ") + (where)).c_str ()); \
    } catch (tl::Exception &ex) { \
      __eclass = rb_eRuntimeError; \
      __error_msg = rb_str_new2 ((ex.msg () + tl::translate (" in ") + (where)).c_str ()); \
    } catch (...) { \
      __eclass = rb_eRuntimeError; \
      __error_msg = rb_str_new2 ((tl::translate ("Unspecific exception in ") + (where)).c_str ()); \
    } \
  } \
  if (__eclass == rb_eSystemExit) { \
    /*  HINT: we do the rb_raise outside any destructor code - sometimes this longjmp seems not to work properly */ \
    VALUE args [2]; \
    args [0] = INT2NUM (__estatus); \
    args [1] = __error_msg; \
    rb_exc_raise (rb_class_new_instance(2, args, __eclass)); \
  } else if (__eclass != Qnil) { \
    /*  HINT: we do the rb_raise outside any destructor code - sometimes this longjmp seems not to work properly */ \
    VALUE args [1]; \
    args [0] = __error_msg; \
    rb_exc_raise (rb_class_new_instance(1, args, __eclass)); \
  }

VALUE gsi_init (VALUE self)
{
  /* 
   * This was supposed to be a check if the object can be created with new.
   * Unfortunately we can't do this, because a initialization happens even if an object is 
   * returned as reference: 
   
  RBA_TRY

    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    const gsi::ClassBase *cls_decl_self = p->cls_decl ();
    if (! cls_decl_self->can_default_create ()) {
        throw tl::Exception (tl::translate ("No constructor provided for class '%s'").c_str (), cls_decl_self->name ().c_str ());
    }

  RBA_CATCH(rb_class_name (self) + "::initialize")

  */

  return self;
}

VALUE gsi_destroy (VALUE self)
{
  RBA_TRY

    //  Destroy the object
    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    p->destroy ();

  RBA_CATCH(rb_class_name (self) + "::destroy")

  return Qnil;
}

VALUE gsi_create (VALUE self)
{
  RBA_TRY

    //  return true if the object was destroyed already
    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    //  this potentially instantiates the object if not done yet
    p->obj (self);

  RBA_CATCH(rb_class_name (self) + "::create")

  return self;
}

VALUE gsi_destroyed (VALUE self)
{
  bool r = false; 

  RBA_TRY

    //  return true if the object was destroyed already
    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    r = p->destroyed ();

  RBA_CATCH(rb_class_name (self) + "::destroyed")

  return c2ruby<bool>::get (r);
}

VALUE gsi_is_const (VALUE self)
{
  bool r = false; 

  RBA_TRY

    //  return true if the object was destroyed already
    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    r = p->const_ref ();

  RBA_CATCH(rb_class_name (self) + "::is_const?")

  return c2ruby<bool>::get (r);
}

VALUE gsi_init_copy (VALUE self, VALUE src)
{
  RBA_TRY

    //  Compare if the classes are identical
    Proxy *p;

    Data_Get_Struct (src, Proxy, p);
    const gsi::ClassBase *cls_decl_src = p->cls_decl ();
    void *obj_src = p->obj (src);

    Data_Get_Struct (self, Proxy, p);
    const gsi::ClassBase *cls_decl_self = p->cls_decl ();
    void *obj_self = p->obj (self);

    if (cls_decl_src != cls_decl_self) {
      throw tl::Exception (tl::sprintf (tl::translate ("Type is not identical on copy")));
    } 
    if (! cls_decl_self->can_copy ()) {
      throw tl::Exception (tl::translate ("No copy constructor provided for class '%s'"), cls_decl_self->name ());
    }
    cls_decl_self->assign (obj_self, obj_src);

  RBA_CATCH(rb_class_name (self) + "::initialize_copy")

  return self;
}

void free_proxy (void *p)
{
  delete ((Proxy *) p);
}


VALUE alloc_proxy (VALUE klass)
{
  const gsi::ClassBase *cls = find_cclass (klass);
  Proxy *proxy = new Proxy (cls);
  return Data_Wrap_Struct (klass, 0, free_proxy, proxy);
}

struct YieldIterFunc
{
  YieldIterFunc (gsi::SerialArgs &retlist)
    : m_retlist (retlist)
  { }

  template <class R> void op() { yield_iter<R>::yield (m_retlist); }

private:
  gsi::SerialArgs &m_retlist;
};

/**
 *  @brief Gets the method name from a method id
 */
std::string 
method_name_from_id (int mid, VALUE self)
{
  const gsi::ClassBase *cls_decl;
  Proxy *p = 0;

  if (TYPE (self) == T_CLASS) {
    //  we have a static method
    cls_decl = find_cclass (self);
  } else {
    //  we have an instance method
    Data_Get_Struct (self, Proxy, p);
    cls_decl = p->cls_decl ();
  }

  //  locate the method in the base classes method table if necessary
  while (mid < int (cls_decl->method_offset ())) {
    tl_assert (cls_decl->base ());
    cls_decl = cls_decl->base ();
  }

  mid -= cls_decl->method_offset ();

  gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); 
  if (om != cls_decl->end_overloaded_methods (mid)) {
    return cls_decl->name () + "::" + cls_decl->overloaded_name (mid);
  } else {
    return cls_decl->name () + "::<unknown method>";
  }
}

VALUE
method_adaptor (int mid, int argc, VALUE *argv, VALUE self, bool ctor)
{
  VALUE ret = Qnil;

  std::list<gsi::TempObject> tmp_stack;

  const gsi::ClassBase *cls_decl;
  Proxy *p = 0;

  if (TYPE (self) == T_CLASS) {
    //  we have a static method
    cls_decl = find_cclass (self);
  } else {
    //  we have an instance method
    Data_Get_Struct (self, Proxy, p);
    cls_decl = p->cls_decl ();
  }

  //  locate the method in the base classes method table if necessary
  while (mid < int (cls_decl->method_offset ())) {
    tl_assert (cls_decl->base ());
    cls_decl = cls_decl->base ();
  }

  mid -= cls_decl->method_offset ();

  //  get number of candidates by argument count
  const gsi::MethodBase *meth = 0;
  unsigned int candidates = 0;
  for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
    const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
    if (m->is_event()) {
      if (rb_block_given_p ()) {
        //  events do not have parameters, but accept a Proc object -> no overloading -> take this one.
        ++candidates;
        meth = m;
        break;
      } else {
        throw tl::Exception (tl::translate ("An event needs a block"));
      }
    } else if (m->is_callback()) {
      //  ignore callbacks
    } else if (std::distance (m->begin_arguments (), m->end_arguments ()) == argc) {
      ++candidates;
      meth = m;
    }
  }

  //  no method found, but the ctor was requested - implement that method as replacement for the default "initialize"
  if (! meth && argc == 0 && ctor) {
    return Qnil;
  }

  //  no candidate -> error
  if (! meth) {

    std::set<unsigned int> nargs;
    for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
      const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
      if (! m->is_callback () && ! m->is_event ()) {
        nargs.insert (std::distance (m->begin_arguments (), m->end_arguments ()));
      }
    }

    std::string nargs_s;
    for (std::set<unsigned int>::const_iterator na = nargs.begin (); na != nargs.end (); ++na) {
      if (na != nargs.begin ()) {
        nargs_s += "/";
      }
      nargs_s += tl::to_string (*na);
    }

    throw tl::Exception (tl::sprintf (tl::translate ("Invalid number of arguments (got %d, expected %s)"), argc, nargs_s));

  }

  //  more than one candidate -> refine by checking the arguments
  if (candidates > 1) {

    //  two passes where the second is with loose checking
    int pass = 0;

    do {

      meth = 0;
      candidates = 0;
      bool const_matching = true;
      for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
        const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
        if (! m->is_callback () && ! m->is_event ()) {
          bool is_valid = (std::distance (m->begin_arguments (), m->end_arguments ()) == argc);
          VALUE *av = argv;
          for (gsi::MethodBase::argument_iterator a = m->begin_arguments (); is_valid && a != m->end_arguments (); ++a, ++av) {
            if (! test_arg (*a, *av, pass != 0 /*loose in the second pass*/)) {
              is_valid = false;
            }
          }
          if (is_valid && p) {
            //  constness matching candidates have precedence
            if (m->is_const () != p->const_ref ()) {
              if (const_matching && candidates > 0) {
                is_valid = false;
              } else {
                const_matching = false;
              }
            } else if (! const_matching) {
              const_matching = true;
              candidates = 0;
            }
          }
          if (is_valid) {
            ++candidates;
            meth = m;
          }
        }
      }

      ++pass;

    } while (! meth && pass < 2);

  }

  if (! meth) {
    throw tl::Exception (tl::translate ("No overload with matching arguments"));
  }

  if (candidates > 1) {
    throw tl::Exception (tl::translate ("Ambiguous overload variants - multiple method declarations match arguments"));
  }

  if (p && p->const_ref () && ! meth->is_const ()) {
    throw tl::Exception (tl::translate ("Cannot call non-const on a const reference"));
  }

  if (meth->is_event ()) {

    if (p) {
      //  Hint: this potentially instantiates the object
      p->obj (self);
      p->register_event_proc(meth, rb_block_proc ());
    }

  } else if (ctor && meth->ret_type ().type == gsi::T_object_new) {

    tl_assert (p != 0);

    //  This is a non-static constructor ("new" renamed to "initialize"): it does not create a 
    //  new Ruby object but just a new C++ object which replaces the old one.

    //  TODO: when the yield block raises an error, the buffer allocated here does not get cleaned up!
    gsi::SerialArgs retlist (meth->retsize ());

    {
      gsi::SerialArgs arglist (meth->argsize ());
      VALUE *av = argv;
      for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a, ++av) {
        push_arg (*a, arglist, *av, tmp_stack);
      }

      meth->call (0, arglist, retlist);
    }

    void *obj = retlist.get_value<void *> (gsi::type_traits<void *>::tag ());
    if (obj == 0) {
      p->set (0, false, false, false, Qnil);
    } else {
      p->set (obj, true, false, false, self);
    }

  } else {

    void *obj = 0;
    if (p) {
      //  Hint: this potentially instantiates the object
      obj = p->obj (self);
    }

    //  TODO: when the yield block raises an error, the buffer allocated here does not get cleaned up!
    gsi::SerialArgs retlist (meth->retsize ());

    {
      gsi::SerialArgs arglist (meth->argsize ());
      VALUE *av = argv;
      for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a, ++av) {
        push_arg (*a, arglist, *av, tmp_stack);
      }

      meth->call (obj, arglist, retlist);
    }

    if (meth->ret_type ().is_iter) {

      YieldIterFunc f (retlist);
      if (switch_type (meth->ret_type ().type, f)) {
        //  handled in switch
      } else if (meth->ret_type ().type == gsi::T_object) {

        const gsi::ClassBase *cls = meth->ret_type ().cls;
        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (cls);
        tl_assert (c != rev_cls_map.end ());

        if (meth->ret_type ().is_ref || meth->ret_type ().is_cref || meth->ret_type ().is_ptr || meth->ret_type ().is_cptr) {

          void *iter = c->first->create_obj_iterator (retlist);

          try {

            while (! c->first->obj_iterator_at_end (iter)) {

              void *cobj = const_cast<void *> (c->first->get_iterator_obj (iter));

              //  specialize the class if possible
              const gsi::ClassBase *cls_act = cls->subclass_decl (cobj);

              VALUE obj;
              if (cobj == 0) {

                obj = Qnil;

              } else if (cls_act->client_data (cobj) != 0 && 
                         ((RBAClientData *) cls_act->client_data (cobj))->get_ref () != Qnil) {

                //  we have a that is located in C++ space but is supposed to get attached
                //  a Ruby object. If it already has, we simply return a reference to this
                obj = ((RBAClientData *) cls_act->client_data (cobj))->get_ref ();

                //  correct constness if the object is not supposed to be const
                correct_constness (obj, meth->ret_type ().is_cref || meth->ret_type ().is_cptr);

              } else {

                //  the object returned can be of a derived class. Hence we need to look up the class again
                std::map <const gsi::ClassBase *, VALUE>::const_iterator cact = rev_cls_map.find (cls_act);
                tl_assert (cact != rev_cls_map.end ());

                //  create the object
                obj = rb_class_new_instance_checked (0, 0, cact->second);
                Proxy *p;
                Data_Get_Struct (obj, Proxy, p);
                p->set (cobj, false, meth->ret_type ().is_cref || meth->ret_type ().is_cptr, meth->ret_type ().is_ptr, obj);

              }

              rb_yield (obj);
              c->first->inc_obj_iterator (iter);

            }

            c->first->delete_obj_iterator (iter);

          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }

        } else {

          void *iter = c->first->create_obj_iterator (retlist);

          try {

            while (! c->first->obj_iterator_at_end (iter)) {

              VALUE obj = rb_class_new_instance_checked (0, 0, c->second);
              Proxy *p;
              Data_Get_Struct (obj, Proxy, p);
              p->set (c->first->take_iterator_obj (iter), true, false, false, obj);
              rb_yield (obj);
              c->first->inc_obj_iterator (iter);

            }

            c->first->delete_obj_iterator (iter);

          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }

        }

      } else if (meth->ret_type ().type == gsi::T_object_new) {
        tl_assert (false);  //  not implemented
      }

    } else {
      ret = pop_arg (meth->ret_type (), retlist);
    }

  }

  return ret;
}

template <int N> 
VALUE method_adaptor (int argc, VALUE *argv, VALUE self)
{
  VALUE ret = Qnil;

  RBA_TRY

    ret = method_adaptor (N, argc, argv, self, false);

  RBA_CATCH(method_name_from_id (N, self))

  return ret;
}

VALUE (*(method_adaptors [])) (int, VALUE *, VALUE) =
{
  &method_adaptor<0x000>, &method_adaptor<0x001>, &method_adaptor<0x002>, &method_adaptor<0x003>, &method_adaptor<0x004>, &method_adaptor<0x005>, &method_adaptor<0x006>, &method_adaptor<0x007>,
  &method_adaptor<0x008>, &method_adaptor<0x009>, &method_adaptor<0x00a>, &method_adaptor<0x00b>, &method_adaptor<0x00c>, &method_adaptor<0x00d>, &method_adaptor<0x00e>, &method_adaptor<0x00f>,
  &method_adaptor<0x010>, &method_adaptor<0x011>, &method_adaptor<0x012>, &method_adaptor<0x013>, &method_adaptor<0x014>, &method_adaptor<0x015>, &method_adaptor<0x016>, &method_adaptor<0x017>,
  &method_adaptor<0x018>, &method_adaptor<0x019>, &method_adaptor<0x01a>, &method_adaptor<0x01b>, &method_adaptor<0x01c>, &method_adaptor<0x01d>, &method_adaptor<0x01e>, &method_adaptor<0x01f>,
  &method_adaptor<0x020>, &method_adaptor<0x021>, &method_adaptor<0x022>, &method_adaptor<0x023>, &method_adaptor<0x024>, &method_adaptor<0x025>, &method_adaptor<0x026>, &method_adaptor<0x027>,
  &method_adaptor<0x028>, &method_adaptor<0x029>, &method_adaptor<0x02a>, &method_adaptor<0x02b>, &method_adaptor<0x02c>, &method_adaptor<0x02d>, &method_adaptor<0x02e>, &method_adaptor<0x02f>,
  &method_adaptor<0x030>, &method_adaptor<0x031>, &method_adaptor<0x032>, &method_adaptor<0x033>, &method_adaptor<0x034>, &method_adaptor<0x035>, &method_adaptor<0x036>, &method_adaptor<0x037>,
  &method_adaptor<0x038>, &method_adaptor<0x039>, &method_adaptor<0x03a>, &method_adaptor<0x03b>, &method_adaptor<0x03c>, &method_adaptor<0x03d>, &method_adaptor<0x03e>, &method_adaptor<0x03f>,
  &method_adaptor<0x040>, &method_adaptor<0x041>, &method_adaptor<0x042>, &method_adaptor<0x043>, &method_adaptor<0x044>, &method_adaptor<0x045>, &method_adaptor<0x046>, &method_adaptor<0x047>,
  &method_adaptor<0x048>, &method_adaptor<0x049>, &method_adaptor<0x04a>, &method_adaptor<0x04b>, &method_adaptor<0x04c>, &method_adaptor<0x04d>, &method_adaptor<0x04e>, &method_adaptor<0x04f>,
  &method_adaptor<0x050>, &method_adaptor<0x051>, &method_adaptor<0x052>, &method_adaptor<0x053>, &method_adaptor<0x054>, &method_adaptor<0x055>, &method_adaptor<0x056>, &method_adaptor<0x057>,
  &method_adaptor<0x058>, &method_adaptor<0x059>, &method_adaptor<0x05a>, &method_adaptor<0x05b>, &method_adaptor<0x05c>, &method_adaptor<0x05d>, &method_adaptor<0x05e>, &method_adaptor<0x05f>,
  &method_adaptor<0x060>, &method_adaptor<0x061>, &method_adaptor<0x062>, &method_adaptor<0x063>, &method_adaptor<0x064>, &method_adaptor<0x065>, &method_adaptor<0x066>, &method_adaptor<0x067>,
  &method_adaptor<0x068>, &method_adaptor<0x069>, &method_adaptor<0x06a>, &method_adaptor<0x06b>, &method_adaptor<0x06c>, &method_adaptor<0x06d>, &method_adaptor<0x06e>, &method_adaptor<0x06f>,
  &method_adaptor<0x070>, &method_adaptor<0x071>, &method_adaptor<0x072>, &method_adaptor<0x073>, &method_adaptor<0x074>, &method_adaptor<0x075>, &method_adaptor<0x076>, &method_adaptor<0x077>,
  &method_adaptor<0x078>, &method_adaptor<0x079>, &method_adaptor<0x07a>, &method_adaptor<0x07b>, &method_adaptor<0x07c>, &method_adaptor<0x07d>, &method_adaptor<0x07e>, &method_adaptor<0x07f>,
  &method_adaptor<0x080>, &method_adaptor<0x081>, &method_adaptor<0x082>, &method_adaptor<0x083>, &method_adaptor<0x084>, &method_adaptor<0x085>, &method_adaptor<0x086>, &method_adaptor<0x087>,
  &method_adaptor<0x088>, &method_adaptor<0x089>, &method_adaptor<0x08a>, &method_adaptor<0x08b>, &method_adaptor<0x08c>, &method_adaptor<0x08d>, &method_adaptor<0x08e>, &method_adaptor<0x08f>,
  &method_adaptor<0x090>, &method_adaptor<0x091>, &method_adaptor<0x092>, &method_adaptor<0x093>, &method_adaptor<0x094>, &method_adaptor<0x095>, &method_adaptor<0x096>, &method_adaptor<0x097>,
  &method_adaptor<0x098>, &method_adaptor<0x099>, &method_adaptor<0x09a>, &method_adaptor<0x09b>, &method_adaptor<0x09c>, &method_adaptor<0x09d>, &method_adaptor<0x09e>, &method_adaptor<0x09f>,
  &method_adaptor<0x0a0>, &method_adaptor<0x0a1>, &method_adaptor<0x0a2>, &method_adaptor<0x0a3>, &method_adaptor<0x0a4>, &method_adaptor<0x0a5>, &method_adaptor<0x0a6>, &method_adaptor<0x0a7>,
  &method_adaptor<0x0a8>, &method_adaptor<0x0a9>, &method_adaptor<0x0aa>, &method_adaptor<0x0ab>, &method_adaptor<0x0ac>, &method_adaptor<0x0ad>, &method_adaptor<0x0ae>, &method_adaptor<0x0af>,
  &method_adaptor<0x0b0>, &method_adaptor<0x0b1>, &method_adaptor<0x0b2>, &method_adaptor<0x0b3>, &method_adaptor<0x0b4>, &method_adaptor<0x0b5>, &method_adaptor<0x0b6>, &method_adaptor<0x0b7>,
  &method_adaptor<0x0b8>, &method_adaptor<0x0b9>, &method_adaptor<0x0ba>, &method_adaptor<0x0bb>, &method_adaptor<0x0bc>, &method_adaptor<0x0bd>, &method_adaptor<0x0be>, &method_adaptor<0x0bf>,
  &method_adaptor<0x0c0>, &method_adaptor<0x0c1>, &method_adaptor<0x0c2>, &method_adaptor<0x0c3>, &method_adaptor<0x0c4>, &method_adaptor<0x0c5>, &method_adaptor<0x0c6>, &method_adaptor<0x0c7>,
  &method_adaptor<0x0c8>, &method_adaptor<0x0c9>, &method_adaptor<0x0ca>, &method_adaptor<0x0cb>, &method_adaptor<0x0cc>, &method_adaptor<0x0cd>, &method_adaptor<0x0ce>, &method_adaptor<0x0cf>,
  &method_adaptor<0x0d0>, &method_adaptor<0x0d1>, &method_adaptor<0x0d2>, &method_adaptor<0x0d3>, &method_adaptor<0x0d4>, &method_adaptor<0x0d5>, &method_adaptor<0x0d6>, &method_adaptor<0x0d7>,
  &method_adaptor<0x0d8>, &method_adaptor<0x0d9>, &method_adaptor<0x0da>, &method_adaptor<0x0db>, &method_adaptor<0x0dc>, &method_adaptor<0x0dd>, &method_adaptor<0x0de>, &method_adaptor<0x0df>,
  &method_adaptor<0x0e0>, &method_adaptor<0x0e1>, &method_adaptor<0x0e2>, &method_adaptor<0x0e3>, &method_adaptor<0x0e4>, &method_adaptor<0x0e5>, &method_adaptor<0x0e6>, &method_adaptor<0x0e7>,
  &method_adaptor<0x0e8>, &method_adaptor<0x0e9>, &method_adaptor<0x0ea>, &method_adaptor<0x0eb>, &method_adaptor<0x0ec>, &method_adaptor<0x0ed>, &method_adaptor<0x0ee>, &method_adaptor<0x0ef>,
  &method_adaptor<0x0f0>, &method_adaptor<0x0f1>, &method_adaptor<0x0f2>, &method_adaptor<0x0f3>, &method_adaptor<0x0f4>, &method_adaptor<0x0f5>, &method_adaptor<0x0f6>, &method_adaptor<0x0f7>,
  &method_adaptor<0x0f8>, &method_adaptor<0x0f9>, &method_adaptor<0x0fa>, &method_adaptor<0x0fb>, &method_adaptor<0x0fc>, &method_adaptor<0x0fd>, &method_adaptor<0x0fe>, &method_adaptor<0x0ff>,
  &method_adaptor<0x100>, &method_adaptor<0x101>, &method_adaptor<0x102>, &method_adaptor<0x103>, &method_adaptor<0x104>, &method_adaptor<0x105>, &method_adaptor<0x106>, &method_adaptor<0x107>,
  &method_adaptor<0x108>, &method_adaptor<0x109>, &method_adaptor<0x10a>, &method_adaptor<0x10b>, &method_adaptor<0x10c>, &method_adaptor<0x10d>, &method_adaptor<0x10e>, &method_adaptor<0x10f>,
  &method_adaptor<0x110>, &method_adaptor<0x111>, &method_adaptor<0x112>, &method_adaptor<0x113>, &method_adaptor<0x114>, &method_adaptor<0x115>, &method_adaptor<0x116>, &method_adaptor<0x117>,
  &method_adaptor<0x118>, &method_adaptor<0x119>, &method_adaptor<0x11a>, &method_adaptor<0x11b>, &method_adaptor<0x11c>, &method_adaptor<0x11d>, &method_adaptor<0x11e>, &method_adaptor<0x11f>,
  &method_adaptor<0x120>, &method_adaptor<0x121>, &method_adaptor<0x122>, &method_adaptor<0x123>, &method_adaptor<0x124>, &method_adaptor<0x125>, &method_adaptor<0x126>, &method_adaptor<0x127>,
  &method_adaptor<0x128>, &method_adaptor<0x129>, &method_adaptor<0x12a>, &method_adaptor<0x12b>, &method_adaptor<0x12c>, &method_adaptor<0x12d>, &method_adaptor<0x12e>, &method_adaptor<0x12f>,
  &method_adaptor<0x130>, &method_adaptor<0x131>, &method_adaptor<0x132>, &method_adaptor<0x133>, &method_adaptor<0x134>, &method_adaptor<0x135>, &method_adaptor<0x136>, &method_adaptor<0x137>,
  &method_adaptor<0x138>, &method_adaptor<0x139>, &method_adaptor<0x13a>, &method_adaptor<0x13b>, &method_adaptor<0x13c>, &method_adaptor<0x13d>, &method_adaptor<0x13e>, &method_adaptor<0x13f>,
  &method_adaptor<0x140>, &method_adaptor<0x141>, &method_adaptor<0x142>, &method_adaptor<0x143>, &method_adaptor<0x144>, &method_adaptor<0x145>, &method_adaptor<0x146>, &method_adaptor<0x147>,
  &method_adaptor<0x148>, &method_adaptor<0x149>, &method_adaptor<0x14a>, &method_adaptor<0x14b>, &method_adaptor<0x14c>, &method_adaptor<0x14d>, &method_adaptor<0x14e>, &method_adaptor<0x14f>,
  &method_adaptor<0x150>, &method_adaptor<0x151>, &method_adaptor<0x152>, &method_adaptor<0x153>, &method_adaptor<0x154>, &method_adaptor<0x155>, &method_adaptor<0x156>, &method_adaptor<0x157>,
  &method_adaptor<0x158>, &method_adaptor<0x159>, &method_adaptor<0x15a>, &method_adaptor<0x15b>, &method_adaptor<0x15c>, &method_adaptor<0x15d>, &method_adaptor<0x15e>, &method_adaptor<0x15f>,
  &method_adaptor<0x160>, &method_adaptor<0x161>, &method_adaptor<0x162>, &method_adaptor<0x163>, &method_adaptor<0x164>, &method_adaptor<0x165>, &method_adaptor<0x166>, &method_adaptor<0x167>,
  &method_adaptor<0x168>, &method_adaptor<0x169>, &method_adaptor<0x16a>, &method_adaptor<0x16b>, &method_adaptor<0x16c>, &method_adaptor<0x16d>, &method_adaptor<0x16e>, &method_adaptor<0x16f>,
  &method_adaptor<0x170>, &method_adaptor<0x171>, &method_adaptor<0x172>, &method_adaptor<0x173>, &method_adaptor<0x174>, &method_adaptor<0x175>, &method_adaptor<0x176>, &method_adaptor<0x177>,
  &method_adaptor<0x178>, &method_adaptor<0x179>, &method_adaptor<0x17a>, &method_adaptor<0x17b>, &method_adaptor<0x17c>, &method_adaptor<0x17d>, &method_adaptor<0x17e>, &method_adaptor<0x17f>,
  &method_adaptor<0x180>, &method_adaptor<0x181>, &method_adaptor<0x182>, &method_adaptor<0x183>, &method_adaptor<0x184>, &method_adaptor<0x185>, &method_adaptor<0x186>, &method_adaptor<0x187>,
  &method_adaptor<0x188>, &method_adaptor<0x189>, &method_adaptor<0x18a>, &method_adaptor<0x18b>, &method_adaptor<0x18c>, &method_adaptor<0x18d>, &method_adaptor<0x18e>, &method_adaptor<0x18f>,
  &method_adaptor<0x190>, &method_adaptor<0x191>, &method_adaptor<0x192>, &method_adaptor<0x193>, &method_adaptor<0x194>, &method_adaptor<0x195>, &method_adaptor<0x196>, &method_adaptor<0x197>,
  &method_adaptor<0x198>, &method_adaptor<0x199>, &method_adaptor<0x19a>, &method_adaptor<0x19b>, &method_adaptor<0x19c>, &method_adaptor<0x19d>, &method_adaptor<0x19e>, &method_adaptor<0x19f>,
  &method_adaptor<0x1a0>, &method_adaptor<0x1a1>, &method_adaptor<0x1a2>, &method_adaptor<0x1a3>, &method_adaptor<0x1a4>, &method_adaptor<0x1a5>, &method_adaptor<0x1a6>, &method_adaptor<0x1a7>,
  &method_adaptor<0x1a8>, &method_adaptor<0x1a9>, &method_adaptor<0x1aa>, &method_adaptor<0x1ab>, &method_adaptor<0x1ac>, &method_adaptor<0x1ad>, &method_adaptor<0x1ae>, &method_adaptor<0x1af>,
  &method_adaptor<0x1b0>, &method_adaptor<0x1b1>, &method_adaptor<0x1b2>, &method_adaptor<0x1b3>, &method_adaptor<0x1b4>, &method_adaptor<0x1b5>, &method_adaptor<0x1b6>, &method_adaptor<0x1b7>,
  &method_adaptor<0x1b8>, &method_adaptor<0x1b9>, &method_adaptor<0x1ba>, &method_adaptor<0x1bb>, &method_adaptor<0x1bc>, &method_adaptor<0x1bd>, &method_adaptor<0x1be>, &method_adaptor<0x1bf>,
  &method_adaptor<0x1c0>, &method_adaptor<0x1c1>, &method_adaptor<0x1c2>, &method_adaptor<0x1c3>, &method_adaptor<0x1c4>, &method_adaptor<0x1c5>, &method_adaptor<0x1c6>, &method_adaptor<0x1c7>,
  &method_adaptor<0x1c8>, &method_adaptor<0x1c9>, &method_adaptor<0x1ca>, &method_adaptor<0x1cb>, &method_adaptor<0x1cc>, &method_adaptor<0x1cd>, &method_adaptor<0x1ce>, &method_adaptor<0x1cf>,
  &method_adaptor<0x1d0>, &method_adaptor<0x1d1>, &method_adaptor<0x1d2>, &method_adaptor<0x1d3>, &method_adaptor<0x1d4>, &method_adaptor<0x1d5>, &method_adaptor<0x1d6>, &method_adaptor<0x1d7>,
  &method_adaptor<0x1d8>, &method_adaptor<0x1d9>, &method_adaptor<0x1da>, &method_adaptor<0x1db>, &method_adaptor<0x1dc>, &method_adaptor<0x1dd>, &method_adaptor<0x1de>, &method_adaptor<0x1df>,
  &method_adaptor<0x1e0>, &method_adaptor<0x1e1>, &method_adaptor<0x1e2>, &method_adaptor<0x1e3>, &method_adaptor<0x1e4>, &method_adaptor<0x1e5>, &method_adaptor<0x1e6>, &method_adaptor<0x1e7>,
  &method_adaptor<0x1e8>, &method_adaptor<0x1e9>, &method_adaptor<0x1ea>, &method_adaptor<0x1eb>, &method_adaptor<0x1ec>, &method_adaptor<0x1ed>, &method_adaptor<0x1ee>, &method_adaptor<0x1ef>,
  &method_adaptor<0x1f0>, &method_adaptor<0x1f1>, &method_adaptor<0x1f2>, &method_adaptor<0x1f3>, &method_adaptor<0x1f4>, &method_adaptor<0x1f5>, &method_adaptor<0x1f6>, &method_adaptor<0x1f7>,
  &method_adaptor<0x1f8>, &method_adaptor<0x1f9>, &method_adaptor<0x1fa>, &method_adaptor<0x1fb>, &method_adaptor<0x1fc>, &method_adaptor<0x1fd>, &method_adaptor<0x1fe>, &method_adaptor<0x1ff>,
  &method_adaptor<0x200>, &method_adaptor<0x201>, &method_adaptor<0x202>, &method_adaptor<0x203>, &method_adaptor<0x204>, &method_adaptor<0x205>, &method_adaptor<0x206>, &method_adaptor<0x207>,
  &method_adaptor<0x208>, &method_adaptor<0x209>, &method_adaptor<0x20a>, &method_adaptor<0x20b>, &method_adaptor<0x20c>, &method_adaptor<0x20d>, &method_adaptor<0x20e>, &method_adaptor<0x20f>,
  &method_adaptor<0x210>, &method_adaptor<0x211>, &method_adaptor<0x212>, &method_adaptor<0x213>, &method_adaptor<0x214>, &method_adaptor<0x215>, &method_adaptor<0x216>, &method_adaptor<0x217>,
  &method_adaptor<0x218>, &method_adaptor<0x219>, &method_adaptor<0x21a>, &method_adaptor<0x21b>, &method_adaptor<0x21c>, &method_adaptor<0x21d>, &method_adaptor<0x21e>, &method_adaptor<0x21f>,
  &method_adaptor<0x220>, &method_adaptor<0x221>, &method_adaptor<0x222>, &method_adaptor<0x223>, &method_adaptor<0x224>, &method_adaptor<0x225>, &method_adaptor<0x226>, &method_adaptor<0x227>,
  &method_adaptor<0x228>, &method_adaptor<0x229>, &method_adaptor<0x22a>, &method_adaptor<0x22b>, &method_adaptor<0x22c>, &method_adaptor<0x22d>, &method_adaptor<0x22e>, &method_adaptor<0x22f>,
  &method_adaptor<0x230>, &method_adaptor<0x231>, &method_adaptor<0x232>, &method_adaptor<0x233>, &method_adaptor<0x234>, &method_adaptor<0x235>, &method_adaptor<0x236>, &method_adaptor<0x237>,
  &method_adaptor<0x238>, &method_adaptor<0x239>, &method_adaptor<0x23a>, &method_adaptor<0x23b>, &method_adaptor<0x23c>, &method_adaptor<0x23d>, &method_adaptor<0x23e>, &method_adaptor<0x23f>,
  &method_adaptor<0x240>, &method_adaptor<0x241>, &method_adaptor<0x242>, &method_adaptor<0x243>, &method_adaptor<0x244>, &method_adaptor<0x245>, &method_adaptor<0x246>, &method_adaptor<0x247>,
  &method_adaptor<0x248>, &method_adaptor<0x249>, &method_adaptor<0x24a>, &method_adaptor<0x24b>, &method_adaptor<0x24c>, &method_adaptor<0x24d>, &method_adaptor<0x24e>, &method_adaptor<0x24f>,
  &method_adaptor<0x250>, &method_adaptor<0x251>, &method_adaptor<0x252>, &method_adaptor<0x253>, &method_adaptor<0x254>, &method_adaptor<0x255>, &method_adaptor<0x256>, &method_adaptor<0x257>,
  &method_adaptor<0x258>, &method_adaptor<0x259>, &method_adaptor<0x25a>, &method_adaptor<0x25b>, &method_adaptor<0x25c>, &method_adaptor<0x25d>, &method_adaptor<0x25e>, &method_adaptor<0x25f>,
  &method_adaptor<0x260>, &method_adaptor<0x261>, &method_adaptor<0x262>, &method_adaptor<0x263>, &method_adaptor<0x264>, &method_adaptor<0x265>, &method_adaptor<0x266>, &method_adaptor<0x267>,
  &method_adaptor<0x268>, &method_adaptor<0x269>, &method_adaptor<0x26a>, &method_adaptor<0x26b>, &method_adaptor<0x26c>, &method_adaptor<0x26d>, &method_adaptor<0x26e>, &method_adaptor<0x26f>,
  &method_adaptor<0x270>, &method_adaptor<0x271>, &method_adaptor<0x272>, &method_adaptor<0x273>, &method_adaptor<0x274>, &method_adaptor<0x275>, &method_adaptor<0x276>, &method_adaptor<0x277>,
  &method_adaptor<0x278>, &method_adaptor<0x279>, &method_adaptor<0x27a>, &method_adaptor<0x27b>, &method_adaptor<0x27c>, &method_adaptor<0x27d>, &method_adaptor<0x27e>, &method_adaptor<0x27f>,
  &method_adaptor<0x280>, &method_adaptor<0x281>, &method_adaptor<0x282>, &method_adaptor<0x283>, &method_adaptor<0x284>, &method_adaptor<0x285>, &method_adaptor<0x286>, &method_adaptor<0x287>,
  &method_adaptor<0x288>, &method_adaptor<0x289>, &method_adaptor<0x28a>, &method_adaptor<0x28b>, &method_adaptor<0x28c>, &method_adaptor<0x28d>, &method_adaptor<0x28e>, &method_adaptor<0x28f>,
  &method_adaptor<0x290>, &method_adaptor<0x291>, &method_adaptor<0x292>, &method_adaptor<0x293>, &method_adaptor<0x294>, &method_adaptor<0x295>, &method_adaptor<0x296>, &method_adaptor<0x297>,
  &method_adaptor<0x298>, &method_adaptor<0x299>, &method_adaptor<0x29a>, &method_adaptor<0x29b>, &method_adaptor<0x29c>, &method_adaptor<0x29d>, &method_adaptor<0x29e>, &method_adaptor<0x29f>,
  &method_adaptor<0x2a0>, &method_adaptor<0x2a1>, &method_adaptor<0x2a2>, &method_adaptor<0x2a3>, &method_adaptor<0x2a4>, &method_adaptor<0x2a5>, &method_adaptor<0x2a6>, &method_adaptor<0x2a7>,
  &method_adaptor<0x2a8>, &method_adaptor<0x2a9>, &method_adaptor<0x2aa>, &method_adaptor<0x2ab>, &method_adaptor<0x2ac>, &method_adaptor<0x2ad>, &method_adaptor<0x2ae>, &method_adaptor<0x2af>,
  &method_adaptor<0x2b0>, &method_adaptor<0x2b1>, &method_adaptor<0x2b2>, &method_adaptor<0x2b3>, &method_adaptor<0x2b4>, &method_adaptor<0x2b5>, &method_adaptor<0x2b6>, &method_adaptor<0x2b7>,
  &method_adaptor<0x2b8>, &method_adaptor<0x2b9>, &method_adaptor<0x2ba>, &method_adaptor<0x2bb>, &method_adaptor<0x2bc>, &method_adaptor<0x2bd>, &method_adaptor<0x2be>, &method_adaptor<0x2bf>,
  &method_adaptor<0x2c0>, &method_adaptor<0x2c1>, &method_adaptor<0x2c2>, &method_adaptor<0x2c3>, &method_adaptor<0x2c4>, &method_adaptor<0x2c5>, &method_adaptor<0x2c6>, &method_adaptor<0x2c7>,
  &method_adaptor<0x2c8>, &method_adaptor<0x2c9>, &method_adaptor<0x2ca>, &method_adaptor<0x2cb>, &method_adaptor<0x2cc>, &method_adaptor<0x2cd>, &method_adaptor<0x2ce>, &method_adaptor<0x2cf>,
  &method_adaptor<0x2d0>, &method_adaptor<0x2d1>, &method_adaptor<0x2d2>, &method_adaptor<0x2d3>, &method_adaptor<0x2d4>, &method_adaptor<0x2d5>, &method_adaptor<0x2d6>, &method_adaptor<0x2d7>,
  &method_adaptor<0x2d8>, &method_adaptor<0x2d9>, &method_adaptor<0x2da>, &method_adaptor<0x2db>, &method_adaptor<0x2dc>, &method_adaptor<0x2dd>, &method_adaptor<0x2de>, &method_adaptor<0x2df>,
  &method_adaptor<0x2e0>, &method_adaptor<0x2e1>, &method_adaptor<0x2e2>, &method_adaptor<0x2e3>, &method_adaptor<0x2e4>, &method_adaptor<0x2e5>, &method_adaptor<0x2e6>, &method_adaptor<0x2e7>,
  &method_adaptor<0x2e8>, &method_adaptor<0x2e9>, &method_adaptor<0x2ea>, &method_adaptor<0x2eb>, &method_adaptor<0x2ec>, &method_adaptor<0x2ed>, &method_adaptor<0x2ee>, &method_adaptor<0x2ef>,
  &method_adaptor<0x2f0>, &method_adaptor<0x2f1>, &method_adaptor<0x2f2>, &method_adaptor<0x2f3>, &method_adaptor<0x2f4>, &method_adaptor<0x2f5>, &method_adaptor<0x2f6>, &method_adaptor<0x2f7>,
  &method_adaptor<0x2f8>, &method_adaptor<0x2f9>, &method_adaptor<0x2fa>, &method_adaptor<0x2fb>, &method_adaptor<0x2fc>, &method_adaptor<0x2fd>, &method_adaptor<0x2fe>, &method_adaptor<0x2ff>,
  &method_adaptor<0x300>, &method_adaptor<0x301>, &method_adaptor<0x302>, &method_adaptor<0x303>, &method_adaptor<0x304>, &method_adaptor<0x305>, &method_adaptor<0x306>, &method_adaptor<0x307>,
  &method_adaptor<0x308>, &method_adaptor<0x309>, &method_adaptor<0x30a>, &method_adaptor<0x30b>, &method_adaptor<0x30c>, &method_adaptor<0x30d>, &method_adaptor<0x30e>, &method_adaptor<0x30f>,
  &method_adaptor<0x310>, &method_adaptor<0x311>, &method_adaptor<0x312>, &method_adaptor<0x313>, &method_adaptor<0x314>, &method_adaptor<0x315>, &method_adaptor<0x316>, &method_adaptor<0x317>,
  &method_adaptor<0x318>, &method_adaptor<0x319>, &method_adaptor<0x31a>, &method_adaptor<0x31b>, &method_adaptor<0x31c>, &method_adaptor<0x31d>, &method_adaptor<0x31e>, &method_adaptor<0x31f>,
  &method_adaptor<0x320>, &method_adaptor<0x321>, &method_adaptor<0x322>, &method_adaptor<0x323>, &method_adaptor<0x324>, &method_adaptor<0x325>, &method_adaptor<0x326>, &method_adaptor<0x327>,
  &method_adaptor<0x328>, &method_adaptor<0x329>, &method_adaptor<0x32a>, &method_adaptor<0x32b>, &method_adaptor<0x32c>, &method_adaptor<0x32d>, &method_adaptor<0x32e>, &method_adaptor<0x32f>,
  &method_adaptor<0x330>, &method_adaptor<0x331>, &method_adaptor<0x332>, &method_adaptor<0x333>, &method_adaptor<0x334>, &method_adaptor<0x335>, &method_adaptor<0x336>, &method_adaptor<0x337>,
  &method_adaptor<0x338>, &method_adaptor<0x339>, &method_adaptor<0x33a>, &method_adaptor<0x33b>, &method_adaptor<0x33c>, &method_adaptor<0x33d>, &method_adaptor<0x33e>, &method_adaptor<0x33f>,
  &method_adaptor<0x340>, &method_adaptor<0x341>, &method_adaptor<0x342>, &method_adaptor<0x343>, &method_adaptor<0x344>, &method_adaptor<0x345>, &method_adaptor<0x346>, &method_adaptor<0x347>,
  &method_adaptor<0x348>, &method_adaptor<0x349>, &method_adaptor<0x34a>, &method_adaptor<0x34b>, &method_adaptor<0x34c>, &method_adaptor<0x34d>, &method_adaptor<0x34e>, &method_adaptor<0x34f>,
  &method_adaptor<0x350>, &method_adaptor<0x351>, &method_adaptor<0x352>, &method_adaptor<0x353>, &method_adaptor<0x354>, &method_adaptor<0x355>, &method_adaptor<0x356>, &method_adaptor<0x357>,
  &method_adaptor<0x358>, &method_adaptor<0x359>, &method_adaptor<0x35a>, &method_adaptor<0x35b>, &method_adaptor<0x35c>, &method_adaptor<0x35d>, &method_adaptor<0x35e>, &method_adaptor<0x35f>,
  &method_adaptor<0x360>, &method_adaptor<0x361>, &method_adaptor<0x362>, &method_adaptor<0x363>, &method_adaptor<0x364>, &method_adaptor<0x365>, &method_adaptor<0x366>, &method_adaptor<0x367>,
  &method_adaptor<0x368>, &method_adaptor<0x369>, &method_adaptor<0x36a>, &method_adaptor<0x36b>, &method_adaptor<0x36c>, &method_adaptor<0x36d>, &method_adaptor<0x36e>, &method_adaptor<0x36f>,
  &method_adaptor<0x370>, &method_adaptor<0x371>, &method_adaptor<0x372>, &method_adaptor<0x373>, &method_adaptor<0x374>, &method_adaptor<0x375>, &method_adaptor<0x376>, &method_adaptor<0x377>,
  &method_adaptor<0x378>, &method_adaptor<0x379>, &method_adaptor<0x37a>, &method_adaptor<0x37b>, &method_adaptor<0x37c>, &method_adaptor<0x37d>, &method_adaptor<0x37e>, &method_adaptor<0x37f>,
  &method_adaptor<0x380>, &method_adaptor<0x381>, &method_adaptor<0x382>, &method_adaptor<0x383>, &method_adaptor<0x384>, &method_adaptor<0x385>, &method_adaptor<0x386>, &method_adaptor<0x387>,
  &method_adaptor<0x388>, &method_adaptor<0x389>, &method_adaptor<0x38a>, &method_adaptor<0x38b>, &method_adaptor<0x38c>, &method_adaptor<0x38d>, &method_adaptor<0x38e>, &method_adaptor<0x38f>,
  &method_adaptor<0x390>, &method_adaptor<0x391>, &method_adaptor<0x392>, &method_adaptor<0x393>, &method_adaptor<0x394>, &method_adaptor<0x395>, &method_adaptor<0x396>, &method_adaptor<0x397>,
  &method_adaptor<0x398>, &method_adaptor<0x399>, &method_adaptor<0x39a>, &method_adaptor<0x39b>, &method_adaptor<0x39c>, &method_adaptor<0x39d>, &method_adaptor<0x39e>, &method_adaptor<0x39f>,
  &method_adaptor<0x3a0>, &method_adaptor<0x3a1>, &method_adaptor<0x3a2>, &method_adaptor<0x3a3>, &method_adaptor<0x3a4>, &method_adaptor<0x3a5>, &method_adaptor<0x3a6>, &method_adaptor<0x3a7>,
  &method_adaptor<0x3a8>, &method_adaptor<0x3a9>, &method_adaptor<0x3aa>, &method_adaptor<0x3ab>, &method_adaptor<0x3ac>, &method_adaptor<0x3ad>, &method_adaptor<0x3ae>, &method_adaptor<0x3af>,
  &method_adaptor<0x3b0>, &method_adaptor<0x3b1>, &method_adaptor<0x3b2>, &method_adaptor<0x3b3>, &method_adaptor<0x3b4>, &method_adaptor<0x3b5>, &method_adaptor<0x3b6>, &method_adaptor<0x3b7>,
  &method_adaptor<0x3b8>, &method_adaptor<0x3b9>, &method_adaptor<0x3ba>, &method_adaptor<0x3bb>, &method_adaptor<0x3bc>, &method_adaptor<0x3bd>, &method_adaptor<0x3be>, &method_adaptor<0x3bf>,
  &method_adaptor<0x3c0>, &method_adaptor<0x3c1>, &method_adaptor<0x3c2>, &method_adaptor<0x3c3>, &method_adaptor<0x3c4>, &method_adaptor<0x3c5>, &method_adaptor<0x3c6>, &method_adaptor<0x3c7>,
  &method_adaptor<0x3c8>, &method_adaptor<0x3c9>, &method_adaptor<0x3ca>, &method_adaptor<0x3cb>, &method_adaptor<0x3cc>, &method_adaptor<0x3cd>, &method_adaptor<0x3ce>, &method_adaptor<0x3cf>,
  &method_adaptor<0x3d0>, &method_adaptor<0x3d1>, &method_adaptor<0x3d2>, &method_adaptor<0x3d3>, &method_adaptor<0x3d4>, &method_adaptor<0x3d5>, &method_adaptor<0x3d6>, &method_adaptor<0x3d7>,
  &method_adaptor<0x3d8>, &method_adaptor<0x3d9>, &method_adaptor<0x3da>, &method_adaptor<0x3db>, &method_adaptor<0x3dc>, &method_adaptor<0x3dd>, &method_adaptor<0x3de>, &method_adaptor<0x3df>,
  &method_adaptor<0x3e0>, &method_adaptor<0x3e1>, &method_adaptor<0x3e2>, &method_adaptor<0x3e3>, &method_adaptor<0x3e4>, &method_adaptor<0x3e5>, &method_adaptor<0x3e6>, &method_adaptor<0x3e7>,
  &method_adaptor<0x3e8>, &method_adaptor<0x3e9>, &method_adaptor<0x3ea>, &method_adaptor<0x3eb>, &method_adaptor<0x3ec>, &method_adaptor<0x3ed>, &method_adaptor<0x3ee>, &method_adaptor<0x3ef>,
  &method_adaptor<0x3f0>, &method_adaptor<0x3f1>, &method_adaptor<0x3f2>, &method_adaptor<0x3f3>, &method_adaptor<0x3f4>, &method_adaptor<0x3f5>, &method_adaptor<0x3f6>, &method_adaptor<0x3f7>,
  &method_adaptor<0x3f8>, &method_adaptor<0x3f9>, &method_adaptor<0x3fa>, &method_adaptor<0x3fb>, &method_adaptor<0x3fc>, &method_adaptor<0x3fd>, &method_adaptor<0x3fe>, &method_adaptor<0x3ff>,
};

//  zero-arguments method adaptors

template <int N> 
VALUE method_adaptor_ctor (int argc, VALUE *argv, VALUE self)
{
  VALUE ret = Qnil;

  RBA_TRY

    ret = method_adaptor (N, argc, argv, self, true);

  RBA_CATCH(method_name_from_id (N, self))

  return ret;
}

VALUE (*(method_adaptors_ctor [])) (int, VALUE *, VALUE) =
{
  &method_adaptor_ctor<0x000>, &method_adaptor_ctor<0x001>, &method_adaptor_ctor<0x002>, &method_adaptor_ctor<0x003>, &method_adaptor_ctor<0x004>, &method_adaptor_ctor<0x005>, &method_adaptor_ctor<0x006>, &method_adaptor_ctor<0x007>,
  &method_adaptor_ctor<0x008>, &method_adaptor_ctor<0x009>, &method_adaptor_ctor<0x00a>, &method_adaptor_ctor<0x00b>, &method_adaptor_ctor<0x00c>, &method_adaptor_ctor<0x00d>, &method_adaptor_ctor<0x00e>, &method_adaptor_ctor<0x00f>,
  &method_adaptor_ctor<0x010>, &method_adaptor_ctor<0x011>, &method_adaptor_ctor<0x012>, &method_adaptor_ctor<0x013>, &method_adaptor_ctor<0x014>, &method_adaptor_ctor<0x015>, &method_adaptor_ctor<0x016>, &method_adaptor_ctor<0x017>,
  &method_adaptor_ctor<0x018>, &method_adaptor_ctor<0x019>, &method_adaptor_ctor<0x01a>, &method_adaptor_ctor<0x01b>, &method_adaptor_ctor<0x01c>, &method_adaptor_ctor<0x01d>, &method_adaptor_ctor<0x01e>, &method_adaptor_ctor<0x01f>,
  &method_adaptor_ctor<0x020>, &method_adaptor_ctor<0x021>, &method_adaptor_ctor<0x022>, &method_adaptor_ctor<0x023>, &method_adaptor_ctor<0x024>, &method_adaptor_ctor<0x025>, &method_adaptor_ctor<0x026>, &method_adaptor_ctor<0x027>,
  &method_adaptor_ctor<0x028>, &method_adaptor_ctor<0x029>, &method_adaptor_ctor<0x02a>, &method_adaptor_ctor<0x02b>, &method_adaptor_ctor<0x02c>, &method_adaptor_ctor<0x02d>, &method_adaptor_ctor<0x02e>, &method_adaptor_ctor<0x02f>,
  &method_adaptor_ctor<0x030>, &method_adaptor_ctor<0x031>, &method_adaptor_ctor<0x032>, &method_adaptor_ctor<0x033>, &method_adaptor_ctor<0x034>, &method_adaptor_ctor<0x035>, &method_adaptor_ctor<0x036>, &method_adaptor_ctor<0x037>,
  &method_adaptor_ctor<0x038>, &method_adaptor_ctor<0x039>, &method_adaptor_ctor<0x03a>, &method_adaptor_ctor<0x03b>, &method_adaptor_ctor<0x03c>, &method_adaptor_ctor<0x03d>, &method_adaptor_ctor<0x03e>, &method_adaptor_ctor<0x03f>,
  &method_adaptor_ctor<0x040>, &method_adaptor_ctor<0x041>, &method_adaptor_ctor<0x042>, &method_adaptor_ctor<0x043>, &method_adaptor_ctor<0x044>, &method_adaptor_ctor<0x045>, &method_adaptor_ctor<0x046>, &method_adaptor_ctor<0x047>,
  &method_adaptor_ctor<0x048>, &method_adaptor_ctor<0x049>, &method_adaptor_ctor<0x04a>, &method_adaptor_ctor<0x04b>, &method_adaptor_ctor<0x04c>, &method_adaptor_ctor<0x04d>, &method_adaptor_ctor<0x04e>, &method_adaptor_ctor<0x04f>,
  &method_adaptor_ctor<0x050>, &method_adaptor_ctor<0x051>, &method_adaptor_ctor<0x052>, &method_adaptor_ctor<0x053>, &method_adaptor_ctor<0x054>, &method_adaptor_ctor<0x055>, &method_adaptor_ctor<0x056>, &method_adaptor_ctor<0x057>,
  &method_adaptor_ctor<0x058>, &method_adaptor_ctor<0x059>, &method_adaptor_ctor<0x05a>, &method_adaptor_ctor<0x05b>, &method_adaptor_ctor<0x05c>, &method_adaptor_ctor<0x05d>, &method_adaptor_ctor<0x05e>, &method_adaptor_ctor<0x05f>,
  &method_adaptor_ctor<0x060>, &method_adaptor_ctor<0x061>, &method_adaptor_ctor<0x062>, &method_adaptor_ctor<0x063>, &method_adaptor_ctor<0x064>, &method_adaptor_ctor<0x065>, &method_adaptor_ctor<0x066>, &method_adaptor_ctor<0x067>,
  &method_adaptor_ctor<0x068>, &method_adaptor_ctor<0x069>, &method_adaptor_ctor<0x06a>, &method_adaptor_ctor<0x06b>, &method_adaptor_ctor<0x06c>, &method_adaptor_ctor<0x06d>, &method_adaptor_ctor<0x06e>, &method_adaptor_ctor<0x06f>,
  &method_adaptor_ctor<0x070>, &method_adaptor_ctor<0x071>, &method_adaptor_ctor<0x072>, &method_adaptor_ctor<0x073>, &method_adaptor_ctor<0x074>, &method_adaptor_ctor<0x075>, &method_adaptor_ctor<0x076>, &method_adaptor_ctor<0x077>,
  &method_adaptor_ctor<0x078>, &method_adaptor_ctor<0x079>, &method_adaptor_ctor<0x07a>, &method_adaptor_ctor<0x07b>, &method_adaptor_ctor<0x07c>, &method_adaptor_ctor<0x07d>, &method_adaptor_ctor<0x07e>, &method_adaptor_ctor<0x07f>,
  &method_adaptor_ctor<0x080>, &method_adaptor_ctor<0x081>, &method_adaptor_ctor<0x082>, &method_adaptor_ctor<0x083>, &method_adaptor_ctor<0x084>, &method_adaptor_ctor<0x085>, &method_adaptor_ctor<0x086>, &method_adaptor_ctor<0x087>,
  &method_adaptor_ctor<0x088>, &method_adaptor_ctor<0x089>, &method_adaptor_ctor<0x08a>, &method_adaptor_ctor<0x08b>, &method_adaptor_ctor<0x08c>, &method_adaptor_ctor<0x08d>, &method_adaptor_ctor<0x08e>, &method_adaptor_ctor<0x08f>,
  &method_adaptor_ctor<0x090>, &method_adaptor_ctor<0x091>, &method_adaptor_ctor<0x092>, &method_adaptor_ctor<0x093>, &method_adaptor_ctor<0x094>, &method_adaptor_ctor<0x095>, &method_adaptor_ctor<0x096>, &method_adaptor_ctor<0x097>,
  &method_adaptor_ctor<0x098>, &method_adaptor_ctor<0x099>, &method_adaptor_ctor<0x09a>, &method_adaptor_ctor<0x09b>, &method_adaptor_ctor<0x09c>, &method_adaptor_ctor<0x09d>, &method_adaptor_ctor<0x09e>, &method_adaptor_ctor<0x09f>,
  &method_adaptor_ctor<0x0a0>, &method_adaptor_ctor<0x0a1>, &method_adaptor_ctor<0x0a2>, &method_adaptor_ctor<0x0a3>, &method_adaptor_ctor<0x0a4>, &method_adaptor_ctor<0x0a5>, &method_adaptor_ctor<0x0a6>, &method_adaptor_ctor<0x0a7>,
  &method_adaptor_ctor<0x0a8>, &method_adaptor_ctor<0x0a9>, &method_adaptor_ctor<0x0aa>, &method_adaptor_ctor<0x0ab>, &method_adaptor_ctor<0x0ac>, &method_adaptor_ctor<0x0ad>, &method_adaptor_ctor<0x0ae>, &method_adaptor_ctor<0x0af>,
  &method_adaptor_ctor<0x0b0>, &method_adaptor_ctor<0x0b1>, &method_adaptor_ctor<0x0b2>, &method_adaptor_ctor<0x0b3>, &method_adaptor_ctor<0x0b4>, &method_adaptor_ctor<0x0b5>, &method_adaptor_ctor<0x0b6>, &method_adaptor_ctor<0x0b7>,
  &method_adaptor_ctor<0x0b8>, &method_adaptor_ctor<0x0b9>, &method_adaptor_ctor<0x0ba>, &method_adaptor_ctor<0x0bb>, &method_adaptor_ctor<0x0bc>, &method_adaptor_ctor<0x0bd>, &method_adaptor_ctor<0x0be>, &method_adaptor_ctor<0x0bf>,
  &method_adaptor_ctor<0x0c0>, &method_adaptor_ctor<0x0c1>, &method_adaptor_ctor<0x0c2>, &method_adaptor_ctor<0x0c3>, &method_adaptor_ctor<0x0c4>, &method_adaptor_ctor<0x0c5>, &method_adaptor_ctor<0x0c6>, &method_adaptor_ctor<0x0c7>,
  &method_adaptor_ctor<0x0c8>, &method_adaptor_ctor<0x0c9>, &method_adaptor_ctor<0x0ca>, &method_adaptor_ctor<0x0cb>, &method_adaptor_ctor<0x0cc>, &method_adaptor_ctor<0x0cd>, &method_adaptor_ctor<0x0ce>, &method_adaptor_ctor<0x0cf>,
  &method_adaptor_ctor<0x0d0>, &method_adaptor_ctor<0x0d1>, &method_adaptor_ctor<0x0d2>, &method_adaptor_ctor<0x0d3>, &method_adaptor_ctor<0x0d4>, &method_adaptor_ctor<0x0d5>, &method_adaptor_ctor<0x0d6>, &method_adaptor_ctor<0x0d7>,
  &method_adaptor_ctor<0x0d8>, &method_adaptor_ctor<0x0d9>, &method_adaptor_ctor<0x0da>, &method_adaptor_ctor<0x0db>, &method_adaptor_ctor<0x0dc>, &method_adaptor_ctor<0x0dd>, &method_adaptor_ctor<0x0de>, &method_adaptor_ctor<0x0df>,
  &method_adaptor_ctor<0x0e0>, &method_adaptor_ctor<0x0e1>, &method_adaptor_ctor<0x0e2>, &method_adaptor_ctor<0x0e3>, &method_adaptor_ctor<0x0e4>, &method_adaptor_ctor<0x0e5>, &method_adaptor_ctor<0x0e6>, &method_adaptor_ctor<0x0e7>,
  &method_adaptor_ctor<0x0e8>, &method_adaptor_ctor<0x0e9>, &method_adaptor_ctor<0x0ea>, &method_adaptor_ctor<0x0eb>, &method_adaptor_ctor<0x0ec>, &method_adaptor_ctor<0x0ed>, &method_adaptor_ctor<0x0ee>, &method_adaptor_ctor<0x0ef>,
  &method_adaptor_ctor<0x0f0>, &method_adaptor_ctor<0x0f1>, &method_adaptor_ctor<0x0f2>, &method_adaptor_ctor<0x0f3>, &method_adaptor_ctor<0x0f4>, &method_adaptor_ctor<0x0f5>, &method_adaptor_ctor<0x0f6>, &method_adaptor_ctor<0x0f7>,
  &method_adaptor_ctor<0x0f8>, &method_adaptor_ctor<0x0f9>, &method_adaptor_ctor<0x0fa>, &method_adaptor_ctor<0x0fb>, &method_adaptor_ctor<0x0fc>, &method_adaptor_ctor<0x0fd>, &method_adaptor_ctor<0x0fe>, &method_adaptor_ctor<0x0ff>,
  &method_adaptor_ctor<0x100>, &method_adaptor_ctor<0x101>, &method_adaptor_ctor<0x102>, &method_adaptor_ctor<0x103>, &method_adaptor_ctor<0x104>, &method_adaptor_ctor<0x105>, &method_adaptor_ctor<0x106>, &method_adaptor_ctor<0x107>,
  &method_adaptor_ctor<0x108>, &method_adaptor_ctor<0x109>, &method_adaptor_ctor<0x10a>, &method_adaptor_ctor<0x10b>, &method_adaptor_ctor<0x10c>, &method_adaptor_ctor<0x10d>, &method_adaptor_ctor<0x10e>, &method_adaptor_ctor<0x10f>,
  &method_adaptor_ctor<0x110>, &method_adaptor_ctor<0x111>, &method_adaptor_ctor<0x112>, &method_adaptor_ctor<0x113>, &method_adaptor_ctor<0x114>, &method_adaptor_ctor<0x115>, &method_adaptor_ctor<0x116>, &method_adaptor_ctor<0x117>,
  &method_adaptor_ctor<0x118>, &method_adaptor_ctor<0x119>, &method_adaptor_ctor<0x11a>, &method_adaptor_ctor<0x11b>, &method_adaptor_ctor<0x11c>, &method_adaptor_ctor<0x11d>, &method_adaptor_ctor<0x11e>, &method_adaptor_ctor<0x11f>,
  &method_adaptor_ctor<0x120>, &method_adaptor_ctor<0x121>, &method_adaptor_ctor<0x122>, &method_adaptor_ctor<0x123>, &method_adaptor_ctor<0x124>, &method_adaptor_ctor<0x125>, &method_adaptor_ctor<0x126>, &method_adaptor_ctor<0x127>,
  &method_adaptor_ctor<0x128>, &method_adaptor_ctor<0x129>, &method_adaptor_ctor<0x12a>, &method_adaptor_ctor<0x12b>, &method_adaptor_ctor<0x12c>, &method_adaptor_ctor<0x12d>, &method_adaptor_ctor<0x12e>, &method_adaptor_ctor<0x12f>,
  &method_adaptor_ctor<0x130>, &method_adaptor_ctor<0x131>, &method_adaptor_ctor<0x132>, &method_adaptor_ctor<0x133>, &method_adaptor_ctor<0x134>, &method_adaptor_ctor<0x135>, &method_adaptor_ctor<0x136>, &method_adaptor_ctor<0x137>,
  &method_adaptor_ctor<0x138>, &method_adaptor_ctor<0x139>, &method_adaptor_ctor<0x13a>, &method_adaptor_ctor<0x13b>, &method_adaptor_ctor<0x13c>, &method_adaptor_ctor<0x13d>, &method_adaptor_ctor<0x13e>, &method_adaptor_ctor<0x13f>,
  &method_adaptor_ctor<0x140>, &method_adaptor_ctor<0x141>, &method_adaptor_ctor<0x142>, &method_adaptor_ctor<0x143>, &method_adaptor_ctor<0x144>, &method_adaptor_ctor<0x145>, &method_adaptor_ctor<0x146>, &method_adaptor_ctor<0x147>,
  &method_adaptor_ctor<0x148>, &method_adaptor_ctor<0x149>, &method_adaptor_ctor<0x14a>, &method_adaptor_ctor<0x14b>, &method_adaptor_ctor<0x14c>, &method_adaptor_ctor<0x14d>, &method_adaptor_ctor<0x14e>, &method_adaptor_ctor<0x14f>,
  &method_adaptor_ctor<0x150>, &method_adaptor_ctor<0x151>, &method_adaptor_ctor<0x152>, &method_adaptor_ctor<0x153>, &method_adaptor_ctor<0x154>, &method_adaptor_ctor<0x155>, &method_adaptor_ctor<0x156>, &method_adaptor_ctor<0x157>,
  &method_adaptor_ctor<0x158>, &method_adaptor_ctor<0x159>, &method_adaptor_ctor<0x15a>, &method_adaptor_ctor<0x15b>, &method_adaptor_ctor<0x15c>, &method_adaptor_ctor<0x15d>, &method_adaptor_ctor<0x15e>, &method_adaptor_ctor<0x15f>,
  &method_adaptor_ctor<0x160>, &method_adaptor_ctor<0x161>, &method_adaptor_ctor<0x162>, &method_adaptor_ctor<0x163>, &method_adaptor_ctor<0x164>, &method_adaptor_ctor<0x165>, &method_adaptor_ctor<0x166>, &method_adaptor_ctor<0x167>,
  &method_adaptor_ctor<0x168>, &method_adaptor_ctor<0x169>, &method_adaptor_ctor<0x16a>, &method_adaptor_ctor<0x16b>, &method_adaptor_ctor<0x16c>, &method_adaptor_ctor<0x16d>, &method_adaptor_ctor<0x16e>, &method_adaptor_ctor<0x16f>,
  &method_adaptor_ctor<0x170>, &method_adaptor_ctor<0x171>, &method_adaptor_ctor<0x172>, &method_adaptor_ctor<0x173>, &method_adaptor_ctor<0x174>, &method_adaptor_ctor<0x175>, &method_adaptor_ctor<0x176>, &method_adaptor_ctor<0x177>,
  &method_adaptor_ctor<0x178>, &method_adaptor_ctor<0x179>, &method_adaptor_ctor<0x17a>, &method_adaptor_ctor<0x17b>, &method_adaptor_ctor<0x17c>, &method_adaptor_ctor<0x17d>, &method_adaptor_ctor<0x17e>, &method_adaptor_ctor<0x17f>,
  &method_adaptor_ctor<0x180>, &method_adaptor_ctor<0x181>, &method_adaptor_ctor<0x182>, &method_adaptor_ctor<0x183>, &method_adaptor_ctor<0x184>, &method_adaptor_ctor<0x185>, &method_adaptor_ctor<0x186>, &method_adaptor_ctor<0x187>,
  &method_adaptor_ctor<0x188>, &method_adaptor_ctor<0x189>, &method_adaptor_ctor<0x18a>, &method_adaptor_ctor<0x18b>, &method_adaptor_ctor<0x18c>, &method_adaptor_ctor<0x18d>, &method_adaptor_ctor<0x18e>, &method_adaptor_ctor<0x18f>,
  &method_adaptor_ctor<0x190>, &method_adaptor_ctor<0x191>, &method_adaptor_ctor<0x192>, &method_adaptor_ctor<0x193>, &method_adaptor_ctor<0x194>, &method_adaptor_ctor<0x195>, &method_adaptor_ctor<0x196>, &method_adaptor_ctor<0x197>,
  &method_adaptor_ctor<0x198>, &method_adaptor_ctor<0x199>, &method_adaptor_ctor<0x19a>, &method_adaptor_ctor<0x19b>, &method_adaptor_ctor<0x19c>, &method_adaptor_ctor<0x19d>, &method_adaptor_ctor<0x19e>, &method_adaptor_ctor<0x19f>,
  &method_adaptor_ctor<0x1a0>, &method_adaptor_ctor<0x1a1>, &method_adaptor_ctor<0x1a2>, &method_adaptor_ctor<0x1a3>, &method_adaptor_ctor<0x1a4>, &method_adaptor_ctor<0x1a5>, &method_adaptor_ctor<0x1a6>, &method_adaptor_ctor<0x1a7>,
  &method_adaptor_ctor<0x1a8>, &method_adaptor_ctor<0x1a9>, &method_adaptor_ctor<0x1aa>, &method_adaptor_ctor<0x1ab>, &method_adaptor_ctor<0x1ac>, &method_adaptor_ctor<0x1ad>, &method_adaptor_ctor<0x1ae>, &method_adaptor_ctor<0x1af>,
  &method_adaptor_ctor<0x1b0>, &method_adaptor_ctor<0x1b1>, &method_adaptor_ctor<0x1b2>, &method_adaptor_ctor<0x1b3>, &method_adaptor_ctor<0x1b4>, &method_adaptor_ctor<0x1b5>, &method_adaptor_ctor<0x1b6>, &method_adaptor_ctor<0x1b7>,
  &method_adaptor_ctor<0x1b8>, &method_adaptor_ctor<0x1b9>, &method_adaptor_ctor<0x1ba>, &method_adaptor_ctor<0x1bb>, &method_adaptor_ctor<0x1bc>, &method_adaptor_ctor<0x1bd>, &method_adaptor_ctor<0x1be>, &method_adaptor_ctor<0x1bf>,
  &method_adaptor_ctor<0x1c0>, &method_adaptor_ctor<0x1c1>, &method_adaptor_ctor<0x1c2>, &method_adaptor_ctor<0x1c3>, &method_adaptor_ctor<0x1c4>, &method_adaptor_ctor<0x1c5>, &method_adaptor_ctor<0x1c6>, &method_adaptor_ctor<0x1c7>,
  &method_adaptor_ctor<0x1c8>, &method_adaptor_ctor<0x1c9>, &method_adaptor_ctor<0x1ca>, &method_adaptor_ctor<0x1cb>, &method_adaptor_ctor<0x1cc>, &method_adaptor_ctor<0x1cd>, &method_adaptor_ctor<0x1ce>, &method_adaptor_ctor<0x1cf>,
  &method_adaptor_ctor<0x1d0>, &method_adaptor_ctor<0x1d1>, &method_adaptor_ctor<0x1d2>, &method_adaptor_ctor<0x1d3>, &method_adaptor_ctor<0x1d4>, &method_adaptor_ctor<0x1d5>, &method_adaptor_ctor<0x1d6>, &method_adaptor_ctor<0x1d7>,
  &method_adaptor_ctor<0x1d8>, &method_adaptor_ctor<0x1d9>, &method_adaptor_ctor<0x1da>, &method_adaptor_ctor<0x1db>, &method_adaptor_ctor<0x1dc>, &method_adaptor_ctor<0x1dd>, &method_adaptor_ctor<0x1de>, &method_adaptor_ctor<0x1df>,
  &method_adaptor_ctor<0x1e0>, &method_adaptor_ctor<0x1e1>, &method_adaptor_ctor<0x1e2>, &method_adaptor_ctor<0x1e3>, &method_adaptor_ctor<0x1e4>, &method_adaptor_ctor<0x1e5>, &method_adaptor_ctor<0x1e6>, &method_adaptor_ctor<0x1e7>,
  &method_adaptor_ctor<0x1e8>, &method_adaptor_ctor<0x1e9>, &method_adaptor_ctor<0x1ea>, &method_adaptor_ctor<0x1eb>, &method_adaptor_ctor<0x1ec>, &method_adaptor_ctor<0x1ed>, &method_adaptor_ctor<0x1ee>, &method_adaptor_ctor<0x1ef>,
  &method_adaptor_ctor<0x1f0>, &method_adaptor_ctor<0x1f1>, &method_adaptor_ctor<0x1f2>, &method_adaptor_ctor<0x1f3>, &method_adaptor_ctor<0x1f4>, &method_adaptor_ctor<0x1f5>, &method_adaptor_ctor<0x1f6>, &method_adaptor_ctor<0x1f7>,
  &method_adaptor_ctor<0x1f8>, &method_adaptor_ctor<0x1f9>, &method_adaptor_ctor<0x1fa>, &method_adaptor_ctor<0x1fb>, &method_adaptor_ctor<0x1fc>, &method_adaptor_ctor<0x1fd>, &method_adaptor_ctor<0x1fe>, &method_adaptor_ctor<0x1ff>,
  &method_adaptor_ctor<0x200>, &method_adaptor_ctor<0x201>, &method_adaptor_ctor<0x202>, &method_adaptor_ctor<0x203>, &method_adaptor_ctor<0x204>, &method_adaptor_ctor<0x205>, &method_adaptor_ctor<0x206>, &method_adaptor_ctor<0x207>,
  &method_adaptor_ctor<0x208>, &method_adaptor_ctor<0x209>, &method_adaptor_ctor<0x20a>, &method_adaptor_ctor<0x20b>, &method_adaptor_ctor<0x20c>, &method_adaptor_ctor<0x20d>, &method_adaptor_ctor<0x20e>, &method_adaptor_ctor<0x20f>,
  &method_adaptor_ctor<0x210>, &method_adaptor_ctor<0x211>, &method_adaptor_ctor<0x212>, &method_adaptor_ctor<0x213>, &method_adaptor_ctor<0x214>, &method_adaptor_ctor<0x215>, &method_adaptor_ctor<0x216>, &method_adaptor_ctor<0x217>,
  &method_adaptor_ctor<0x218>, &method_adaptor_ctor<0x219>, &method_adaptor_ctor<0x21a>, &method_adaptor_ctor<0x21b>, &method_adaptor_ctor<0x21c>, &method_adaptor_ctor<0x21d>, &method_adaptor_ctor<0x21e>, &method_adaptor_ctor<0x21f>,
  &method_adaptor_ctor<0x220>, &method_adaptor_ctor<0x221>, &method_adaptor_ctor<0x222>, &method_adaptor_ctor<0x223>, &method_adaptor_ctor<0x224>, &method_adaptor_ctor<0x225>, &method_adaptor_ctor<0x226>, &method_adaptor_ctor<0x227>,
  &method_adaptor_ctor<0x228>, &method_adaptor_ctor<0x229>, &method_adaptor_ctor<0x22a>, &method_adaptor_ctor<0x22b>, &method_adaptor_ctor<0x22c>, &method_adaptor_ctor<0x22d>, &method_adaptor_ctor<0x22e>, &method_adaptor_ctor<0x22f>,
  &method_adaptor_ctor<0x230>, &method_adaptor_ctor<0x231>, &method_adaptor_ctor<0x232>, &method_adaptor_ctor<0x233>, &method_adaptor_ctor<0x234>, &method_adaptor_ctor<0x235>, &method_adaptor_ctor<0x236>, &method_adaptor_ctor<0x237>,
  &method_adaptor_ctor<0x238>, &method_adaptor_ctor<0x239>, &method_adaptor_ctor<0x23a>, &method_adaptor_ctor<0x23b>, &method_adaptor_ctor<0x23c>, &method_adaptor_ctor<0x23d>, &method_adaptor_ctor<0x23e>, &method_adaptor_ctor<0x23f>,
  &method_adaptor_ctor<0x240>, &method_adaptor_ctor<0x241>, &method_adaptor_ctor<0x242>, &method_adaptor_ctor<0x243>, &method_adaptor_ctor<0x244>, &method_adaptor_ctor<0x245>, &method_adaptor_ctor<0x246>, &method_adaptor_ctor<0x247>,
  &method_adaptor_ctor<0x248>, &method_adaptor_ctor<0x249>, &method_adaptor_ctor<0x24a>, &method_adaptor_ctor<0x24b>, &method_adaptor_ctor<0x24c>, &method_adaptor_ctor<0x24d>, &method_adaptor_ctor<0x24e>, &method_adaptor_ctor<0x24f>,
  &method_adaptor_ctor<0x250>, &method_adaptor_ctor<0x251>, &method_adaptor_ctor<0x252>, &method_adaptor_ctor<0x253>, &method_adaptor_ctor<0x254>, &method_adaptor_ctor<0x255>, &method_adaptor_ctor<0x256>, &method_adaptor_ctor<0x257>,
  &method_adaptor_ctor<0x258>, &method_adaptor_ctor<0x259>, &method_adaptor_ctor<0x25a>, &method_adaptor_ctor<0x25b>, &method_adaptor_ctor<0x25c>, &method_adaptor_ctor<0x25d>, &method_adaptor_ctor<0x25e>, &method_adaptor_ctor<0x25f>,
  &method_adaptor_ctor<0x260>, &method_adaptor_ctor<0x261>, &method_adaptor_ctor<0x262>, &method_adaptor_ctor<0x263>, &method_adaptor_ctor<0x264>, &method_adaptor_ctor<0x265>, &method_adaptor_ctor<0x266>, &method_adaptor_ctor<0x267>,
  &method_adaptor_ctor<0x268>, &method_adaptor_ctor<0x269>, &method_adaptor_ctor<0x26a>, &method_adaptor_ctor<0x26b>, &method_adaptor_ctor<0x26c>, &method_adaptor_ctor<0x26d>, &method_adaptor_ctor<0x26e>, &method_adaptor_ctor<0x26f>,
  &method_adaptor_ctor<0x270>, &method_adaptor_ctor<0x271>, &method_adaptor_ctor<0x272>, &method_adaptor_ctor<0x273>, &method_adaptor_ctor<0x274>, &method_adaptor_ctor<0x275>, &method_adaptor_ctor<0x276>, &method_adaptor_ctor<0x277>,
  &method_adaptor_ctor<0x278>, &method_adaptor_ctor<0x279>, &method_adaptor_ctor<0x27a>, &method_adaptor_ctor<0x27b>, &method_adaptor_ctor<0x27c>, &method_adaptor_ctor<0x27d>, &method_adaptor_ctor<0x27e>, &method_adaptor_ctor<0x27f>,
  &method_adaptor_ctor<0x280>, &method_adaptor_ctor<0x281>, &method_adaptor_ctor<0x282>, &method_adaptor_ctor<0x283>, &method_adaptor_ctor<0x284>, &method_adaptor_ctor<0x285>, &method_adaptor_ctor<0x286>, &method_adaptor_ctor<0x287>,
  &method_adaptor_ctor<0x288>, &method_adaptor_ctor<0x289>, &method_adaptor_ctor<0x28a>, &method_adaptor_ctor<0x28b>, &method_adaptor_ctor<0x28c>, &method_adaptor_ctor<0x28d>, &method_adaptor_ctor<0x28e>, &method_adaptor_ctor<0x28f>,
  &method_adaptor_ctor<0x290>, &method_adaptor_ctor<0x291>, &method_adaptor_ctor<0x292>, &method_adaptor_ctor<0x293>, &method_adaptor_ctor<0x294>, &method_adaptor_ctor<0x295>, &method_adaptor_ctor<0x296>, &method_adaptor_ctor<0x297>,
  &method_adaptor_ctor<0x298>, &method_adaptor_ctor<0x299>, &method_adaptor_ctor<0x29a>, &method_adaptor_ctor<0x29b>, &method_adaptor_ctor<0x29c>, &method_adaptor_ctor<0x29d>, &method_adaptor_ctor<0x29e>, &method_adaptor_ctor<0x29f>,
  &method_adaptor_ctor<0x2a0>, &method_adaptor_ctor<0x2a1>, &method_adaptor_ctor<0x2a2>, &method_adaptor_ctor<0x2a3>, &method_adaptor_ctor<0x2a4>, &method_adaptor_ctor<0x2a5>, &method_adaptor_ctor<0x2a6>, &method_adaptor_ctor<0x2a7>,
  &method_adaptor_ctor<0x2a8>, &method_adaptor_ctor<0x2a9>, &method_adaptor_ctor<0x2aa>, &method_adaptor_ctor<0x2ab>, &method_adaptor_ctor<0x2ac>, &method_adaptor_ctor<0x2ad>, &method_adaptor_ctor<0x2ae>, &method_adaptor_ctor<0x2af>,
  &method_adaptor_ctor<0x2b0>, &method_adaptor_ctor<0x2b1>, &method_adaptor_ctor<0x2b2>, &method_adaptor_ctor<0x2b3>, &method_adaptor_ctor<0x2b4>, &method_adaptor_ctor<0x2b5>, &method_adaptor_ctor<0x2b6>, &method_adaptor_ctor<0x2b7>,
  &method_adaptor_ctor<0x2b8>, &method_adaptor_ctor<0x2b9>, &method_adaptor_ctor<0x2ba>, &method_adaptor_ctor<0x2bb>, &method_adaptor_ctor<0x2bc>, &method_adaptor_ctor<0x2bd>, &method_adaptor_ctor<0x2be>, &method_adaptor_ctor<0x2bf>,
  &method_adaptor_ctor<0x2c0>, &method_adaptor_ctor<0x2c1>, &method_adaptor_ctor<0x2c2>, &method_adaptor_ctor<0x2c3>, &method_adaptor_ctor<0x2c4>, &method_adaptor_ctor<0x2c5>, &method_adaptor_ctor<0x2c6>, &method_adaptor_ctor<0x2c7>,
  &method_adaptor_ctor<0x2c8>, &method_adaptor_ctor<0x2c9>, &method_adaptor_ctor<0x2ca>, &method_adaptor_ctor<0x2cb>, &method_adaptor_ctor<0x2cc>, &method_adaptor_ctor<0x2cd>, &method_adaptor_ctor<0x2ce>, &method_adaptor_ctor<0x2cf>,
  &method_adaptor_ctor<0x2d0>, &method_adaptor_ctor<0x2d1>, &method_adaptor_ctor<0x2d2>, &method_adaptor_ctor<0x2d3>, &method_adaptor_ctor<0x2d4>, &method_adaptor_ctor<0x2d5>, &method_adaptor_ctor<0x2d6>, &method_adaptor_ctor<0x2d7>,
  &method_adaptor_ctor<0x2d8>, &method_adaptor_ctor<0x2d9>, &method_adaptor_ctor<0x2da>, &method_adaptor_ctor<0x2db>, &method_adaptor_ctor<0x2dc>, &method_adaptor_ctor<0x2dd>, &method_adaptor_ctor<0x2de>, &method_adaptor_ctor<0x2df>,
  &method_adaptor_ctor<0x2e0>, &method_adaptor_ctor<0x2e1>, &method_adaptor_ctor<0x2e2>, &method_adaptor_ctor<0x2e3>, &method_adaptor_ctor<0x2e4>, &method_adaptor_ctor<0x2e5>, &method_adaptor_ctor<0x2e6>, &method_adaptor_ctor<0x2e7>,
  &method_adaptor_ctor<0x2e8>, &method_adaptor_ctor<0x2e9>, &method_adaptor_ctor<0x2ea>, &method_adaptor_ctor<0x2eb>, &method_adaptor_ctor<0x2ec>, &method_adaptor_ctor<0x2ed>, &method_adaptor_ctor<0x2ee>, &method_adaptor_ctor<0x2ef>,
  &method_adaptor_ctor<0x2f0>, &method_adaptor_ctor<0x2f1>, &method_adaptor_ctor<0x2f2>, &method_adaptor_ctor<0x2f3>, &method_adaptor_ctor<0x2f4>, &method_adaptor_ctor<0x2f5>, &method_adaptor_ctor<0x2f6>, &method_adaptor_ctor<0x2f7>,
  &method_adaptor_ctor<0x2f8>, &method_adaptor_ctor<0x2f9>, &method_adaptor_ctor<0x2fa>, &method_adaptor_ctor<0x2fb>, &method_adaptor_ctor<0x2fc>, &method_adaptor_ctor<0x2fd>, &method_adaptor_ctor<0x2fe>, &method_adaptor_ctor<0x2ff>,
  &method_adaptor_ctor<0x300>, &method_adaptor_ctor<0x301>, &method_adaptor_ctor<0x302>, &method_adaptor_ctor<0x303>, &method_adaptor_ctor<0x304>, &method_adaptor_ctor<0x305>, &method_adaptor_ctor<0x306>, &method_adaptor_ctor<0x307>,
  &method_adaptor_ctor<0x308>, &method_adaptor_ctor<0x309>, &method_adaptor_ctor<0x30a>, &method_adaptor_ctor<0x30b>, &method_adaptor_ctor<0x30c>, &method_adaptor_ctor<0x30d>, &method_adaptor_ctor<0x30e>, &method_adaptor_ctor<0x30f>,
  &method_adaptor_ctor<0x310>, &method_adaptor_ctor<0x311>, &method_adaptor_ctor<0x312>, &method_adaptor_ctor<0x313>, &method_adaptor_ctor<0x314>, &method_adaptor_ctor<0x315>, &method_adaptor_ctor<0x316>, &method_adaptor_ctor<0x317>,
  &method_adaptor_ctor<0x318>, &method_adaptor_ctor<0x319>, &method_adaptor_ctor<0x31a>, &method_adaptor_ctor<0x31b>, &method_adaptor_ctor<0x31c>, &method_adaptor_ctor<0x31d>, &method_adaptor_ctor<0x31e>, &method_adaptor_ctor<0x31f>,
  &method_adaptor_ctor<0x320>, &method_adaptor_ctor<0x321>, &method_adaptor_ctor<0x322>, &method_adaptor_ctor<0x323>, &method_adaptor_ctor<0x324>, &method_adaptor_ctor<0x325>, &method_adaptor_ctor<0x326>, &method_adaptor_ctor<0x327>,
  &method_adaptor_ctor<0x328>, &method_adaptor_ctor<0x329>, &method_adaptor_ctor<0x32a>, &method_adaptor_ctor<0x32b>, &method_adaptor_ctor<0x32c>, &method_adaptor_ctor<0x32d>, &method_adaptor_ctor<0x32e>, &method_adaptor_ctor<0x32f>,
  &method_adaptor_ctor<0x330>, &method_adaptor_ctor<0x331>, &method_adaptor_ctor<0x332>, &method_adaptor_ctor<0x333>, &method_adaptor_ctor<0x334>, &method_adaptor_ctor<0x335>, &method_adaptor_ctor<0x336>, &method_adaptor_ctor<0x337>,
  &method_adaptor_ctor<0x338>, &method_adaptor_ctor<0x339>, &method_adaptor_ctor<0x33a>, &method_adaptor_ctor<0x33b>, &method_adaptor_ctor<0x33c>, &method_adaptor_ctor<0x33d>, &method_adaptor_ctor<0x33e>, &method_adaptor_ctor<0x33f>,
  &method_adaptor_ctor<0x340>, &method_adaptor_ctor<0x341>, &method_adaptor_ctor<0x342>, &method_adaptor_ctor<0x343>, &method_adaptor_ctor<0x344>, &method_adaptor_ctor<0x345>, &method_adaptor_ctor<0x346>, &method_adaptor_ctor<0x347>,
  &method_adaptor_ctor<0x348>, &method_adaptor_ctor<0x349>, &method_adaptor_ctor<0x34a>, &method_adaptor_ctor<0x34b>, &method_adaptor_ctor<0x34c>, &method_adaptor_ctor<0x34d>, &method_adaptor_ctor<0x34e>, &method_adaptor_ctor<0x34f>,
  &method_adaptor_ctor<0x350>, &method_adaptor_ctor<0x351>, &method_adaptor_ctor<0x352>, &method_adaptor_ctor<0x353>, &method_adaptor_ctor<0x354>, &method_adaptor_ctor<0x355>, &method_adaptor_ctor<0x356>, &method_adaptor_ctor<0x357>,
  &method_adaptor_ctor<0x358>, &method_adaptor_ctor<0x359>, &method_adaptor_ctor<0x35a>, &method_adaptor_ctor<0x35b>, &method_adaptor_ctor<0x35c>, &method_adaptor_ctor<0x35d>, &method_adaptor_ctor<0x35e>, &method_adaptor_ctor<0x35f>,
  &method_adaptor_ctor<0x360>, &method_adaptor_ctor<0x361>, &method_adaptor_ctor<0x362>, &method_adaptor_ctor<0x363>, &method_adaptor_ctor<0x364>, &method_adaptor_ctor<0x365>, &method_adaptor_ctor<0x366>, &method_adaptor_ctor<0x367>,
  &method_adaptor_ctor<0x368>, &method_adaptor_ctor<0x369>, &method_adaptor_ctor<0x36a>, &method_adaptor_ctor<0x36b>, &method_adaptor_ctor<0x36c>, &method_adaptor_ctor<0x36d>, &method_adaptor_ctor<0x36e>, &method_adaptor_ctor<0x36f>,
  &method_adaptor_ctor<0x370>, &method_adaptor_ctor<0x371>, &method_adaptor_ctor<0x372>, &method_adaptor_ctor<0x373>, &method_adaptor_ctor<0x374>, &method_adaptor_ctor<0x375>, &method_adaptor_ctor<0x376>, &method_adaptor_ctor<0x377>,
  &method_adaptor_ctor<0x378>, &method_adaptor_ctor<0x379>, &method_adaptor_ctor<0x37a>, &method_adaptor_ctor<0x37b>, &method_adaptor_ctor<0x37c>, &method_adaptor_ctor<0x37d>, &method_adaptor_ctor<0x37e>, &method_adaptor_ctor<0x37f>,
  &method_adaptor_ctor<0x380>, &method_adaptor_ctor<0x381>, &method_adaptor_ctor<0x382>, &method_adaptor_ctor<0x383>, &method_adaptor_ctor<0x384>, &method_adaptor_ctor<0x385>, &method_adaptor_ctor<0x386>, &method_adaptor_ctor<0x387>,
  &method_adaptor_ctor<0x388>, &method_adaptor_ctor<0x389>, &method_adaptor_ctor<0x38a>, &method_adaptor_ctor<0x38b>, &method_adaptor_ctor<0x38c>, &method_adaptor_ctor<0x38d>, &method_adaptor_ctor<0x38e>, &method_adaptor_ctor<0x38f>,
  &method_adaptor_ctor<0x390>, &method_adaptor_ctor<0x391>, &method_adaptor_ctor<0x392>, &method_adaptor_ctor<0x393>, &method_adaptor_ctor<0x394>, &method_adaptor_ctor<0x395>, &method_adaptor_ctor<0x396>, &method_adaptor_ctor<0x397>,
  &method_adaptor_ctor<0x398>, &method_adaptor_ctor<0x399>, &method_adaptor_ctor<0x39a>, &method_adaptor_ctor<0x39b>, &method_adaptor_ctor<0x39c>, &method_adaptor_ctor<0x39d>, &method_adaptor_ctor<0x39e>, &method_adaptor_ctor<0x39f>,
  &method_adaptor_ctor<0x3a0>, &method_adaptor_ctor<0x3a1>, &method_adaptor_ctor<0x3a2>, &method_adaptor_ctor<0x3a3>, &method_adaptor_ctor<0x3a4>, &method_adaptor_ctor<0x3a5>, &method_adaptor_ctor<0x3a6>, &method_adaptor_ctor<0x3a7>,
  &method_adaptor_ctor<0x3a8>, &method_adaptor_ctor<0x3a9>, &method_adaptor_ctor<0x3aa>, &method_adaptor_ctor<0x3ab>, &method_adaptor_ctor<0x3ac>, &method_adaptor_ctor<0x3ad>, &method_adaptor_ctor<0x3ae>, &method_adaptor_ctor<0x3af>,
  &method_adaptor_ctor<0x3b0>, &method_adaptor_ctor<0x3b1>, &method_adaptor_ctor<0x3b2>, &method_adaptor_ctor<0x3b3>, &method_adaptor_ctor<0x3b4>, &method_adaptor_ctor<0x3b5>, &method_adaptor_ctor<0x3b6>, &method_adaptor_ctor<0x3b7>,
  &method_adaptor_ctor<0x3b8>, &method_adaptor_ctor<0x3b9>, &method_adaptor_ctor<0x3ba>, &method_adaptor_ctor<0x3bb>, &method_adaptor_ctor<0x3bc>, &method_adaptor_ctor<0x3bd>, &method_adaptor_ctor<0x3be>, &method_adaptor_ctor<0x3bf>,
  &method_adaptor_ctor<0x3c0>, &method_adaptor_ctor<0x3c1>, &method_adaptor_ctor<0x3c2>, &method_adaptor_ctor<0x3c3>, &method_adaptor_ctor<0x3c4>, &method_adaptor_ctor<0x3c5>, &method_adaptor_ctor<0x3c6>, &method_adaptor_ctor<0x3c7>,
  &method_adaptor_ctor<0x3c8>, &method_adaptor_ctor<0x3c9>, &method_adaptor_ctor<0x3ca>, &method_adaptor_ctor<0x3cb>, &method_adaptor_ctor<0x3cc>, &method_adaptor_ctor<0x3cd>, &method_adaptor_ctor<0x3ce>, &method_adaptor_ctor<0x3cf>,
  &method_adaptor_ctor<0x3d0>, &method_adaptor_ctor<0x3d1>, &method_adaptor_ctor<0x3d2>, &method_adaptor_ctor<0x3d3>, &method_adaptor_ctor<0x3d4>, &method_adaptor_ctor<0x3d5>, &method_adaptor_ctor<0x3d6>, &method_adaptor_ctor<0x3d7>,
  &method_adaptor_ctor<0x3d8>, &method_adaptor_ctor<0x3d9>, &method_adaptor_ctor<0x3da>, &method_adaptor_ctor<0x3db>, &method_adaptor_ctor<0x3dc>, &method_adaptor_ctor<0x3dd>, &method_adaptor_ctor<0x3de>, &method_adaptor_ctor<0x3df>,
  &method_adaptor_ctor<0x3e0>, &method_adaptor_ctor<0x3e1>, &method_adaptor_ctor<0x3e2>, &method_adaptor_ctor<0x3e3>, &method_adaptor_ctor<0x3e4>, &method_adaptor_ctor<0x3e5>, &method_adaptor_ctor<0x3e6>, &method_adaptor_ctor<0x3e7>,
  &method_adaptor_ctor<0x3e8>, &method_adaptor_ctor<0x3e9>, &method_adaptor_ctor<0x3ea>, &method_adaptor_ctor<0x3eb>, &method_adaptor_ctor<0x3ec>, &method_adaptor_ctor<0x3ed>, &method_adaptor_ctor<0x3ee>, &method_adaptor_ctor<0x3ef>,
  &method_adaptor_ctor<0x3f0>, &method_adaptor_ctor<0x3f1>, &method_adaptor_ctor<0x3f2>, &method_adaptor_ctor<0x3f3>, &method_adaptor_ctor<0x3f4>, &method_adaptor_ctor<0x3f5>, &method_adaptor_ctor<0x3f6>, &method_adaptor_ctor<0x3f7>,
  &method_adaptor_ctor<0x3f8>, &method_adaptor_ctor<0x3f9>, &method_adaptor_ctor<0x3fa>, &method_adaptor_ctor<0x3fb>, &method_adaptor_ctor<0x3fc>, &method_adaptor_ctor<0x3fd>, &method_adaptor_ctor<0x3fe>, &method_adaptor_ctor<0x3ff>,
};

// --------------------------------------------------------------------------
//  RubyInterpreter implementation

static VALUE 
stdout_write (VALUE self, VALUE a)
{
  if (sp_current_console) {
    sp_current_console->write_str (StringValuePtr (a), Console::OS_stdout);
  }
  return Qnil;
}

static VALUE 
stdout_flush (VALUE self)
{
  //  nothing yet ..
  return Qnil;
}

static VALUE 
stderr_write (VALUE self, VALUE a)
{
  if (sp_current_console) {
    sp_current_console->write_str (StringValuePtr (a), Console::OS_stderr);
  }
  return Qnil;
}

static VALUE 
stderr_flush (VALUE self)
{
  //  nothing yet ..
  return Qnil;
}

struct RubyConstDescriptor 
{
  VALUE klass;
  int mid;
  std::string name;
};

extern "C" void ruby_prog_init();
static void
rba_init ()
{
  const char *program_name = "klayout";

  int argc = 3;
  char *argv[3];
  argv[0] = (char *)program_name;
  // Hint: to keep ruby_options from reading stdin, we simulate a "-e" option with an empty script
  argv[1] = (char *)"-e";
  argv[2] = (char *)"";

#if HAVE_RUBY_VERSION_CODE>=10900
  //  Make sure we call ruby_sysinit because otherwise the program will crash
  char **argvp = argv;
  int argcp = argc;
  ruby_sysinit (&argcp, &argvp);
#endif

  //  Keep ruby_init from changing the SIGINT handler.
  //  The reason for ruby_init doing this is probably to enable the input methods
  //  to receive Ctrl+C. We however do not need this functionality and users complain 
  //  about now having Ctrl+C to kill the application.
  void (*org_sigint) (int) = signal (SIGINT, SIG_DFL);
  ruby_init ();
  signal (SIGINT, org_sigint);

  //  Remove setters for $0 and $PROGRAM_NAME (still both are linked) because
  //  the setter does strange things with the process and the argv, specifically argv[0] above. 
  static VALUE argv0 = Qnil;
  argv0 = rb_gv_get("0");
  rb_define_hooked_variable("$0", &argv0, 0, 0);
  rb_define_hooked_variable("$PROGRAM_NAME", &argv0, 0, 0);
  
  ruby_options(argc, argv);

  typedef VALUE (*ruby_func)(ANYARGS);

  VALUE module = rb_define_module ("RBA");

  //  register the cline handler at the GSI side
  gsi::ObjectBase::register_client_handler (new RBAClientHandler ());

  //  save all constants for later (we cannot declare them while we are still producing classes
  //  because of the enum representative classes and enum constants are important)
  std::vector <RubyConstDescriptor> constants;

  bool more_classes = true;
  while (more_classes) {

    more_classes = false;
    for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {

      //  don't handle classes twice
      if (rev_cls_map.find (&*c) != rev_cls_map.end ()) {
        continue;
      }

      VALUE super = rb_cObject;
      if (c->base () != 0) {
        std::map <const gsi::ClassBase *, VALUE>::const_iterator cb = rev_cls_map.find (c->base ());
        if (cb == rev_cls_map.end ()) {
          //  can't produce this class yet. The base class needs to be handled first.
          more_classes = true;
          continue;
        } else {
          super = cb->second;
        }
      }

      //  there should be only main declarations since we merged
      tl_assert (c->declaration () == &*c);

      VALUE klass = rb_define_class_under (module, c->name ().c_str (), super);
      cls_map.insert (std::make_pair (klass, &*c));
      rev_cls_map.insert (std::make_pair (&*c, klass));

      rb_define_alloc_func (klass, alloc_proxy);

      bool has_default_initialize = false;

      if (c->can_copy ()) {
        rb_define_method (klass, "initialize_copy", (ruby_func) &gsi_init_copy, 1);
        rb_define_method (klass, "assign", (ruby_func) &gsi_init_copy, 1);
      }
      rb_define_method (klass, "create", (ruby_func) &gsi_create, 0);
      if (c->can_destroy ()) {
        rb_define_method (klass, "destroy", (ruby_func) &gsi_destroy, 0);
      }
      rb_define_method (klass, "destroyed?", (ruby_func) &gsi_destroyed, 0);
      // for backward compatibility:
      rb_define_method (klass, "destroyed", (ruby_func) &gsi_destroyed, 0);
      rb_define_method (klass, "is_const_object?", (ruby_func) &gsi_is_const, 0);

      unsigned int mo = c->method_offset ();

      //  Hint: we need to do static methods before the non-static ones because 
      //  rb_define_module_function creates an private instance method.
      //  If the do the non-static methods afterwards we will make it a public one again.
      //  The order of the names will be "name(non-static), name(static), ..." because 
      //  the static flag is the second member of the key (string, bool) pair.
      for (gsi::ClassBase::name_map_iterator nm = c->end_name_map (); nm != c->begin_name_map (); ) {

        --nm;

        unsigned int mid = nm->second;

        //  the first overloaded method determines the type of method - this implies that all
        //  overloaded methods must be static or non-static and protected or public.
        const gsi::MethodBase *m_first = c->begin_methods () [*c->begin_overloaded_methods (mid)];

        bool is_protected = false;
        extract_protected(m_first->name ().c_str (), is_protected);

        if (! m_first->is_callback ()) {

          tl_assert (mid + mo < sizeof (method_adaptors) / sizeof (method_adaptors[0]));
          if (! nm->first.second /*not static*/) {

            if (! is_protected) {
              rb_define_method (klass, nm->first.first.c_str (), (ruby_func) method_adaptors[mid + mo], -1);
            } else {
              rb_define_protected_method (klass, nm->first.first.c_str (), (ruby_func) method_adaptors[mid + mo], -1);
            }

          } else if (m_first->ret_type ().type == gsi::T_object_new && nm->first.first == "new") {

            if (m_first->ret_type ().type == gsi::T_object) {
              tl::warn << "'new' method " << c->name () << "::" << m_first->name () << " is not declared as a constructor";
            }

            has_default_initialize = true;
            if (! is_protected) {
              rb_define_method (klass, "initialize", (ruby_func) method_adaptors_ctor[mid + mo], -1);
            } else {
              rb_define_method (klass, "initialize", (ruby_func) method_adaptors_ctor[mid + mo], -1);
              rb_define_protected_method (klass, "initialize", (ruby_func) method_adaptors_ctor[mid + mo], -1);
            }

          } else if (isupper (nm->first.first [0])) {
            
            if (c->begin_overloaded_methods (mid) + 1 == c->end_overloaded_methods (mid) && m_first->begin_arguments () == m_first->end_arguments ()) {

              //  static methods without arguments which start with a capital letter are treated as constants
              constants.push_back (RubyConstDescriptor ());
              constants.back ().klass = klass;
              constants.back ().mid = mid + mo;
              constants.back ().name = nm->first.first;

            } else {
              tl::warn << "Upper case method name encountered which cannot be used as a constant (more than one overload or at least one argument: " << c->name () << "::" << nm->first.first;
            }

          } else {

            if (is_protected) {
              tl::warn << "'new' method cannot be protected in class " << c->name ();
            }

            rb_define_module_function (klass, nm->first.first.c_str (), (ruby_func) method_adaptors[mid + mo], -1);

          }

        }

      }

      if (! has_default_initialize) {
        rb_define_method (klass, "initialize", (ruby_func) &gsi_init, 0);
      }

    }

  }

  //  now make the constants
  for (std::vector <RubyConstDescriptor>::const_iterator c = constants.begin (); c != constants.end (); ++c) {
    VALUE argv [1] = { Qnil };
    VALUE ret = (*method_adaptors[c->mid]) (0, argv, c->klass);
    rb_define_const (c->klass, c->name.c_str (), ret);
  }

  //  define new classes for handling stdout and stderr.
  //  use IO as the basis for that.
  stdout_klass = rb_define_class_under (module, "RBAstdout", rb_cIO);
  rb_define_method (stdout_klass, "write", (ruby_func) &stdout_write, 1);
  rb_define_method (stdout_klass, "flush", (ruby_func) &stdout_flush, 0);
  stderr_klass = rb_define_class_under (module, "RBAstderr", rb_cIO);
  rb_define_method (stderr_klass, "write", (ruby_func) &stderr_write, 1);
  rb_define_method (stderr_klass, "flush", (ruby_func) &stderr_flush, 0);

  //  register the saved stdout/stderr objects with the garbage collector
  rb_global_variable (& saved_stdout);
  rb_global_variable (& saved_stderr);

  //  create the handler objects
  VALUE empty_args [] = { INT2NUM (0) };
  saved_stderr = rb_class_new_instance_checked (1, empty_args, stderr_klass);
  saved_stdout = rb_class_new_instance_checked (1, empty_args, stdout_klass);
}

RubyInterpreter::RubyInterpreter () 
{ 
  tl_assert (! rba_ruby_destroyed);
  tl_assert (! rba_interpreter);
  rba_interpreter = this;
  rba_init ();

  //  hopefully this one always remains there:
  ruby_top_self = rb_eval_string ("self");
  rb_gc_register_address (&ruby_top_self);
}

RubyInterpreter::~RubyInterpreter () 
{
  rb_gc_unregister_address (&ruby_top_self);
  ruby_top_self = Qnil;

  ruby_finalize ();
  gsi::ObjectBase::register_client_handler (0);
  rba_ruby_destroyed = true;
  rba_interpreter = 0;
}

std::string
RubyInterpreter::version () const
{
  try {
    return const_cast<RubyInterpreter *> (this)->eval_as_string ("RUBY_VERSION.to_s+'-p'+RUBY_PATCHLEVEL.to_s+' ('+RUBY_PLATFORM+')'");
  } catch (...) {
    return std::string ("unknown");
  }
}

void 
RubyInterpreter::initialize ()
{
  // Hint: this macro must be called from as high as possible on the stack. 
  // Ruby does not work properly if any ruby function is called from above the
  // stack. Therefore this method must be called as high as possible in the call
  // stack (i.e. from main).
  // One effect of the fail is that a "require" does not work properly and will
  // abort with a segmentation fault.
#if defined(RUBY_INIT_STACK) // for ruby <=1.8.5.
  RUBY_INIT_STACK;
#endif
}

void 
RubyInterpreter::set_debugger_scope (const std::string &filename)
{
  s_debugger_scope = filename;
}

void 
RubyInterpreter::remove_debugger_scope ()
{
  s_debugger_scope.clear ();
}

void 
RubyInterpreter::ignore_next_exception ()
{
  if (sp_current_exec_handler) {
    s_ignore_next_exception = true;
  }
}

void 
RubyInterpreter::add_path (const std::string &path)
{
  VALUE pv = rb_gv_get ("$:");
  if (pv != Qnil && TYPE (pv) == T_ARRAY) {
    rb_ary_push (pv, rb_str_new (path.c_str (), path.size ()));
  }
}

void
RubyInterpreter::require (const std::string &filename_utf8)
{
  RUBY_BEGIN_EXEC 

    std::string fl (rb_cstring_from_utf8 (filename_utf8));

    rb_set_errinfo (Qnil);
    int error = 0;
    rb_protect_init (); // see above
    rb_protect ((VALUE (*)(VALUE))rb_require, (VALUE)fl.c_str (), &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC 
}

void
RubyInterpreter::load_file (const std::string &filename_utf8)
{
  RUBY_BEGIN_EXEC 

    std::string fl (rb_cstring_from_utf8 (filename_utf8));

    set_progname (rb_str_new (fl.c_str (), fl.size ()));

    rb_set_errinfo (Qnil);
    int error = 0;
    int wrap = 0; // TODO: make variable?
    rb_protect_init (); // see above
    rb_load_protect (rb_str_new (fl.c_str (), fl.size ()), wrap, &error);
    if (error) {
      rba_check_error ();
    }

  RUBY_END_EXEC 
}

void
RubyInterpreter::eval_string (const char *expr, const char *file, int line)
{
  do_eval_string (expr, file, line);
}

std::string
RubyInterpreter::eval_as_string (const char *expr, const char *file, int line)
{
  VALUE res = do_eval_string (expr, file, line);
  if (res != Qnil) {
    VALUE res_s = rba_safe_obj_as_string (res);
    return std::string (StringValuePtr (res_s));
  } else {
    return std::string ();
  }
}

void
RubyInterpreter::eval_string_and_print (const char *expr, const char *file, int line)
{
  VALUE res = do_eval_string (expr, file, line);
  if (sp_current_console && res != Qnil) {
    VALUE res_s = rba_safe_obj_as_string (res);
    sp_current_console->write_str (StringValuePtr (res_s), Console::OS_stdout);
    sp_current_console->write_str ("\n", Console::OS_stdout);
  }
}

void
RubyInterpreter::define_variable (const std::string &name, const std::string &value)
{
  rb_gv_set (name.c_str (), rb_str_new (value.c_str (), value.size ()));
}

bool
RubyInterpreter::ruby_available () const
{
  return true;
}

void 
RubyInterpreter::push_console (Console *console)
{
  if (! sp_current_console) { 
    std::swap (saved_stderr, rb_stderr);
    std::swap (saved_stdout, rb_stdout);
  } else {
    s_consoles.push_back (sp_current_console);
  }

  sp_current_console = console;
}

void 
RubyInterpreter::remove_console (Console *console)
{
  if (sp_current_console == console) {

    if (s_consoles.empty ()) {
      sp_current_console = 0;
      std::swap (saved_stderr, rb_stderr);
      std::swap (saved_stdout, rb_stdout);
    } else {
      sp_current_console = s_consoles.back ();
      s_consoles.pop_back ();
    }

  } else {

    for (std::vector<Console *>::iterator c = s_consoles.begin (); c != s_consoles.end (); ++c) {
      if (*c == console) {
        s_consoles.erase (c);
        break;
      }
    }

  }
}

static size_t
prepare_trace (const char *fn)
{
  s_in_trace = true;
  if (s_current_binding != Qnil) {
    rb_gc_unregister_address (&s_current_binding);
    s_current_binding = Qnil;
  }

  std::map<const char *, size_t>::const_iterator f = s_file_id_map.find (fn);
  if (f == s_file_id_map.end ()) {
    f = s_file_id_map.insert (std::make_pair (fn, sp_current_exec_handler->id_for_path (fn))).first;
  } 

  return f->second;
}

static void 
release_binding ()
{
  s_in_trace = false;
  if (s_current_binding != Qnil) {
    rb_gc_unregister_address (&s_current_binding);
    s_current_binding = Qnil;
  }
}

static void
#if HAVE_RUBY_VERSION_CODE<10900
trace_callback (rb_event_t event, NODE *, VALUE self, ID id, VALUE klass)
#else
trace_callback (rb_event_flag_t event, VALUE data, VALUE self, ID id, VALUE klass)
#endif
{
  if (sp_current_exec_handler && !s_in_trace) {

    if ((event & RUBY_EVENT_LINE) != 0) {

      //  see below for a description of s_block_exceptions
      s_block_exceptions = false;

      int line = rb_sourceline ();
      size_t file_id = prepare_trace (rb_sourcefile ());

      RBA_TRY

        if (s_exit_on_next) {
          release_binding ();
          throw ExitException (0);
        }

        try {
          RubyStackTraceProvider st_provider (s_debugger_scope);
          sp_current_exec_handler->trace (file_id, line, &st_provider);
          release_binding ();
        } catch (...) {
          release_binding ();
          throw;
        }

      RBA_CATCH("trace callback")

    } else if ((event & RUBY_EVENT_CALL) != 0) {
      sp_current_exec_handler->push_call_stack ();
    } else if ((event & RUBY_EVENT_RETURN) != 0) {
      sp_current_exec_handler->pop_call_stack ();
    } else if ((event & RUBY_EVENT_RAISE) != 0 && ! s_block_exceptions) {

      VALUE lasterr = rb_errinfo ();

      //  Don't catch syntax errors (it does not make sense to stop in the debugger for them).
      //  Sometimes lasterr is nil - we must suppress these errors as well.
      if (lasterr != Qnil && CLASS_OF (lasterr) != rb_eSyntaxError) {

        //  If the next exception shall be ignored, do so
        if (s_ignore_next_exception) {
          s_ignore_next_exception = false;
        } else {

          try {

            int line = rb_sourceline ();
            size_t file_id = prepare_trace (rb_sourcefile ());

            std::string eclass = "<unknown>";
            std::string emsg = "<unknown>";

            VALUE klass = rb_class_path (CLASS_OF (lasterr));
            eclass = std::string (RSTRING_PTR(klass), RSTRING_LEN(klass));
            VALUE message = rb_obj_as_string(lasterr);
            emsg = std::string (RSTRING_PTR(message), RSTRING_LEN(message));

            RubyStackTraceProvider st_provider (s_debugger_scope);
            sp_current_exec_handler->exception_thrown (file_id, line, eclass, emsg, &st_provider);

            release_binding ();

          } catch (ExitException &) {
            release_binding ();
            s_exit_on_next = true; // delayed handling of ExitException
          } catch (...) {
            //  ignore other exceptions - we can't raise a ruby exception from the trace callback
            //  because this causes a fatal error (exception reentered)
            release_binding ();
          }

        }

        //  Ruby tends to call this callback twice - once from rb_f_raise and then 
        //  from rb_exc_raise. We use the s_block_exceptions flag to suppress the 
        //  second one
        s_block_exceptions = true;

      }

    }

  }

}

void 
RubyInterpreter::push_exec_handler (ExecutionHandler *h)
{
  if (sp_current_exec_handler) {

    s_exec_handlers.push_back (sp_current_exec_handler);

  } else {

    rb_remove_event_hook(trace_callback);
#if HAVE_RUBY_VERSION_CODE<10900
    rb_add_event_hook(trace_callback, RUBY_EVENT_ALL);
#else
    rb_add_event_hook(trace_callback, RUBY_EVENT_ALL, Qnil);
#endif

  }

  sp_current_exec_handler = h;
  s_file_id_map.clear ();

  //  if we happen to push the exec handler inside the execution, 
  //  signal start of execution
  if (s_current_exec_level > 0) {
    sp_current_exec_handler->start_exec ();
  }
}

void 
RubyInterpreter::remove_exec_handler (ExecutionHandler *exec_handler)
{
  if (sp_current_exec_handler == exec_handler) {

    //  if we happen to remove the exec handler inside the execution, 
    //  signal end of execution
    if (s_current_exec_level > 0) {
      sp_current_exec_handler->end_exec ();
    }

    if (s_exec_handlers.empty ()) {
      sp_current_exec_handler = 0;
      rb_remove_event_hook(trace_callback);
    } else {
      sp_current_exec_handler = s_exec_handlers.back ();
      s_exec_handlers.pop_back ();
    }

  } else {

    for (std::vector<ExecutionHandler *>::iterator eh = s_exec_handlers.begin (); eh != s_exec_handlers.end (); ++eh) {
      if (*eh == exec_handler) {
        s_exec_handlers.erase (eh);
        break;
      }
    }

  }

}

}

#else

namespace rba
{

RubyInterpreter::RubyInterpreter () 
{ 
  // .. nothing ..
}

RubyInterpreter::~RubyInterpreter () 
{
  // .. nothing ..
}

void 
RubyInterpreter::add_path (const std::string &path)
{
  // .. nothing ..
}

void
RubyInterpreter::require (const std::string &filename)
{
  // .. nothing ..
}

void 
RubyInterpreter::set_debugger_scope (const std::string &filename)
{
  // .. nothing ..
}

void 
RubyInterpreter::remove_debugger_scope ()
{
  // .. nothing ..
}

void 
RubyInterpreter::ignore_next_exception ()
{
  // .. nothing ..
}

void
RubyInterpreter::load_file (const std::string &filename)
{
  // .. nothing ..
}

void
RubyInterpreter::eval_string (const char *expr, const char *file, int line)
{
  // .. nothing ..
}

void
RubyInterpreter::eval_string_and_print (const char *expr, const char *file, int line)
{
  // .. nothing ..
}

void
RubyInterpreter::define_variable (const std::string &name, const std::string &value)
{
  // .. nothing ..
}

bool
RubyInterpreter::ruby_available () const
{
  return false;
}

void 
RubyInterpreter::initialize ()
{
  // .. nothing ..
}

void 
RubyInterpreter::push_exec_handler (ExecutionHandler *exec_handler)
{
  // .. nothing ..
}

void
RubyInterpreter::remove_exec_handler (ExecutionHandler *exec_handler)
{
  // .. nothing ..
}

void 
RubyInterpreter::push_console (Console *console)
{
  // .. nothing ..
}

void 
RubyInterpreter::remove_console (Console *console)
{
  // .. nothing ..
}

std::string
RubyInterpreter::version () const
{
  return std::string ();
}

std::string
RubyInterpreter::eval_as_string (const char *expr, const char *file, int line)
{
  return std::string ();
}

}

#endif

